/*
DECAF (Distributed, Environment-Centered, Agent Framework) is a toolkit 
which allows a well-defined software engineering approach to building 
multi-agent systems. It is essentially, the internal "operating system" 
of a software agent, to which application programmers have strictly 
limited access.

This material is based upon work supported by the National Science 
Foundation under Grant No. IIS-9812764. Any opinions, findings and 
conclusions or recomendations expressed in this material are those of 
the authors and do not necessarily reflect the views of the National 
Science Foundation (NSF).

Copyright (C) 2002-3 University of Delaware

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For more information, visit http://www.eecis.udel.edu/~decaf/, 
or contat by e-mail at decaf@eecis.udel.edu
*/

import java.io.*;
import java.util.*;
import java.net.*;


public class Agent extends AgentDef{

    //    FileWriter fwtemp;

	/** Stores the command line arguments in a static variable and creates a new agent. */
	public static void main(String[] args) {
		CommandLine = args;
		Agent myAgent = new Agent();
	}

	/**
	  * Properly starts up the agent, making decisions and filling in data based on command
	  * line values and defaults.
	  */
	public Agent() {


	    //try {
	    //
	    //    fwtemp = new FileWriter("kqmlOut", true);
	    //} catch (IOException e) {
	    //System.out.println(e);
	    //}
 
		boolean help = (Boolean.valueOf(parseCommandLine("-h", true))).booleanValue();
		if (help) {
			System.out.println("A variety of command line options are available:\n" +
			       "-agn\t<Agent Name>\n-agh\t<Agent Host>\n-agp\t<Agent Port>\n" +
			       "-anh\t<Agent Name Server Host>\n-anp\t<ANS Port>\n" +
			       "-p\t<plan file>\n-l\t<log file>\n" +
			       "-era\t<Error Reporting Agent Name>\n" +
			       "-gui\tInclude this switch to start the gui with other options.\n" + 
				   "-go\tOnly checked when -gui is used.  Starts Agent without setup window.\n");
			System.exit(0);
		}

		Util.setAgent(this);
		this.debugMsgs = new LinkedListQ();

		String temp;
		//if true for -gui set boolean
		boolean guiOn = false;
		guiOn = (Boolean.valueOf(parseCommandLine("-gui", true))).booleanValue();

		String AgentHost = null;
		temp = parseCommandLine("-agh");
		if (temp == null) {
			try {
				AgentHost = InetAddress.getLocalHost().getHostName();
			} catch (UnknownHostException e) {
				AgentHost = "localHost";
			}
		} else {
			AgentHost = temp;
		}

		String ANShost = null;
		temp = parseCommandLine("-anh");
		if (temp == null) {
			try {
				ANShost = InetAddress.getLocalHost().getHostName();
			} catch (UnknownHostException e) {
				ANShost = "localHost";
			}
		} else {
			ANShost = temp;
		}

		int AgentPort = 0;
		try {
			AgentPort = startListening(Integer.parseInt(parseCommandLine("-agp")));
		} catch (java.lang.NumberFormatException e) {}
		if (AgentPort == 0) {
			AgentPort = startListening(4000);
		}

		int ANSport = 0;
		try {
			ANSport = Integer.parseInt(parseCommandLine("-anp"));
		} catch (java.lang.NumberFormatException e) {}
		if ( ANSport == 0) {
			ANSport = 6677;
		}

		String AgentName;
		temp = parseCommandLine("-agn");
		if (temp == null) {
			StringTokenizer tokens = new StringTokenizer(AgentHost, ".", false);
			AgentName = new String(tokens.nextToken() + "Ag" + AgentPort);
		} else {
			AgentName = temp;
		}
	
		String PlanFile;
		temp = parseCommandLine("-p");
		if (temp == null) {
			PlanFile = "agent.lsp";
		} else {
			PlanFile = temp;
		}

		String LogFile;
		temp = parseCommandLine("-l");
		if (temp == null) {
			LogFile = "";
		} else {
			LogFile = temp;
		}

		String ERAname = parseCommandLine("-era");
      
		// Intialize regardless of GUI or non-GUI
		agentPlan = new Plan(debugMsgs);
		setupDebugSwitch();
		
		// If GUI then open gui window otherwise continue with agent startup
		if (guiOn || CommandLine.length == 0) {
			// GUI
			gui = new AgentWindow(this, debugSwitch, PlanFile, LogFile, ANShost, ANSport, 
								AgentHost, AgentPort, AgentName, ERAname);
			notInitialized = false;
			String setupQ = parseCommandLine("-go", true);
			if (setupQ.equalsIgnoreCase("true")) {
				gui.setupWithoutWindow();
			}
		} else {
			// no GUI
			gui = null;
			this.setAgent(AgentName, AgentHost, AgentPort, PlanFile);
			this.initialize();
	
			//See if logging is on, so we can start as early as possible
			if (!LogFile.equals("")) {
				OutMsgs.startLogging(LogFile);
			}
		
			//register with ANS
			if (registerWithANS(ANShost, ANSport)) { 
				DebugAgent("Registered");
			} else {
				DebugAgent("Could not Register with ANS - Exiting");
				System.exit(1);
			}

			if (ERAname != null) {
				setERAName(ERAname);
			}
		}
	}

	//== Variable Initialization Functions ==//
	
	/** Starts up all the internal DECAF threads */
	public synchronized void initialize() {
		// Initialize all the Data Strucutres
		DebugAgent("Initializing Data Structures....");
		Objectives = new LinkedListQ();
		DebugAgent(" \nObjectives ...");
		ActionRes = new LinkedListQ();
		DebugAgent(" \nActionRes ...");
		Agenda = new LinkedListQ();
		DebugAgent(" \nAgenda ...");
		DebugAgent(" \nPending Queue...");
		TaskQueue = new LinkedListQ();
		DebugAgent(" \nTask Queue...");
		Pending = new Hashtable();
		DebugAgent(" \nincomingQueue ...");
		incomingQueue = new LinkedListQ();
		SchdQ = new LinkedListQ();
		WITQ = new LinkedListQ();

		DebugAgent(" D.S. Initialization Complete!\n");
		// Initializing all neccessary Threads
		// Starting Output Messages Thread
		DebugAgent("Starting Agents Threads....");
		OutMsgs = new MsgOutput(gui, debugSwitch, debugMsgs, DebuggerSem);
		OutMsgs.start();
		DebugAgent(" \nMsgOutput ...");

		// Starting the Dispatcher Thread
		DispatcherThread Dispatcher = new DispatcherThread(incomingQueue, 
									Objectives, Agenda, Pending, this, 
									DispatcherSem, PlannerSem, ExecutorSem);
		Dispatcher.start();
		DebugAgent(" \nDispatcher Started ...");

		// Starting the Executor thread
		Executor Execute = new Executor(this, Agenda, ActionRes, Pending, 
                                        ExecutorSem, AgendaManagerSem);
		Execute.start();
		DebugAgent("\nExecution Thread Started ...");

		// Starting the Planner thread
		PlannerThread PL = new PlannerThread(this, Objectives, SchdQ, WITQ, taskTemplates, 
											PlannerSem, SchedSem, GPGPSem);
		PL.start();
		DebugAgent("\nPlanner Thread Started...");

		// Starting the GPGP Thread
		GPGPModule GPGP = new GPGPModule(this, WITQ, SchdQ, GPGPSem, SchedSem); 
        GPGP.start();
        DebugAgent("\nGPGP Thread Started...");
        
        //Starting Scheduler Thread
		Scheduler TS = new Scheduler(this, SchdQ, TaskQueue, WITQ,
								SchedSem, AgendaManagerSem, GPGPSem );
		TS.start();
		DebugAgent("Schedueler Started.");

		//Starting Agenda Manager Thread
		AgendaManager SC = new AgendaManager(this, TaskQueue, ActionRes, Agenda, 
											AgendaManagerSem, ExecutorSem);
		SC.start();

		DebugAgent("\nAgenda Manager thread Started...");
		DebugAgent("\nAgent Initialisation Complete!!!!\n");
	}

	/** Intializes a default set of debug switches. */
	private void setupDebugSwitch() {
		debugSwitch.put(new String("INQ_MSG"), new DebugSwitchCell("Incoming Messages", "INQ_MSG", true));
		debugSwitch.put(new String("OUTMESSAGE"), new DebugSwitchCell("Outgoing Messages", "OUTMESSAGE", true));
		debugSwitch.put(new String("EXCEPTION"), new DebugSwitchCell("Exceptions", "EXCEPTION", true));
		debugSwitch.put(new String("DISPATCHER"), new DebugSwitchCell("Dispatcher", "DISPATCHER", true));
		debugSwitch.put(new String("PLANNER"), new DebugSwitchCell("Planner", "PLANNER", true));
		debugSwitch.put(new String("SCHEDULER"), new DebugSwitchCell("Scheduler", "SCHEDULER", true));
		debugSwitch.put(new String("EXECUTOR"), new DebugSwitchCell("Executor", "EXECUTOR", true));
		debugSwitch.put(new String("EXECUTOR_THREAD"), new DebugSwitchCell("Executor Thread", "EXECUTOR_THREAD", true));
		debugSwitch.put(new String("AGENT"), new DebugSwitchCell("Agent", "AGENT", true));
		debugSwitch.put(new String("AGENDAMANAGER"), new DebugSwitchCell("Agenda Manager", "AGENDAMANAGER", true));
	}
	
	/** Synchronizes the internal agent variables to those set elsewhere. */
	public synchronized void setAgent(String Name, String Host, int Port, String file) {
		agentName = Name;
		agentHost = Host;
		agentPort = Port;
		setTaskTemplate(file);
		DebugAgent( "Agent parameters are set.\nAgent Name: " + agentName + 
					"\nAgent Port: " + agentPort + "\nAgent Host: " + agentHost + 
					"\nPlan file " + agentPlan.getPlanFile());
    }
    
	/** Syncs the internal ANS variables with ones set elsewhere */
	public synchronized void setAns(String Name, int Port) {
		ansHost = Name;
		ansPort = Port;
		DebugAgent("ANS parameter are set.\n ANS Server: " + ansHost + " ANS Port: " + ansPort + "\n");
	}
	
	/** Creates a taskTemplates hashtable from the given file.*/
	public void setTaskTemplate(String file) {
		agentPlan.setPlanFile(file);
		agentPlan.readFile();
		this.taskTemplates = agentPlan.gettaskTemplate();
	}
	
	/** Sends out the startup task for the agent. */
	public void runStartupTask() {
		/* uncomment this block for RMA
		if (gui == null) {
			this.mrm = new MRM(this, false);
		} else {
			this.mrm = new MRM(this, true);
			//load the Resource Manager Plans.
			this.addTaskTemplate("RMA.lsp");
		}
		*/
		String StartupKQML = new String();
		StartupKQML = "(achieve " + ":sender " + agentName + " " + ":receiver " + agentName + " "
					+ ":language DECAF " + ":content (:task Startup))";
		if (this.taskTemplates.get("_Startup") != null) {
			incomingQueue.add (new KQMLmsg(StartupKQML));
			DispatcherSem.V();
		}        
	}
	
	////==== Functions to Start Connections with Agent System ====////

	/** Creates a socket for the agent to listen to. */
	public int startListening(int port) {
		boolean foundport = false;
		int count = 999900;
		int i = 0;
		while (!foundport && (i < count) ) {
			try {
				listen_socket = new ServerSocket(port);
				foundport = true;
			} catch(IOException e10) {
				i++;
				port++;
				if (port > 8000) port = 4000;
			}
			if (i >= count) {
				System.out.println("Exception Creating server socket");
				System.exit(1);
			} // end
		} // end while loop    
		return port;
	}
	
	/**
	 * Registers with an Agent Name Server.
	 * @param x The host address of the new name server.
	 * @param y The port for the name server.
	 */
	public synchronized boolean registerWithANS(String x, int y) {
		if (isRegistered) {
			this.unregister();                  // & unregister with old ANS
			this.stopListening();               // if already registered, let go of old port
			this.isRegistered = false;                  // not in control of the the agent
		} // end if already registered

		setAns(x, y); 

		if (listen_socket != null) {
			try {
				this.listen_socket.close();   //unlock socket so listening thread can get it
			} catch (IOException e) {}
		}

		boolean proceed = false;
		try {
			ansComm = new CommunicateWithANS(ansHost, ansPort);
			String stat = ansComm.register(agentName, agentHost, agentPort);
			if(stat.equals("success")) proceed = true;
		} catch(Exception e) {
			DebugAgent("ANS not found. Cannot register with ANS.\n Please check ANS parameter and try again\n");
			DebugAgent(e + "\n");
			DebugAgent("ANS not found");
		}	    
	
		if (proceed) {
			isRegistered = true;
			// Start listening to the port
			ListenThread = new Receive(incomingQueue, agentPort, debugMsgs, DispatcherSem);
			runStartupTask();
			// end if can go through
		} else {
			isRegistered = false;
			DebugAgent("Couldn't register with ANS");
		}
		DebugAgent("Registered with ANS");

		if (isRegistered) {
			return true;
		} else {
			return false;
		}
	}
	
	////==== Functions to close connection with system ====////

	/** Stops listeneing to the current port and cleans up */
	public void stopListening() {
		ListenThread.cleanUp();
	}

	/** Unregisters with the ANS */
	public void unregister() {
		try {
			String stat = ansComm.unregister(agentName);
		} catch (Exception e) { }
		DebugAgent("Unregistered");  
	}

	/** Sends the shutdown task to the agent and then exits when done. */
	public void shutdownAgent() {
		if (taskTemplates != null && taskTemplates.get("_Shutdown") != null) {
			String ShutdownKQML = "(achieve " + ":sender " + agentName + " " + ":receiver "
						+ agentName + " " + ":language Java " + ":content (:task Shutdown))";
			incomingQueue.add(new KQMLmsg(ShutdownKQML));
			DispatcherSem.V();        
		} else {
			this.exit();
		}
	}
    
	/** Unregisters from the ANS and actually exits. */
	public void exit() {
		if (eraName != null && !eraName.equals("")) {
			unsetERAname();
		}
		if (isRegistered) this.unregister();
		System.exit(0);
	}
	
	////==== Functions to send messages ====////
	
	/** 
	  * Sends the given String message, after converting to a KQMLmsg object.
	  */
	public void send(String message) {   
		KQMLmsg kqml = new KQMLmsg(message);
		send(kqml);
	}

	/** Sends the given KQML object. */
	public void send(KQMLmsg kqml) {
		if (!kqml.getValue("reply-with").equals("")) { 
		    SendQ.add(kqml.toString()); //don't send the message yet.
		} else {
			sendNow(kqml.toString());
		}
	}
	
	/** Sends all the messages in the SendQ */
	public void sendMessages() {
		while (!SendQ.isEmpty()) {
			sendNow(((String )SendQ.removeHead()));
		}
	}

	/** Sends the given message to itself. */
	public void sendLocal(KQMLmsg k) {
		k.addFieldValuePair("sender", this.getName());
		k.addFieldValuePair("receiver", this.getName());
		incomingQueue.add(k);
		DispatcherSem.V(); //dispatcher notify
	}

	/** Sends an error message. */
	public void sendError(KQMLmsg k, String error) {
		KQMLmsg j = k;
		j.addFieldValuePair("performative", "error");
		j.addFieldValuePair("content", error);
		j.addFieldValuePair("receiver", k.getValue("sender"));
		j.addFieldValuePair("sender", k.getValue("receiver"));
		j.addFieldValuePair("in-reply-to", k.getValue("reply-with"));
		j.addFieldValuePair("reply-with", "");
		send(j.getKQMLString());
		this.DebugAgent("Sent error: " + j.getKQMLString());
	}

	/** 
	  * Converts a message to a KQMLmsg object and then calls sendNow(KQMLmsg).
	  * @deprecated Users requiring immediate sending should call sendNow(KQMLmsg).
	  */
	public void sendNow(String message) {
		KQMLmsg kqml = new KQMLmsg(message);
		sendNow(kqml);
	}
	/** Actually performs the task of sending a message. */
	public void sendNow(KQMLmsg kqml) {

	    String receiver = kqml.getValue("receiver");
	    String sender = kqml.getValue("sender");

	    /*TH for counting msgs sent between agents*/
	    if ( ! sender.equals(receiver) ){
	    //try{
	    //	fwtemp.write(kqml.toString());
	    //	fwtemp.write("\n\n");
	    //	fwtemp.flush();
	    //} catch (IOException e){}
	    //kqml.incrementExtCount(); //MsgCount.incrMsgCount();//
	    }
		// Get receiver's name and socket
	    Location toAgent = null;
	    Socket rAgent = null;

		// Lookup receiver's Location
		try {
			//toAgent = ansComm.lookup(receiver);
			toAgent = ansComm.lookupInCache(receiver);
		} catch (Exception ea) {
			DebugExceptionAgent("No ANS found:\n" + ea);
		}
		DebugAgent("Lookup with cache is OK for getting "+receiver+"\n");

		boolean BadAddress = true;

		if(toAgent != null && toAgent.isValidAgentAddress()) {
			BadAddress = false;
			int Port = toAgent.getPort();
			String Host = toAgent.getHostName();

			// Trying to set the socket to talk to the receiver
			try {
				rAgent = new Socket(Host, Port);
			} catch(Exception MayHaveMoved) {
				// Cached address may not be reliable
				// retry with surely a loopup
				BadAddress = true;
				DebugAgent(agentName + " has bad cached address for " + receiver +
							"\nThe Port is " + Port + " and the Host is " + Host +
							"\n" + MayHaveMoved);
				ansComm.removeName(receiver);
				try {
					toAgent = ansComm.lookup(receiver);
				} catch (Exception ea) {
					DebugExceptionAgent("No ANS found:\n" + ea);
				}
				DebugExceptionAgent("Lookup without cache is OK for "+receiver+"!!!\n");
				System.out.println("Lookup without cache is OK for "+receiver+"!!!\n");
				if(toAgent==null) {
					System.out.println("Except toAgent was null");
				}
				if(!toAgent.isValidAgentAddress()) {
					System.out.println("Except agent address was invalid.");
				}
	
				if(toAgent != null && toAgent.isValidAgentAddress()) {
					Port = toAgent.getPort();
					Host = toAgent.getHostName();
	
					// Try to set the socket to talk to the receiver
					try {
						rAgent = new Socket(Host, Port);
						BadAddress = false;
					} catch (UnknownHostException eh) {
						DebugExceptionAgent("Host unknown:" + eh.toString());
					} catch (IOException ioe) {
						DebugExceptionAgent("IOException:" + ioe.toString());
					}
				}
			}
		}
	
		if(!BadAddress) {
			DebugAgent("Socket is set OK!!!\n");
			// Sending a message
			try {

			    //TH added to find large time lags (> 5 sec) occuring in msg sends
			    if (super.SEND_MSG_TIMES)
				MsgUtilities.addContentFieldValuePair(kqml, "sendTime", 
								      Long.toString(System.currentTimeMillis()));

			    PrintWriter out = new PrintWriter(new OutputStreamWriter(rAgent.getOutputStream()), true);
			    out.println(kqml.getKQMLString());
			    // this doesn't remove newlines! --> out.println(message);
			    out.close();
			} catch (IOException ei) {
				DebugExceptionAgent("IOException" + ei.toString());
			}
		} else {
			// toAgent is null or toAgent does not have valid address
			String performative = kqml.getValue("performative");
			if(receiver==null||receiver.equals("")) {
				System.out.println("receiver is null or blank.  did not send ## " +
								   kqml.getKQMLString()+"##");
				if(rAgent!=null)
					try {
						rAgent.close();
					} catch(Exception ee) {//This is usually a null pointer exception
						System.out.println("receiver null, agent not, & close fails. GoFigure.");
					}
				return;
			}
			if(performative.equals("error")) {
				System.out.println("Circular Errors with "+kqml.getKQMLString());
				System.out.println("Agent.java no longer likes Circular Errors.");
			} else {
				sendError(kqml, ":receiver " + receiver + " :message Receiving Agent not found"
						  + " :undeliverableMessage " + kqml.getKQMLString());
			}
		}
		this.Debug(new DebugMsgCell("OUTMESSAGE",kqml.getKQMLString()));
		try {
			rAgent.close();
		} catch(Exception ee) {//This is usually a null pointer exception
			System.out.println("Agent.sendNow error closing connection: "+ee);
		}
	}
	
	////==== Logging Functions ====////
	/** Turns logging on replacing the given file. */
	public void startLogging(String logFile) {
		if (logFile == null || logFile.equals("")) return;
		OutMsgs.startLogging(logFile);
		logging = true;
	}
    
	/** Stops logging */
	public void stopLogging() {
		logging = false;
		OutMsgs.stopLogging();
	}

	/** Returns the current state of logging. */
	public boolean isLogging() { return logging; }
	
	////==== Tracing Functions ====////

	/** 
	 * Starts the appropriate trace variables and intializes the data structures.
	 * @param taskName The name of task to trace.
	 * @param count The number of occurences of the taskName to store at one time.
	 */
	public void startTrace(String taskName, int instances, boolean stop) {
		traceOn = true;
		tTable = new TraceTable(this, taskName, instances, stop);
	}

	/** Stops the trace and returns the data. */
	public void stopTrace() { traceOn = false; }

	/** Returns a boolean to tell wether tracing is on or off.*/
	public boolean isTracing() { return traceOn; }

	/** Returns the trace table currently stored. */
	public TraceTable getTraceTable() { return tTable; }
	
	/** Removes the trace table from the agent's memory.  Mostly to save space. */
	public void resetTraceTable() {
		traceOn = false;
		tTable = null;
	}
	
	////==== ERA Functions ====////
	
	/** Gets the name of the current ERA. */
	public String getERAName() { return eraName; }

	/** Sets the name of the ERA */
	public void setERAName(String name) {
		if (name == null || name.equals("")) {
			if (eraName != null) {
				unsetERAname();
			}
		} else if (eraName == null) {
			eraName = name;
			KQMLmsg setup = makeERAsubscribe();
			setup.addFieldValuePair("receiver", eraName);
			sendNow(setup);
		} else if (!name.equals(eraName)) {
			unsetERAname();
			eraName = name;
			KQMLmsg setup = makeERAsubscribe();
			setup.addFieldValuePair("receiver", eraName);
			sendNow(setup);
		}
	}
	/**
	  * Unsubscribes from the current ERA, and sets the name variable to null.
	  */
	public void unsetERAname() {
		KQMLmsg unsub = new KQMLmsg();
		unsub.addFieldValuePair("performative", "unsubscribe");
		unsub.addFieldValuePair("sender", getName());
		unsub.addFieldValuePair("receiver", eraName);
		unsub.addFieldValuePair("language", "DECAF");
		unsub.addFieldValuePair("ontology", "ErrorReporting");
		String time = String.valueOf(System.currentTimeMillis());
		unsub.addFieldValuePair("content", ":time " + time);
		sendNow(unsub);
		eraName = null;
	}
	
	/**
	  * Private function to create a valid ERA subscribe message.
	  */
	private KQMLmsg makeERAsubscribe() {
		KQMLmsg sub = new KQMLmsg();
		sub.addFieldValuePair("performative", "subscribe");
		sub.addFieldValuePair("sender", getName());
		sub.addFieldValuePair("language", "DECAF");
		sub.addFieldValuePair("ontology", "ErrorReporting");
		String ont = (String)userHash.get("alternateSupportedPingOntologies");
		String time = String.valueOf(System.currentTimeMillis());
		sub.addFieldValuePair("content", ":ontologies " + ont + " :time " + time);
		return sub;
	}
		
	
	////==== Utility Functions for Agent ====////

	/** 
	 * Returns the argument on the command line after x.  Provides
	 * support for simplified calls and backwards compatibility.
	 * @param x The string to look for. Ex: "-agn"
	 * @return A string containing the word after x.
	 */
	public String parseCommandLine(String x) { 
		return parseCommandLine(x, false);
	}

	/**
	 * Parses the command line for x.
	 * @param x The string to look for. Ex: "-anh"
	 * @param isSwitch If true then the return will be "true" and if
	 * not found "false"
	 * @return A String containing either the word after x, "true", or "false"
	 */
	public String parseCommandLine(String x, boolean isSwitch){
		try {
			int length = CommandLine.length;
			for (int i = 0; i < length; i++) {
				if (CommandLine[i].equals(x)) {
					if (isSwitch) {
						return "true";
					} else {
						if ( i+1 > length) {
							return "";
						} else {
							return CommandLine[i+1];
						}
					}
				}
			}
			if (isSwitch) {
				return "false";
			} else {
				return null;
			}
		} catch (java.lang.ArrayIndexOutOfBoundsException e) { 
			return null; 
		} 
	}

	/** Removes an object from the pending hashtable and returns it. */
	public Object removeFromPending(Object key) {
		Object x = Pending.remove(key);
		DebugAgent("Agent - Key removed from pending: " + x);
		return(x);
	}
	
	/** Adds tasks from an external file into DECAF's taskTemplates hashtable. */
	public void addTaskTemplate(String file) {
		agentPlan = new Plan();
		agentPlan.setPlanFile(file);
		agentPlan.readFile();
		Hashtable temp = new Hashtable();
		temp = agentPlan.gettaskTemplate();
		String key;
		TaskCell value;
		for(Enumeration e = temp.keys(); e.hasMoreElements(); ) {
			key = (String) e.nextElement();
			value = (TaskCell) temp.get(key);
			if (this.taskTemplates.containsKey(key))
			    System.out.println("Warning in Agent.java, duplicate template attempted for task: " + key);
			else
			    this.taskTemplates.put(key, value);
		}
	}
	
	/** Adds a debug switch to the current list */
	public void addDebugSwitch(String type, boolean on) {
		DebugSwitchCell temp = new DebugSwitchCell(type, type, on);
		debugSwitch.put(type, temp);
		gui.addDebugCheckbox(temp);
	}
	/** Tells wether a debug switch is on. */
	public boolean DebugOn(String which) {
		DebugSwitchCell X = (DebugSwitchCell)debugSwitch.get(which);
		if (X==null) {
			System.out.println("Error, code invoked Agent.DebugOn(" + which + 
							"), and that flag doesn't exist.  Returning \"false\"");
			return false;
		} else {
			return X.DebugOn();
		}
	}
	
	/** Adds a debug message to the queue to be printed. */
	public void Debug(DebugMsgCell Message) {
		this.debugMsgs.add(Message);
		this.DebuggerSem.V(); //activate debbuger.
	}

	/** Prints out a debug string from "AGENT" */
	public void DebugAgent(String Message) {
		this.Debug(new DebugMsgCell("AGENT", Message));   
	}
	
	/** Prints out an error message if "AGENT" or "EXCEPTION" are turned on. */
	public void DebugExceptionAgent(String Message) {
		if(this.DebugOn("AGENT"))
			this.Debug(new DebugMsgCell("AGENT", Message));
		else if(this.DebugOn("EXCEPTION"))
			this.Debug(new DebugMsgCell("EXCEPTION", Message));
	}
	
	/** Prints out the current size of the internal data structures to the terminal.  */
	public void size() {
		System.out.println(agentName+" Hashtable sizes:"
			+" AgentCharacteristics:"+AgentCharacteristics.size()
			+" Pending:"+Pending.size()
			+" taskTemplates:"+taskTemplates.size()
			+" debugSwitch:"+debugSwitch.size() );
		System.out.println(agentName+" Global LinkedListQ sizes:"
			+" debugMsgs:"+debugMsgs.size()
			+" TaskQueue:"+NuSize(TaskQueue)
			+" SchdQ:"+NuSize(SchdQ)
			+" WITQ:"+NuSize(WITQ)
			+" Objectives:"+NuSize(Objectives)
			+" ActionRes:"+NuSize(ActionRes)
			+" Agenda:"+NuSize(Agenda)
			+" incomingQueue:"+NuSize(incomingQueue)
			+" SendQ:"+NuSize(SendQ) );
	}

	private int NuSize(LinkedListQ LLQ) {
		LLQ.deep=0;
		return LLQ.realSize();
	}

	public int Qsize() {
		int answer =
			debugMsgs.size()
			+ TaskQueue.size()
			+ SchdQ.size()
			+ WITQ.size()
			+ Objectives.size()
			+ ActionRes.size()
			+ Agenda.size()
			+ incomingQueue.size()
			+ SendQ.size();
		return answer;
	}
	
	////==== Variable Accessor Functions ====////
	public int getDebugMsgsQueueSize() {return debugMsgs.size();}
	public int getTaskQueueSize() {return TaskQueue.size();}
	public int getSchdQSize() {return SchdQ.size();}
	public int getWITQSize() {return WITQ.size();}
	public int getObjectivesSize() {return Objectives.size();}
	public int getActionResSize() {return ActionRes.size();}
	public int getAgendaSize() {return Agenda.size();}
	public int getPendingSize() {return Pending.size();}
	public int getIncomingQueueSize() {return incomingQueue.size();}
	public int getTaskTemplatesSize() {return taskTemplates.size();}
	public int getAgentPlanSize() {return agentPlan.size();}
	public int getDebugSwitchSize() {return debugSwitch.size();}
	public int getSendQSize() {return SendQ.size();}

	public Hashtable getTasks(){ return(taskTemplates); }
	public MsgOutput getMsgOutput() { return OutMsgs; }
	public LinkedListQ getTaskQueue() { return TaskQueue; }
	public LinkedListQ getAgenda() { return Agenda; }
	public LinkedListQ getSchdQ() { return SchdQ; }
	public LinkedListQ getObjectives() { return Objectives; }
	public Plan getagentPlan() { return agentPlan; }
	public Hashtable getPending() { return Pending; }

	public String getAgentHost() { return agentHost; }
	public String getAgentName() { return agentName; }
	public String getName() { return agentName; }
	public int getAgentPort() { return agentPort; }
	public String getANSHost() { return ansHost; }
	public int getANSPort() { return ansPort; }

	////==== Agent Variables ====////

	/** A hashtable for the user to save persistent information */
	public static Hashtable userHash = new Hashtable ();

	/** A hashtable for Agent characteristics, ie AgentName, NickName, 
	 * AgentHost, ANSHost, ANSName, MatchMaker, Broker,
	 * Keywords, Capabilities, DatabaseType, ProxyAgent
	 */
	public static Hashtable AgentCharacteristics = new Hashtable();

	/** The name of this agent */
	private String agentName;
	/** The name of the computer the agent is running on */
	private String agentHost;
	/** The port on the computer that the agent is listening on */
	private int agentPort;
	/** A link to the current Agent Window.  Can be null. */
	static protected AgentWindow gui;

	/** The current ANS Host name */
	private String ansHost;
	/** The port that the ANS is using */
	private int ansPort;

	/** The error reporting agent */
	private String eraName;

	// Tracing variables
	private TraceTable tTable;
	private boolean traceOn = false;

	// Agent's Threads and Interfaces
	public CommunicateWithANS ansComm;
	private Receive ListenThread;
	private MsgOutput OutMsgs;
	private DispatcherThread Dispatcher;
	private PlannerThread Planner;
	private Scheduler SC;
	private Executor Execute;
	private GPGPModule GPGP;

	// Agent's Global Data Structures
	private LinkedListQ debugMsgs;		// For output to the agent gui
	private LinkedListQ TaskQueue;		// For Planner to build
	private LinkedListQ SchdQ;			// Scheduler adds, AgendaManager removes
	private LinkedListQ WITQ;			// For Planner to build
	private LinkedListQ Objectives;		// For the dispacher
	private LinkedListQ ActionRes;		// For the P.S., Action Executor
	private LinkedListQ Agenda;			// For the P.S..  Executor and Dispather threads
	private Hashtable Pending;			// used for actions that need a response.
	private LinkedListQ incomingQueue;	// For the Communication Package, and Dispatcher threads
	public Hashtable taskTemplates;	// For the template of the tasks, initally from form plan file
	private Plan agentPlan;
	
	// Need a Sempahore protecting each et of Data Structures
	Semaphore DispatcherSem = new Semaphore();		//Dispatcher notify and wait
	Semaphore PlannerSem = new Semaphore();			//Planner notify and wait
	Semaphore AgendaManagerSem = new Semaphore();	//Scheduler notify and wait
	Semaphore ExecutorSem = new Semaphore();		//Executor notify and wait
	Semaphore DebuggerSem = new Semaphore();		// Notify GUI of new message to display
	Semaphore SchedSem = new Semaphore();			//Taems Scheduler notify and wait
	Semaphore GPGPSem = new Semaphore();			//GPGP Module notify and wait

	/** A monitored Thread Group for the user to use */
	public MonitoredThreadGroup userThreadGroup = new MonitoredThreadGroup("User Threads", this);

	// Adding the debug Switch
	// possible keys for the switch:
	// 0 -- dispatcher
	// 1 -- plan-scheduler
	// 2 -- executor
	// 3 -- run action
	// 4 -- Communication Package
	Hashtable debugSwitch = new Hashtable();
	private LinkedListQ SendQ = new LinkedListQ();
	private ServerSocket listen_socket;
	private boolean logging = false;
	static boolean notInitialized = true;
	static boolean isRegistered = false;

	/** Holds the command line options used when the agent was created */
	static private String[] CommandLine;
}
