import java.util.*;
import java.io.*;

/**
 * AgentDef.java
 *
 *
 * Created: Tue Aug 26 12:47:36 2003
 * 
 * This is a super class for DECAF agents to provide functionality
 * required for bidding and generating text trees.
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public abstract class AgentDef {

    Class plAgentType;
    int level = 0; //reset during startup
    int numAltUtilBids = 2; //default
    int numHighUtilBids = 5; //default
    static final int NUMLEVELS = 4;
    boolean SEND_FAIL_REPORTS = true;
    boolean SEND_MSG_TIMES = false; //is this working?
    int MSEC_ERROR_LIMIT = 5000; //num millisecs past which there is considered to be a transmission error
    static String dataFile1 = "thesisdata";
    static  int total_time = 0; //used for determining how much time was used
    //protected int numLevels = 0; //number of levels in agent system, including this level; estimate going down, data up
    static public int kqmlMsgCount = 0;
    static public int kqmlMsgExtCount = 0;
    public int timeBase = 2;
    public int timeLevelFactor = 4;
    String upperLevelEndTime; //time by which this Agent's upper level tasks should complete
    String lowerLevelEndTime = null; //time by which this Agent's lower level tasks should complete
    private static int simulatedDelay = 1000; //milliseconds
    public static Object synch = new Object();

    private static int jobSuffix = -1; //used to give unique names to jobs
    String dsType = DSType.UNSET; //see interface DSType
    private BidSet localBids = new BidSet();//bids executable by this agent alone
    ConstraintSet constraints; //set in aReqForBids
    String parent = "parentNotSet";
    Vector children = new Vector(5);
    String executeTaskName = "Execute"; //change for numeric agents
    GenAgent myTaskObject = null; //used to identify the task portion of this agent
    boolean hasNumericChildren = false; //true for low level agents that wrap sources
    boolean isNumeric = false; //true for lowest level agents that act as sources
    public Hashtable taskObjects = new Hashtable();
    static String failReportAgentName = "FR";
    private long maxMsgTime = 0;
    private String maxMsgJob = "";

    //public Hashtable bidsMade = new Hashtable(); //make private again
    //public Vector allBidsConsidered;
    //private Hashtable bidsAccepted = new Hashtable(5); //to store each accepted bid with its envelope
    private BidUtility uFcn = null;
    static private Vector msgCopyFields = null;
    static final String copyFields = "symbol price quantity endtime job update parentEnv currEnv";//levels 

    private static boolean DEBUG1 = false;
    private static boolean DEBUG = false;
    private static String loc = " AgentDef.java ";
    private static String em = " Error in AgentDef.java ";
    private static String wm = " Warning in AgentDef.java ";
    String defaultString;

    private Hashtable localStrings = new Hashtable(5); //key=bid, value=string

    
    public AgentDef() {	
	
    } // AgentDef constructor
    
    //this constructor used only for debugging?
    public AgentDef(String n, BidUtility bu) {
	uFcn = bu;
	
    } // AgentDef constructor
    
    public void setSimulatedDelay(int sd){simulatedDelay = sd;}
    public int getSimulatedDelay(){return simulatedDelay;}
    public void setTotalTime(int t){total_time = t;}
    public void setLevel(int i){ level = i; }
    public int getLevel(){ return level; }
    //public void setNumLevels(int i){ numLevels = i; }
    public int getNumLevels(){ return NUMLEVELS + 1 - level; }
    public void setDsType(String p){dsType = p;}
    public String getDsType(){return dsType;}
    public void setPLAgentType(Class type){plAgentType = type;}
    public Class getPLAgentType(){return plAgentType;}
    public void setParent(String p){parent = p;}
    public String getParent(){return parent;}
    public void setChildren(Vector kids){children = kids;}
    public void addChild(String agentName){children.add(agentName);}
    public Iterator getChildren(){return children.iterator();}
    public Vector getChildrenVector(){return children;}
    public int getNumberOfChildren(){return children.size();}

    /**
     * numHighUtilBids: when bidding, an agent will submit the
     * <numHighUtilBids> highest utility options to the parent for
     * consideration. In addition, the agent will submit
     * <numAltUtilBids> other bids that may not have very high
     * utility, but which are very different from the highest U bid,
     * and therefore may be good options if the U fcn were to change.
     **/
    public int getNumHighUtilBids(){return numHighUtilBids;}
    public int getNumAltUtilBids(){return numAltUtilBids;}
    public void setNumHighUtilBids(int n){numHighUtilBids = n;}
    public void setNumAltUtilBids(int n){numAltUtilBids = n;}

    public void setLocalBids(BidSet lb){localBids = lb;}
    //public BidSet getLocalBids(){return localBids;}//returns ALL; see below
    public boolean isLocalBid(Bid b){return localBids.contains(b);}

    //    public void setName(String s){name = s;}
    public void setUtilityFcn(BidUtility bu){uFcn = bu;}
    public BidUtility getUtilityFcn(){return uFcn;}
    //public void addAcceptedBid(Bid b, Envelope e, String job) {

    public void setExecuteTaskName(String etn){executeTaskName = etn;}
    public String getExecuteTaskName(){return executeTaskName;}

    public void setMyTaskObject(GenAgent to){myTaskObject = to;}
    public GenAgent getMyTaskObject(){return myTaskObject;}
    public boolean hasNumericChildren(){return hasNumericChildren;}
    public void setHasNumericChildren(){hasNumericChildren = true;}
    public boolean isNumeric(){return isNumeric;}
    public void setIsNumeric(){isNumeric = true;}

    public String getUpperEndTime(){return upperLevelEndTime;}//String of long
    public String getLowerEndTime(){
	if (lowerLevelEndTime == null)
	    lowerLevelEndTime = Long.toString( Long.parseLong(upperLevelEndTime) - 4000);
	return lowerLevelEndTime;
    }//String of long
    public void setUpperEndTime(String t){upperLevelEndTime = t;}
    public void setLowerEndTime(String t){lowerLevelEndTime = t;}

    //is based on the the timeout of the lower side of the task above
    public int taskUpperTimeout(String endtime){
	long defaultTime = timeLevelFactor * getNumLevels() + timeBase;
	long remaining = getTimeRemaining(endtime)/1000;
	//greedy assumes half of timeLevelFactor is for msgs it took to get to this level
	long greedyTime = remaining - (getLevel() * timeLevelFactor/2);
	if (greedyTime < defaultTime)
	    System.out.println("WARNING in taskUpperTimeout: greedyTime: " + greedyTime + 
			       " is less than defaultTime: " + defaultTime + 
			       " with remainingTime: " + remaining);
	//if (remaining < defaultTime)
	//    return defaultTime;
	return (int)greedyTime;
    }
    //for msgs sent from upper portion of agent to lower; upper portion of agent is PL_BidAgent
    public int upperMsgTimeout(String endtime){return taskUpperTimeout(endtime) - 2;}
    public int taskLowerTimeout(String endtime){return taskUpperTimeout(endtime) - 4;}
    //for msgs sent from lower portion of agent to next agent level down
    public int lowerMsgTimeout(String endtime){return taskUpperTimeout(endtime) - 5;}

    public BidSet getLocalBids(){return localBids;}//returns ALL bids.
    //returns only bids with same symbol; sets realtime bids for internalNumeric agents
    public BidSet getLocalBids(KQMLmsg msg, Agent Local){

	BidSet bs;
	String symbol = msg.getContentFieldValue("symbol");
	GenAgent me = getMyTaskObject();
	if (me != null && me instanceof InternalNumericAgent){
	    bs = ((InternalNumericAgent)me).makeLocalBids(msg.toString(), Local);
	    if (DEBUG1) System.out.println("internalNumeric setting locals: " + bs);
	    return bs;
	}
	else {

	    BidSet allLocalBids = getLocalBids();
	    bs = new BidSet(allLocalBids.getOwner());
	    Iterator i = allLocalBids.iterator();
	    Bid b;
	    while(i.hasNext()){
		b = (Bid)i.next();
		if (b.getSymbol().equals(symbol)){
		    if (DEBUG1) System.out.println(loc + 
						  "getLocalBids(symbol) found bid " + b);
		    bs.add(b);
		}
	    }
	    return bs;
	    
	}
    }//getLocalBids(String symbol)

    /*
     * Parents of numeric agents (hasNumericChildren() == true) have a
     * makeTreeFromData() method that is called from this method. It
     * allows GenAgent parent (expert) to convert numbers (DataSets)
     * into text trees.
     */
    public Vector processChildData(String numericChildData){

	Vector childData = DataSet.readMultipleDataSets(numericChildData);
	    if (DEBUG1) System.out.println(loc + "childData: " + childData);
	Vector trees = new Vector(5);

	GenAgent me = getMyTaskObject();
	RSTTree r = null;
	Iterator i = childData.iterator();
	while(i.hasNext()){
	    DataSet d = (DataSet)i.next();
	    r = me.makeTreeFromData(d);
	    if (r != null){
		if (DEBUG1) System.out.println(loc + "processChildData next tree: \n");
		r.print();	    
		trees.add(r);
	    }
	    else {
		if (DEBUG1) System.out.println(loc + "processChildData null tree from " + d);	    
	    }
	}
	//separate out different data sets
	//call each source's method on its set
	//return vector of RSTTrees
	return trees;
    }//processChildData(String numericChildData)

    /**
     * Use DataSet to generate a tree, unless numeric agent, in which
     * case return the DataSet as string.
     **/
    //use agent's makeTreeFromData in task file (PL_AgentTypeName) to make local trees
    public String generate(Envelope env){
	Bid b = env.getBid();
	if (DEBUG1){
	    System.out.println(" inside generate(), bid is " + b + " localbids are: " + localBids);
	}
	Bid local = getLocalBidUsingEquals(b);
	GenAgent me = getMyTaskObject();
	if (local != null && me != null){
	    if ((me instanceof NumericAgent) || this.hasNumericChildren()){
		if (DEBUG1) System.out.println("generate returning1: " + localStrings.get(local));
		return (String)localStrings.get(local); //is not a tree, but a DataSet (to be updated)
	    }
	    else {
		if (DEBUG1) System.out.println("Generate found exact match for " + 
				   b + "returning: " + (String)localStrings.get(local));
		//String tree = (String)localStrings.get(local);
		//RSTTree rt = RSTTree.readRSTTree(tree);
		DataSet ds = DataSet.readDataSetFromString((String)localStrings.get(local));
		RSTTree rt = me.makeTreeFromData(ds);
		//rt.setBid(b);// wrong.
		if (rt != null){
		    if (getUtilityFcn() != null)
			rt.updateUtility(getUtilityFcn());
		    return rt.getString();
		}
		else return "empty";
	    }
	}
	return defaultString;
    }

    public void clearLocalBids(){
	localBids.clear();
	localStrings.clear();
    }

    /*
     * Local bids are first set by using teh Update class to parse a
     * data file with starting bid info. The file names are currently
     * specified by construction in PL_Start and sent to the relevant
     * agents. Filenames are variants in "IRdata", PEdata" etc.
     */
    public void addLocalBid(Bid b, String s){
	localBids.add(b);
	localStrings.put(b,s);
    }

    /* Incoming Bid b is not identical to local bid, but should be
     * equal(). Returns the local bid.
     */
    public Bid getLocalBidUsingEquals(Bid b){
	Iterator i = getLocalBids().iterator();
	Bid local;
	while(i.hasNext()){
	    local = (Bid)i.next();
	    if (b.equals(local)){
		if (DEBUG1) System.out.println(loc + 
			    "getLocalBidUsingEquals() found bid " + 
					      b + " and local: " + local);
		return local;
	    }
	}
	
	System.out.println(wm + "getLocalBidUsingEquals sources not found for " + b +
			   " in " + getLocalBids());
	return null;
    }
    

    //public Set getBidsMade(){return bidsMade.keySet();}

//public BidSet getBidSetMade(String agentName){
//	return new BidSet(agentName,(Collection)bidsMade.keySet());}

    //public boolean madeBidP(Bid b){
    //	return bidsMade.containsKey(b);
    //}

    //public void setBidsMade(BidSet bs){
    //	Iterator i = bs.iterator();
    //	Bid b;
    //	while (i.hasNext()){
    //	    b = (Bid)i.next();
    //	    bidsMade.put(b, b.getBidSetSource());
    //	    if (b.getBidSetSource() == null)
    //		System.out.println(wm + "bid " + b + " has null source");
    //	}
    //}

    //public BidSet getBidComponents(Bid b){
    //	Iterator i = getBidsMade().iterator();
    //	Bid stored;
    //	while(i.hasNext()){
    //	    stored = (Bid)i.next();
    //	    if (b.equals(stored)){
    //		if (DEBUG1) System.out.println(loc + 
    //			    "getBidComponents found bid " + b + "set " + bidsMade.get(stored));
    //		return (BidSet)bidsMade.get(stored);
    //	    }
    //	}
    //
    //	System.out.println(wm + "getBidComponents sources not found for " + b);
    //	return null;
    //}

    //This doesn't work because reconstituted bids don't have the same hashcode as original
//    public BidSet getBidComponents(Bid b){
//	if (bidsMade.containsKey(b))
//	    return (BidSet)bidsMade.get(b);
//	else
//	    return null;
//    }

    //extracts content from name/content pair returned by Util.getValue()
    public String pcContent(String s){
	String value;
	StringTokenizer st = new StringTokenizer(s);
	if (st.hasMoreTokens()){
	    st.nextToken();
	    if (st.hasMoreTokens())
		value = st.nextToken();
	    else value = "null";
	}
	else value = "null";
	return value;
    }

    private int percentTimeElapsed(){return 24;}//FIX

    public Iterator getMsgCopyFields(){
	if ( msgCopyFields == null){
	    msgCopyFields = new Vector(10);
	    StringTokenizer st = new StringTokenizer(copyFields);
	    while (st.hasMoreTokens())
		msgCopyFields.addElement(st.nextToken());
	}

	return msgCopyFields.iterator();
    }

    public long getTimeRemaining(String endtime){
	long end;
	try {
	    end = Long.parseLong(endtime);
	} catch (NumberFormatException e) {
	    e.printStackTrace();
	    return -999;
	}

	return end - System.currentTimeMillis();
    }

    public long getElapsedTime(String starttime){
	long start = Long.parseLong(starttime);
	return System.currentTimeMillis() - start;
    }

    public long getElapsedTime(long start){
	return System.currentTimeMillis() - start;
    }

    public String getEndTime(String seconds){
	int sec = Integer.parseInt(seconds);
	long endtime = 1000 * sec + System.currentTimeMillis();
	return Long.toString(endtime);
    }
    public String getEndTime(int seconds){
	long endtime = 1000 * seconds + System.currentTimeMillis();
	return Long.toString(endtime);
    }

    static String getUniqueJobName(){
	jobSuffix += 1;
	return "job" + jobSuffix;
    }

    /**
     * For sending failure reports to an agent that collects and prints them.
     **/
  public void sendFailReport(String job, String symbol, String text, String timeRemaining, Agent Local){
      KQMLmsg K = new KQMLmsg();

      text = text.replace(':', '~'); //colons break KQML messages
      
      K.addFieldValuePair("sender", Local.getName());
      K.addFieldValuePair("performative", "achieve");
      K.addFieldValuePair("receiver", failReportAgentName);
      K.addFieldValuePair("language", "DECAF");
      K.addFieldValuePair("ontology", "PL");	
      K.addFieldValuePair("content", ":task FailReport :inFR noValue :job " + job +
			  " :symbol " + symbol + " :report " + text + " in " + Local.getName() +
			  " :timeRemaining " + timeRemaining);

      if (SEND_FAIL_REPORTS) Local.sendNow(K);
  }

    abstract public String getName();

    public void checkMsgTime(KQMLmsg k, String reportFileName, String job){
	if (DEBUG) System.out.println("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
	if (DEBUG) System.out.println("++++++ENTERING CHECK MESSAGE++++++++++++++++++++++++++++++++++++++++++");
	if (DEBUG) System.out.println("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");

	if ( SEND_MSG_TIMES )
	    try{
		String ts = k.getContentFieldValue("sendTime");
		if (! ts.equals("")){
		    long send = getElapsedTime(ts);
		    if (send > maxMsgTime){
			maxMsgTime = send;
			maxMsgJob = job; 
			System.out.println("new max msg time for job " + job + " is " + Long.toString(send));
		    }

		    if (send > MSEC_ERROR_LIMIT){
			FileWriter fw = new FileWriter(reportFileName, true);//same as name in CoordProcess
			fw.write("\nreceiver: " + getName() + " sender " + k.getValue("sender")+ " msg time error in job " + job + " is " + Long.toString(send));
			fw.flush();
			fw.close();
		    }
		}
	    } catch (Exception e){
		System.out.println("Error opening file in AgentDef.checkMsgTime: " + e);
	    }
    }

    static public void main(String[] args){
	Agent Local = new Agent();

	KQMLmsg K = new KQMLmsg();

	K.addFieldValuePair("sender", Local.getName());
	K.addFieldValuePair("performative", "achieve");
	K.addFieldValuePair("receiver", Local.getName());
	K.addFieldValuePair("reply-with", "a");
	K.addFieldValuePair("language", "DECAF");
	K.addFieldValuePair("ontology", "PL");	
	K.addFieldValuePair("content", ":task Coord :msg start :job " + 
			   JobNumber.getJobNumber() + " :levels 4 :endtime " + 
			   Local.getEndTime(90) + 
			   " :symbol IBM :price 126 :quantity 100 :constraints " + 
			   "{cost_max=30.0, length_max=150, length_min=0, time_max=" + 
			   Local.getEndTime(90) + "} " + 
			   ":utilityFcn {cost_coeff=1, cost_target=15, cost_fcn=startPlateauNormal, length_coeff=2, length_target=50, length_fcn=startPlateauNormal, DSr_coeff=3, DSr_target=5, DSv_coeff=1, DSv_target=5, DSg_target=5, DSg_coeff=1, RV_coeff=1, DSr_fcn=linear, DSv_fcn=normPlateauEnd, DSg_fcn=normPlateauEnd}");

	//PL_DebtRisk dr = new PL_DebtRisk();
	PL_PortfolioAlloc pa = new PL_PortfolioAlloc();
	Local.setMyTaskObject(pa);

	BidSet bs = Local.getLocalBids(K, Local);
	
    }

} // AgentDef
