import java.util.*;
import java.io.*;
import java.text.NumberFormat;
/**
 * Bid.java
 *
 * Bid represents expected values of attributes (before execution) or
 * real values of attributes (after execution). BidUtility is the set
 * of coefficients applied to bid to determine the utility of the bid.
 *
 * Length, cost, other simple numbers stored as Strings. Bids, DS
 * (Strength) stored as class.
 *
 * In calcUtility all utility term weights (coeffs) are scaled.
 *
 * Created: Fri Jul 25 13:24:00 2003
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class Bid extends Hashtable {

    BidSet source = null; //a BidSet that this bid represents, typically bids of children are in set.
    String bidder = "unset";
    String lastBU = "unset";
    Double utility = null;
    boolean Uchanged = true; //keeps track of whether utility components have altered
    boolean DSchanged = true; //keeps track of whether DS components have altered
    static final String em = "ERROR in Bid.java: ";
    static final String wm = "WARNING in Bid.java: ";
    static final String loc = "in Bid.java: ";
    static boolean DEBUG = false;
    static boolean DEBUGcalc = false;
    static final double MIN_TIME_LATENCY = 1000; //milliseconds
    /** used to determine size of standard dev for U calc as a fcn of target number **/
    static final double STD_DEV_QUOTIENT = 5;

    static final double DEFAULT_COST_TARGET = 10;
    static final double DEFAULT_LENGTH_TARGET = 100;
    static final double DEFAULT_TIME_TARGET = 60;
    static final double DEFAULT_INFO_TARGET = 10;

    static NumberFormat nf = NumberFormat.getInstance();

    public Bid(){
	super(10);
	initStrengths();
	nf.setMaximumFractionDigits(1);
    } // Bid constructor

    public Bid(String bidderName) {
	super(10);
	nf.setMaximumFractionDigits(1);
	initStrengths();
	bidder = bidderName;
    } // Bid constructor

    public Object put(Object k, Object v){
	Uchanged = true;
	if ((k instanceof String) && ((String)k).startsWith("DS"))
	    DSchanged = true;
	return super.put(k,v);
    }

    public Object clone(){
	Bid result = (Bid)super.clone();
	result.source = source;
	result.bidder = bidder;
	result.utility = utility;
	result.Uchanged = Uchanged;
	if ( ! Uchanged) result.utility = this.utility;
	result.DSchanged = DSchanged;
	return result;
    }

    //is this necessary? 
    public Object deepClone(){
	Bid result = new Bid();
	result.source = source;
	result.bidder = bidder;
	result.utility = utility;
	result.Uchanged = Uchanged;
	if ( ! Uchanged) result.utility = this.utility;
	result.DSchanged = DSchanged;
	Enumeration k = this.keys();
	Object o;
	while (k.hasMoreElements()){
	    o = k.nextElement();
	    result.put( o, this.get(o) );
	}
	return result;
    }

    public double utility(){
	if (utility != null) return utility.doubleValue();
	else return -1;
    }

    public String getSymbol(){return (String)this.get("symbol");}
    public void setSymbol(String s){this.put("symbol", s);}

    //A single bid is often compiled from a set of bids; this set is the BidSet source.
    public void setBidSetSource(BidSet b){source = b;}
    public BidSet getBidSetSource(){return source;}

    public String getBidder(){return bidder;}
    public void setBidder(String b){bidder = b;}

    private void setLastBU(String b){lastBU = b;}

    public boolean equals(Object b){
	if (b instanceof Bid)
	    return this.equals((Bid)b);
	else return false;
    }

    public boolean equals(Bid b){
	if (DEBUG) System.out.println("comparing " + this + " and " + b);
	
	if ( ! this.getBidder().equals(b.getBidder()))
	    return false;

	Iterator i = this.getAttributes();
	Object o;
	while (i.hasNext()){
	    o = i.next();
	    if (DEBUG) 
		System.out.println("o is " + o + " " + get(o)+ " b.o is " + b.get(o));
	    if (! (this.get(o).equals(b.get(o))) )
		return false;
	}
    
	if (DEBUG) System.out.println("returning true");
	return true;
    }

    //does not compare cost_exp, time_exp
    public boolean limitedEquals(Bid b){
	if (DEBUG) System.out.println("comparing " + this + " and " + b);

	if ( ! this.getBidder().equals(b.getBidder()))
	    return false;

	Iterator i = this.getAttributes();
	Object o;
	while (i.hasNext()){
	    o = i.next();
	    if (DEBUG) System.out.println("limitedEquals: o is " + o + " " + get(o)+ " b.o is " + b.get(o));
	    if (o.equals("cost_exp") || o.equals("time_exp"))
		{} //why doesn't continue work here?
	    else if (! (this.get(o).equals(b.get(o))) ){
		if (DEBUG) System.out.println("returning false");
		return false;
	    }
	}
    
	if (DEBUG) System.out.println("returning true");
	return true;
    }

    public double cosineDistance(Bid other){

	Iterator attIter = this.getAttributes();
	double[] myValues = new double[size()];
	double[] otherValues = new double[size()];
	int i = 0;
	String nextAtt;
	double myNextValue;
	while (attIter.hasNext()){
	    nextAtt = (String)attIter.next();

	    if (other.containsKey(nextAtt)){

		if (nextAtt.startsWith("DS")){
		    Strength s = (Strength)this.getValue(nextAtt);
		    myValues[i] = s.getDSSumsForUtility();
		    s = (Strength)other.getValue(nextAtt);
		    otherValues[i] = s.getDSSumsForUtility();
		    if (DEBUG) System.out.println("inside cosineDistance DS");
		}
		else if (nextAtt.startsWith("RV")){
		    RuleValue rv = (RuleValue)this.getValue(nextAtt);
		    myValues[i] = rv.getIntValue();
		    rv = (RuleValue)other.getValue(nextAtt);
		    otherValues[i] = rv.getIntValue();
		}
		else try{
		    if (DEBUG) System.out.println("trying double on " + nextAtt);
		    myValues[i] = Double.parseDouble((String)this.getValue(nextAtt));
		    otherValues[i] = Double.parseDouble((String)other.getValue(nextAtt));
		} catch (NumberFormatException nfe) {
		    if (DEBUG) System.out.println(" nextAtt is: " + nextAtt + " value is: " + 
				       this.getValue(nextAtt));
		    myValues[i] = 0;
		    otherValues[i] = 0;
		}
		i++;
	    }
	}
	if (DEBUG){
	    for (i = 0; i < myValues.length; i++)
		System.out.print(myValues[i] + " ");
	    System.out.println("");
	    
	    for (i = 0; i < otherValues.length; i++)
		System.out.print(otherValues[i] + " ");
	    System.out.println("");
	}
	double similarity = CosineDistance.twoArraySimilarity(myValues, otherValues);

	return 1 - similarity;
    }//cosineDistance(Bid other)

    /**
     * Returns an Iterator of the keySet of this Bid.
     */
    public Iterator getAttributes(){
	return keySet().iterator();
    }
    
    /**
     * Increase the RuleValue term of this Bid for rule <ruleNumber>
     * with value <ruleValue>.
     */
    public void increaseRuleValue(int ruleNumber, int ruleValue){
	RuleValue r = (RuleValue)getValue("RV");
	if (r == null){
	    r = new RuleValue();
	    this.put("RV", r);
	}
	r.increase(ruleNumber, ruleValue);
    }

    private void initStrengths(){
	this.put("DSr_exp", new Strength());
	this.put("DSg_exp", new Strength());
	this.put("DSv_exp", new Strength());
    }

    public Strength getDStot(){
	if (DSchanged)
	    calcDSTotal();
	else if(! this.containsKey("DStot_exp"))
	    calcDSTotal();
	return (Strength)get("DStot_exp");
    }

    public void calcDSTotal(){
	Strength total = new Strength();
	String name;	
	Iterator attributes = this.keySet().iterator();
	while (attributes.hasNext()){
	    name = (String)attributes.next();
	    if (name.startsWith("DS") &&
		! name.equals("DStot_exp")){
		total.add( (Strength)this.getValue(name) );
	    }
	}
	this.put("DStot_exp", total);
	DSchanged = false;
    }

    public Object getValue(String attributeName){
	return this.get(attributeName);
    }

    //This isn't working properly; bypassed. FIX so that it doesn't have to be calculated every time.
    public Double getUtility(BidUtility bu){
	return new Double(calcUtility(bu));
    }
    /*	if (DEBUG) 
	    System.out.println(loc + "getUtility(): " + bu + " " + 
				      Uchanged + "oldU: " + utility);
	if (Uchanged){
	    if (DEBUG) System.out.println("new U: " + new Double(calcUtility(bu)));
	    return new Double(calcUtility(bu));
	}
	else if ( ! lastBU.equals(bu.get("name"))){//gets reset in calcUtility()
	    if (DEBUG) System.out.println("new U: " + new Double(calcUtility(bu)));
	    return new Double(calcUtility(bu));
	}
	else
	    return utility;
	    }*/

    /*
     * time_max in CS is expected to be a large number of milliseconds
     */
    public boolean isWithinConstraints(ConstraintSet cs){
	String loc = " isWithinConstraints ";
	String constraintName, expName;
	Double constraintValue, expValue = null;

	if ( ! cs.containsKey("length_min") )
	    cs.put("length_min", "0");
	if ( ! cs.containsKey("time_min") )
	    cs.put("time_min", "0");

	Iterator attributes = cs.keySet().iterator();
	while (attributes.hasNext()){
	    constraintName = (String)attributes.next();
	    expName = constraintName.substring(0, constraintName.indexOf('_'));
	    expName = expName.concat("_exp");

	    if (expName.startsWith("time") && this.containsKey(expName)){
		double t = Double.parseDouble((String)this.getValue(expName));//num seconds
		long timeExp = ((long)(t * 1000)) + System.currentTimeMillis();
		if (constraintName.endsWith("max")){
		    long timeMax = Long.parseLong((String)cs.getValue(constraintName));
		    if ((timeExp + MIN_TIME_LATENCY) >= timeMax){
			System.out.println(this + " failed constraint: " + constraintName + 
					   ": " + cs.getValue(constraintName));
			System.out.println(" failed constraint: " + expName + 
					   ": " + t);
			System.out.println(" failed constraint: calced timeExp: " + timeExp);
			System.out.println(" failed constraint: calced timeMax: " + timeMax);
			System.out.println("sys time - max: " + (System.currentTimeMillis() - timeMax));
			
			return false;
		    }
		    else continue;
		}	
		else if (constraintName.endsWith("min")){
		    long timeMin = Long.parseLong((String)cs.getValue(constraintName));
		    if (timeExp < timeMin){
			System.out.println(this + " failed constraint: " + constraintName + 
					   " expVal: " + t);
			return false;
		    }
		    else continue;
		}
	    }
	    else if (this.containsKey(expName)){
		try {
		    
		    if (expName.startsWith("DS")){
			Strength s = (Strength)this.getValue(expName);
			expValue = new Double(s.getDSSumsForUtility());
		    }
		    //need to have access to bidUtility to do this
		    //else if (expName.startsWith("utility")){
		    //    expValue = this.getUtility();
		    //}
		    else if (expName.startsWith("RV")){
			RuleValue rv = (RuleValue)this.getValue(expName);
			expValue = new Double(rv.getIntValue());
		    }
		    else
			expValue = Double.valueOf((String)this.getValue(expName));
		
		    constraintValue = Double.valueOf((String)cs.getValue(constraintName));
		    
		    if (constraintName.endsWith("min") &&
			(constraintValue.compareTo(expValue) <= 0))
			continue;//return true;
		    else if (constraintName.endsWith("max") &&
			     (constraintValue.compareTo(expValue) >= 0))
			continue;//return true;
		    else {
			System.out.println(this + " failed constraint: " + constraintName + 
					   " expVal: " + expValue);
			return false;
		    }
		} catch(NullPointerException e){
		    System.out.println(e + loc + "looking up " + constraintName + ", " + 
				       expName + " in " + cs + ", " +this);
		}
	    }
	}
	return true; //if there are no constraints, return true.
    }//isWithinConstraints(ConstraintSet cs)

/** 
 * Find DS term with highest currDS, return utility of that term using
 * currDS instead of sum.
*/
    public double getHiCurrDSUtil(BidUtility bu){
	double best = 0, coeff, target, currValue, nextTermU = 0;
	int char_loc;
	String attName = null, root = null, targetName = null, coeffName = null, 
	    fcn = null,  //is the name of fcn in DistFcn
	    fcnName = null; //is the name by which fcn is found in U
	Iterator attributes = this.getAttributes();

	while (attributes.hasNext()){

	    coeff = 0; //default
	    fcn = null;
	    target = 10; //default
	    currValue = 0;
	    nextTermU = 0;
	    attName = (String)attributes.next();    

	    if (attName.startsWith("DS")){

		char_loc = attName.indexOf('_');
		if (char_loc > -1){ // is -1 if not found
		    root = attName.substring(0, char_loc);
		}
		else { //no underscore
		    root = attName;
		}
		
		coeffName = root.concat("_coeff");
		targetName = root.concat("_target");
		fcnName = root.concat("_fcn");
		
		if (bu.containsKey(coeffName)){
		    coeff = Double.parseDouble((String)bu.get(coeffName));//no need to scale
		}
		if (bu.containsKey(targetName))
		    target = Double.parseDouble((String)bu.get(targetName));
		if (bu.containsKey(fcnName))
		    fcn = (String)bu.get(fcnName);

		Strength s = (Strength)this.getValue(attName);
		currValue = s.absCurr();
		if (DEBUG) System.out.println("peakDS ");
		
		if (fcn == null)
		    fcn = "linear";
		if (DEBUG) System.out.println("peak DS fcn is: " + fcn);
		nextTermU = coeff * DistFcn.f( fcn, currValue, target, target/STD_DEV_QUOTIENT);
		if (DEBUG) System.out.println("leaving peak DS nextTermU is: " + nextTermU);
	    
		if (nextTermU > best){
		    best = nextTermU;
		}
	    }
	}
	return best;
    }

    /**
     * Calculate and update the utility of this bid, given a BidUtility fcn.
     *
     * Attributes are taken from this bid, then corresponding info is
     * located in the BidUtility.
     *
     * ANY CHANGES MADE HERE MUST BE MIRRORED in calcAndPrintUtility()
     */
    public double calcUtility(BidUtility bu){
	if (DEBUGcalc) System.out.println("in calcUtility, bid: " + this + " ufcn: " + bu);
		
	int char_loc;
	double coeff, coeffSum, target, expValue, nextTermU, newUtility = 0;
	String attName = null, root = null, targetName = null, coeffName = null, 
	    fcn = null,  //is the name of fcn in DistFcn
	    fcnName = null; //is the name by which fcn is found in U
	coeffSum = bu.sumCoeffs();

	Iterator attributes = this.getAttributes();
	//if utility function does not have an entry for this item, do not change utility
	while (attributes.hasNext()){
	    nextTermU = 0;
	    coeff = 0; //default
	    fcn = null;
	    target = 10; //default for information attributes
	    
	    expValue = 0;
	    attName = (String)attributes.next();
	    if (DEBUGcalc) System.out.println("next attr: " + attName);

	    /********************* Set names to find in Ufcn ************************/	    
	    char_loc = attName.indexOf('_');
	    if (char_loc > -1){ // is -1 if not found
		root = attName.substring(0, char_loc);
	    }
	    else { //no underscore
		root = attName;
	    }

	    coeffName = root.concat("_coeff");
	    targetName = root.concat("_target");
	    fcnName = root.concat("_fcn");

	    /***************Get coeff and target numbers from Ufcn**********************/

	    if (bu.containsKey(coeffName)){
		coeff = Double.parseDouble((String)bu.get(coeffName));
		coeff = 100 * (coeff / coeffSum); //scale coefficients so U is 0-100
	    }
	    if (bu.containsKey(targetName))
		target = Double.parseDouble((String)bu.get(targetName));
	    else if (targetName.startsWith("cost"))
		target = DEFAULT_COST_TARGET;
	    else if (targetName.startsWith("length"))
		target = DEFAULT_LENGTH_TARGET;
	    else if (targetName.startsWith("time"))
		target = DEFAULT_TIME_TARGET;
	    else //DS or rule value
		target = DEFAULT_INFO_TARGET;

	    if (bu.containsKey(fcnName))
		fcn = (String)bu.get(fcnName);

	    if (DEBUGcalc) System.out.println("fcn is: " + fcn);

	    /*************special term behaviors**************************/

	    if (attName.startsWith("DS")){
		Strength s = (Strength)this.getValue(attName);
		expValue = s.getDSSumsForUtility();
		if (DEBUGcalc) System.out.println("inside DS");
		
		if (fcn == null)
		    fcn = "linear";
		if (DEBUGcalc) System.out.println("DS fcn is: " + fcn);
		nextTermU = coeff * DistFcn.f( fcn, expValue, target, target/STD_DEV_QUOTIENT);
		if (DEBUGcalc) System.out.println("leaving DS nextTermU is: " + nextTermU);
	    }
	    else if (attName.startsWith("RV")){
		RuleValue rv = (RuleValue)this.getValue(attName);
		expValue = rv.getIntValue();
		if (fcn == null)
		    fcn = "linear";
		nextTermU = coeff * DistFcn.f( fcn, expValue, target, target/STD_DEV_QUOTIENT);
	    }
	    else if (attName.startsWith("cost")){
		expValue = Double.parseDouble((String)this.getValue(attName));
		if (fcn == null)
		    fcn = "startPlateauNormal"; //works with neg coeff if cost is bad for U but don't use neg coeff
		nextTermU = coeff * DistFcn.f( fcn, expValue, target, target/STD_DEV_QUOTIENT);
	    }
	    else if (attName.startsWith("time")){
		expValue = Double.parseDouble((String)this.getValue(attName));
		if (fcn == null)
		    fcn = "startPlateauNormal"; 
		nextTermU = coeff * DistFcn.f( fcn, expValue, target, target/STD_DEV_QUOTIENT);
	    }
	    else if (attName.startsWith("topic")
		     || attName.startsWith("symbol")
		     // || attName.startsWith("time")
		     ){
		//continue;
	    }
	    /********************Default Behavior***************************/
	    else{
		expValue = Double.parseDouble((String)this.getValue(attName));

		if (fcn == null)
		    fcn = "normal";
		//else System.out.println(em + "no target for " + attName + " in " + bu);
		nextTermU = coeff * DistFcn.f(fcn, expValue, target, target/STD_DEV_QUOTIENT);
	    }
	    
	    if (DEBUGcalc) System.out.println("in calcUtility, attName: " + attName + " coeff: " + 
					  coeff + " target: "+ target + " termU: " + nextTermU);
		
	    newUtility += nextTermU;
	}
	Uchanged = false;
	utility = new Double(newUtility);
	lastBU = (String)bu.get("name");

	if (DEBUGcalc) System.out.println("leaving calcUtility: " + utility);
		
	return newUtility;
    }//calcUtility(BidUtility bu)

    public String calcAndPrintUtility(BidUtility bu){

	StringBuffer sb = new StringBuffer(200);
		
	int char_loc;
	double coeff, coeffSum, target, expValue, nextTermU, newUtility = 0;
	String attName = null, root = null, targetName = null, coeffName = null, 
	    fcn = null,  //is the name of fcn in DistFcn
	    fcnName = null; //is the name by which fcn is found in U
	coeffSum = bu.sumCoeffs();

	Iterator attributes = this.getAttributes();
	//if utility function does not have an entry for this item, do not change utility
	while (attributes.hasNext()){
	    nextTermU = 0;
	    coeff = 0; //default
	    fcn = null;
	    target = 10; //default
	    expValue = 0;
	    attName = (String)attributes.next();
	    if (DEBUG) System.out.println("next attr: " + attName);

	    /********************* Set names to find in Ufcn ************************/	    
	    char_loc = attName.indexOf('_');
	    if (char_loc > -1){ // is -1 if not found
		root = attName.substring(0, char_loc);
	    }
	    else { //no underscore
		root = attName;
	    }

	    coeffName = root.concat("_coeff");
	    targetName = root.concat("_target");
	    fcnName = root.concat("_fcn");

	    /***************Get coeff and target numbers from Ufcn**********************/

	    if (bu.containsKey(coeffName)){
		coeff = Double.parseDouble((String)bu.get(coeffName));
		coeff = 100 * (coeff / coeffSum); //scale coefficients so U is 0-100
	    }
	    if (bu.containsKey(targetName))
		target = Double.parseDouble((String)bu.get(targetName));
	    else if (targetName.startsWith("cost"))
		target = DEFAULT_COST_TARGET;
	    else if (targetName.startsWith("length"))
		target = DEFAULT_LENGTH_TARGET;
	    else if (targetName.startsWith("time"))
		target = DEFAULT_TIME_TARGET;
	    else //DS or rule value
		target = DEFAULT_INFO_TARGET;

	    if (bu.containsKey(fcnName))
		fcn = (String)bu.get(fcnName);

	    if (DEBUG) System.out.println("fcn is: " + fcn);
	    /*************special term behaviors**************************/
	    if (attName.startsWith("DS")){
		Strength s = (Strength)this.getValue(attName);
		expValue = s.getDSSumsForUtility();
		if (DEBUG) System.out.println("inside DS");
		
		if (fcn == null)
		    fcn = "linear";
		if (DEBUG) System.out.println("DS fcn is: " + fcn);
		nextTermU = coeff * DistFcn.f( fcn, expValue, target, target/STD_DEV_QUOTIENT);
		if (DEBUG) System.out.println("leaving DS nextTermU is: " + nextTermU);
	    }
	    else if (attName.startsWith("RV")){
		RuleValue rv = (RuleValue)this.getValue(attName);
		expValue = rv.getIntValue();
		if (fcn == null)
		    fcn = "linear";
		nextTermU = coeff * DistFcn.f( fcn, expValue, target, target/STD_DEV_QUOTIENT);
	    }
	    else if (attName.startsWith("cost")){
		expValue = Double.parseDouble((String)this.getValue(attName));
		if (fcn == null)
		    fcn = "startPlateauNormal"; //works with neg coeff if cost is bad for U, but don't use neg coeff
		nextTermU = coeff * DistFcn.f( fcn, expValue, target, target/STD_DEV_QUOTIENT);
	    }
	    else if (attName.startsWith("time")){
		expValue = Double.parseDouble((String)this.getValue(attName));
		if (fcn == null)
		    fcn = "startPlateauNormal"; 
		nextTermU = coeff * DistFcn.f( fcn, expValue, target, target/STD_DEV_QUOTIENT);
	    }
	    else if (attName.startsWith("topic")
		     || attName.startsWith("symbol")
		     //|| attName.startsWith("time")
		     ){
		//continue;
	    }
	    /********************Default Behavior***************************/
	    else{
		expValue = Double.parseDouble((String)this.getValue(attName));

		if (fcn == null)
		    fcn = "normal";
		//else System.out.println(em + "no target for " + attName + " in " + bu);
		nextTermU = coeff * DistFcn.f(fcn, expValue, target, target/STD_DEV_QUOTIENT);
	    }
	    
	    if (coeff != 0)
		    sb.append(" " + attName + "=" + 
			      coeff + "*"+ fcn + "(" +expValue + ", " + target + ")=" + nextTermU);
	    
	    newUtility += nextTermU;
	}
	Uchanged = false;
	utility = new Double(newUtility);
	lastBU = (String)bu.get("name");

	sb.insert(0, utility.toString() + " = ");
	System.out.println(sb.toString());
	System.out.println(Double.toString(newUtility));

	return sb.toString();
    }//calcAndPrintUtility(BidUtility bu)

    //Add object specific members to super.toString()
    public String toString(){
	String s = super.toString();
	int size = s.length();
	s = s.substring(0, size - 1);
	return s + ", U=" + utility + ", bidder=" + bidder + ", lastBU=" + lastBU + "}";
	//return s + ", U=" + utility + "}";
    }
   
    /** trim whitespace before calling this method **/
    static public Bid readBidFromString(String s){
    	StringTokenizer st = new StringTokenizer(s, "[], {}=", true); // "{}, =");
	String next = st.nextToken();
	if (next.equals("{"))
	    return readBid(st);
	else
	    System.out.println(em + "starting brace not found; found: " + next);
	return null;
    }

    //Reads bid from st, then if Utility is null, calculates it.
    //Assumes first { already eaten. st must have been created with
    //"true" parameter to count delimters as tokens.
    static public Bid readBid(StringTokenizer st){
	String key, eq, value = null;
	boolean errorFlag = false;
	Bid h = new Bid();
	//if (DEBUG) System.out.println(loc + "inside readBid, st is: " );
	
	while (st.hasMoreTokens()){
	    key = st.nextToken("[], {}=");
	    if (key.equals("}")) 
		return h;
	    else if ( key.equals(",") || key.equals(" ") ) 
		continue;
	    else if ( key.startsWith("DS") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    Strength ds = Strength.readStrength(st);
		    h.put(key, ds);
		}
		else errorFlag = true;
	    }
	    else if ( key.equals("RV") ) {
		//if (DEBUG) System.out.println(loc + "inside readBid RV, st is: " + st.toString());
		eq = st.nextToken();
		if (eq.equals("=")){
		    RuleValue rv = RuleValue.readRuleValue(st);
		    h.put(key, rv);
		}
		else errorFlag = true;
	    }
	    else if ( key.equals("bidder") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    value = st.nextToken();
		    //if (DEBUG) System.out.println("inside bidder, kev: " + key + eq + value);
		    
		    h.setBidder(value);
		}
		else errorFlag = true;
	    }
	    else if ( key.equals("lastBU") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    value = st.nextToken();
		    //if (DEBUG) System.out.println("inside lastBU, kev: " + key + eq + value);
		    
		    h.setLastBU(value);
		}
		else errorFlag = true;
	    }
	    else if (key.equals("U")){ 
		eq = st.nextToken();
		if (eq.equals("=")){
		    value = st.nextToken();
		    //if (value.equals("null"))
			h.utility = null;
			//else h.utility = Double.valueOf(value);
		}
		else errorFlag = true;
	    }
	    else {
		eq = st.nextToken();
		value = st.nextToken();
		h.put(key,value);
	    }
	if (errorFlag)
	    System.out.println(em + " errorFlag set for key: " + key + " eq: " + eq + " value: " + value);
	}
	h.getDStot();
	h.Uchanged = true;
	return h;
    }//readBid(StringTokenizer st)
    
    public static void main(String[] args){

//	ConstraintSet cs = ConstraintSet.readCSFromString( "{cost_max=100.0, length_max=100, length_min=0}");
//	
//	//BidUtility bu = BidUtility.readBidUtilityFromString("{length_coeff=1, DSg_coeff=0, DSv_coeff=0, length_target=20, RV_coeff=1, DSr_target=40, cost_coeff=1, cost_target=0, DSr_coeff=2}");
//	BidUtility bu = BidUtility.readBidUtilityFromString("{length_coeff=7, DSg_coeff=5, DSv_coeff=4, DSv_fcn=linear, length_target=40, DSg_target=5, RV_coeff=1, DSg_fcn=linear, DSr_target=5, DSr_fcn=linear, cost_coeff=1, cost_fcn=startPlateauNormal, DSv_target=5, cost_target=25, length_fcn=startPlateauNormal, DSr_coeff=1}");
//	BidUtility dsBU = bu.dsOnlyClone();
//      //Bid b1 = new Bid("dr");
//      //b1.put("cost_exp", "3"); //dr
//      //b1.put("length_exp", "10");
//      //b1.put("DSr_exp", new Strength(3));
//      //b1.put("DSv_exp", new Strength());
//      //b1.put("DSg_exp", new Strength());
//      //b1.put("RV", new RuleValue());
//      //b1.put("topic", "r");
//
//      BidSet bs = BidSet.readBidSetFromString("[{topic=v, length_exp=12.0, DStot_exp=[[0 0 0][6 6 1]], cost_exp=4, DSr_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][6 6 1]], RV=[[] 0], symbol=URS, U=12.8, bidder=RE}, {topic=v, length_exp=20.0, DStot_exp=[[2 2 1][6 6 1]], cost_exp=4.0, DSr_exp=[[0 0 0][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[2 2 1][6 6 1]], symbol=URS, RV=[[2] 2], U=6.6000000000000005, bidder=Value} owner=Coord]");
//	
//	Vector v = new Vector();
//	v.addAll(bs);
//	for (int i=0; i<v.size(); i++){
//	    Bid b = (Bid)v.elementAt(i);
//	    b.getUtility(bu);
//	    System.out.println(b);
//	    b.getUtility(dsBU);
//	    System.out.println(b);
//	    //b.calcAndPrintUtility(bu);
//	}
//	UtilComp uc = new UtilComp(bu);
//	Collections.sort(v, uc);
//
//       System.out.println("\n\nnew combined bids sorted: " + v);

Bid x = readBidFromString("{topic=v, length_exp=77.0, cost_exp=32.0, DSr_exp=[[3 4.01 2][0.51 0.51 1]], DSg_exp=[[0 0 0][1.25 2.37 2]], DSv_exp=[[1.1 2.2 2][0 0 0]], symbol=IBM, RV=[[] 0], U=null, bidder=Coord, lastBU=unset}");


BidUtility bu = BidUtility.readBidUtilityFromString("{DSr_coeff=4, DSv_coeff=4, DSg_coeff=2, DSg_target=5, DSr_target=5, DSv_target=5, cost_coeff=3, cost_target=30, RV_coeff=0.1, length_coeff=1, length_target=150, length_fcn=startPlateauNormal, DSr_fcn=linear, DSv_fcn=linear, DSg_fcn=linear, cost_fcn=startPlateauNormal}"); 

       //Bid a = readBidFromString("{topic=r, spam=97, length_exp=30.0, cost_exp=6.0, DSr_exp=[[1.39 2.39 2][0 4 0]], DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], symbol=IBM, RV=[[] 0], U=12.182374288832493, bidder=DR, lastBU=bu1}");
       //Bid b = readBidFromString("{topic=r, length_exp=30.0, spam=3.5, cost_exp=6.0, DSr_exp=[[1.39 2.39 2][0 0 0]], DSg_exp=[[6 6 0][0 0 0]], DSv_exp=[[6 0 0][0 0 0]], symbol=IBM, RV=[[] 0], U=null, bidder=DR, lastBU=bu1} ");

       DEBUG = true;

       x.calcAndPrintUtility(bu);

       //System.out.println("distance: " + a.cosineDistance(b));
       //       System.out.println("hcdu: " + b.getHiCurrDSUtil(bu));
       
    }

} // Bid

//{cost_exp=4, RV=[[] 0], DSv_exp=[[0 0 0][0 0 0]], length_exp=10, topic=r, DSr_exp=[[3 3 1][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=null, bidder=IR}, owner=IR]
//IR constraints: {length_max=100, length_min=0, cost_max=100.0}
//in Bid.java: getUtility(): {length_coeff=1, DSg_coeff=0, DSv_coeff=0, length_target=20, RV_coeff=1, DSr_target=40, cost_coeff=1, cost_target=0, DSr_coeff=2} trueoldU: null
//new U: 0.8772012261858588
