// Evan Steinberger
// CISC 466
// Spring 2007

import java.util.ArrayList;

/**
 * BlindEntry is used for building and examining history files.
 * 
 * @author Evan Steinberger
 */

public class BlindEntry
{
	public boolean isValid;
	
	public String date;
	public int totalSources;
	public String profile;
	public String stockSymbol;

	// Numeric data
	public double cr_mean, cr_confidence, cr_std, cr_sources_used;
	public double pe_mean, pe_confidence, pe_std, pe_sources_used;
	public double de_mean, de_confidence, de_std, de_sources_used;
	public double re_mean, re_confidence, re_std, re_sources_used;

	// Outlier Lists
	public ArrayList cr_outliers, pe_outliers, de_outliers, re_outliers;

	private String rawText;
	
	/**
	 * Builds a BlindEntry
	 * 
	 * @param rawText the raw text from a history file
	 */
	public BlindEntry(String rawText)
	{
		this.rawText = rawText;
		this.rawText += "\r\n\r\n";
		this.cr_outliers = new ArrayList();
		this.pe_outliers = new ArrayList();
		this.de_outliers = new ArrayList();
		this.re_outliers = new ArrayList();
		try
		{
			this.stockSymbol = this.rawText.substring(0, this.rawText.indexOf(" "));
			String crOut, peOut, deOut, reOut;
			String tcrMean, tcrCon, tcrStd, tcrSources;
			String tpeMean, tpeCon, tpeStd, tpeSources;
			String tdeMean, tdeCon, tdeStd, tdeSources;
			String treMean, treCon, treStd, treSources;
			String tTotalSources;

			this.date = this.parseDatum("date");
			tTotalSources = this.parseDatum("total sources");
			this.profile = this.parseDatum("profile");
			
			tcrMean = this.parseDatum("cr mean");
			tcrCon = this.parseDatum("cr confidence");
			tcrStd = this.parseDatum("cr std");
			tcrSources = this.parseDatum("cr sources used");
			crOut = this.parseDatum("cr outliers");

			tpeMean = this.parseDatum("pe mean");
			tpeCon = this.parseDatum("pe confidence");
			tpeStd = this.parseDatum("pe std");
			tpeSources = this.parseDatum("pe sources used");
			peOut = this.parseDatum("pe outliers");

			tdeMean = this.parseDatum("de mean");
			tdeCon = this.parseDatum("de confidence");
			tdeStd = this.parseDatum("de std");
			tdeSources = this.parseDatum("de sources used");
			deOut = this.parseDatum("de outliers");

			treMean = this.parseDatum("re mean");
			treCon = this.parseDatum("re confidence");
			treStd = this.parseDatum("re std");
			treSources = this.parseDatum("re sources used");
			reOut = this.parseDatum("re outliers");

			this.cr_mean = Double.parseDouble(tcrMean);
			this.cr_confidence = Double.parseDouble(tcrCon);
			this.cr_std = Double.parseDouble(tcrStd);
			this.cr_sources_used = Double.parseDouble(tcrSources);

			this.pe_mean = Double.parseDouble(tpeMean);
			this.pe_confidence = Double.parseDouble(tpeCon);
			this.pe_std = Double.parseDouble(tpeStd);
			this.pe_sources_used = Double.parseDouble(tpeSources);

			this.de_mean = Double.parseDouble(tdeMean);
			this.de_confidence = Double.parseDouble(tdeCon);
			this.de_std = Double.parseDouble(tdeStd);
			this.de_sources_used = Double.parseDouble(tdeSources);

			this.re_mean = Double.parseDouble(treMean);
			this.re_confidence = Double.parseDouble(treCon);
			this.re_std = Double.parseDouble(treStd);
			this.re_sources_used = Double.parseDouble(treSources);
			
			this.totalSources = Integer.parseInt(tTotalSources);
	
			this.parseOutliers(crOut, cr_outliers);
			this.parseOutliers(peOut, pe_outliers);
			this.parseOutliers(deOut, de_outliers);
			this.parseOutliers(reOut, re_outliers);

			this.isValid = true;
		}
		catch (Exception ex)
		{
			this.isValid = false;
		}
	}
	
	/**
	 * Parses a piece of data
	 * 
	 * @param property the property to search for
	 * @return the parsed piece of data
	 */
	private String parseDatum(String property)
	{
		int delimitLength = property.length() + 2;
		try
		{
			int startIndex;
			String temp;
			startIndex = this.rawText.lastIndexOf(this.stockSymbol + " " + property);
			if (startIndex < 0)
				return "";
			this.rawText = this.rawText.substring(startIndex);
			temp = this.rawText.substring(this.stockSymbol.length() + delimitLength, this.rawText.indexOf("\r\n"));
			return temp;
		}
		catch (Exception ex) { return ""; }
	}

	/**
	 * Parses Outliers
	 * 
	 * @param raw the raw text
	 * @param a an arraylist for adding outliers
	 */
	private void parseOutliers(String raw, ArrayList a)
	{
		if (raw.trim().equals("") || raw.trim().equals("none"))
			return;
		String parts[] = raw.split(",");
		for (String s : parts)
			a.add(s);
	}
}