import java.io.*;
import java.util.*;
/**
 * CosineDistance.java
 *
 * Right now this file is a jumble of misc stuff, including cosine
 * distance alg, but also stuff that finds a random array that is
 * distant from another array.
 *
 * Created: Fri Jun 11 07:37:28 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class CosineDistance {
    static boolean DEBUG = false;

    public CosineDistance() {
	
    } // CosineDistance constructor

    /** 
     * Returns similarity between the vectors of two arrays on unit
     * sphere. 1 means arrays are identical, 0 means no similarity
     **/
    public static double twoArraySimilarity(int[] array1, int[] array2){
    	 double numerator = 0; 
    	 double denom1 = 0;
    	 double denom2 = 0;
    	 double result = 0;
    	 for (int i = 0; i < array1.length; i++ ){
    	   numerator = numerator + (array1[i] * array2[i]);
    	   denom1 = denom1 + (array1[i] * array1[i] );
    	   denom2 = denom2 + (array2[i] * array2[i]);
    	 }
    	 result = numerator / java.lang.Math.sqrt(denom1 * denom2);
    	 //if (DEBUG) System.out.println(" COSINE : " + result);
    	 return result;
    }
    public static double twoArraySimilarity(double[] array1, double[] array2){
    	 double numerator = 0; 
    	 double denom1 = 0;
    	 double denom2 = 0;
    	 double result = 0;
    	 for (int i = 0; i < array1.length; i++ ){
    	   numerator += (array1[i] * array2[i]);
    	   denom1 += (array1[i] * array1[i] );
    	   denom2 += (array2[i] * array2[i]);
    	 }
    	 result = numerator / java.lang.Math.sqrt(denom1 * denom2);
    	 if (DEBUG) System.out.println(" COSINE : " + result);
    	 return result;
    }
    
    /**
     * Makes an array [0,1,2,...length-1] then shuffles it.
     **/
    public static int[] randomArray(int length){
	
	int[] a = new int[length];
	Vector v = new Vector();
	for (int i = 0; i < length; i++)
	    v.addElement(new Integer(i));
	Collections.shuffle(v);
	for (int i = 0; i < length; i++)
	    a[i] = ((Integer)v.elementAt(i)).intValue();
	return a;
    }

    /**
     * Gets an array whose vector is farthest from
     * [0,1,2,..length-1] by testing random arrays "trials" times.
     **/    
    public static int[] getDistantRandomArray(int length, int trials){
	int[] test = new int[length];
	int[] best = test;
	int[] next;
	double distance, bestDistance = 1;
	
	for (int i = 0; i < length; i++)
	    test[i] = i;

	for (int j = 0; j < trials; j++){
	    next = CosineDistance.randomArray(length);
	    distance = twoArraySimilarity(test, next);
	    if (DEBUG) {
		for(int i = 0; i < length; i++)
			System.out.print(next[i] + " ");
		System.out.print("         ");
		for(int i = 0; i < length; i++)
			System.out.print(best[i] + " ");
		
		System.out.print("         " + distance);
		System.out.print("\n");
	    }

	    //if (DEBUG) System.out.println(distance);

	    if (distance <= bestDistance){
		bestDistance = distance;
		best = next;
		
	    }

	}
	return best;
    }


    public static void main(String[] args){
	//int [] a = CosineDistance.randomArray(9);
	//
	//for(int i = 0; i < 9; i++)
	//    System.out.println(a[i]);
	//
	//int[][] b = new int[3][6];
	//for(int i = 0; i<3; i++)
	//    b[i] = CosineDistance.randomArray(6);
	//
	//for(int i = 0; i<3; i++)
	//    for(int j = 0; j < 6; j++)
	//	System.out.println(b[i][j]);

	int [] a = {1,2,3,4,5};
	int [] b = {1,2,3,40,5};
	for(int i = 0; i < a.length; i++)
	    System.out.print(a[i] + " ");
	System.out.println("");
	for(int i = 0; i < b.length; i++)
	    System.out.print(b[i] + " ");
	System.out.println("");
	System.out.println("distance: " + (1 - twoArraySimilarity(a,b)));
	

    }

} // CosineDistance
