// Evan Steinberger
// CISC 466
// Spring 2007


import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
import java.lang.reflect.*;
import java.io.*;

/**
 * A Demo Program for testing IAM outside of MADSUM
 * 
 * @author Evan Steinberger
 */

public class DemoFrame extends JFrame implements ActionListener
{
	// Constants 
	final private int xSize = 640;
	final private int ySize = 520;

	// GUI Components
	private JLabel l_info;
	private JLabel l_blindInfo;
	private JLabel l_grabberInfo;
	private JLabel l_informedInfo;
	private JLabel l_accessInfo;

	private JTextField t_queryfield;
	private JTextField t_madsum;
	private JButton b_submit;
	private JButton b_train;
	private JTextArea t_result;
	private JTextArea t_profileView;

	private JComboBox cb_blind = new JComboBox(this.addToComboBox(".blindprofile"));
	private JComboBox cb_grabber = new JComboBox(this.addToComboBox(".grabberprofile"));
	private JComboBox cb_informed = new JComboBox(this.addToComboBox(".informedprofile"));
	private JComboBox cb_access = new JComboBox(new String[] {"Zero", "TenMin", "Hour", "Day", "Week", "Month"});
	private JComboBox cb_vals = new JComboBox(new String[] {"cr", "pe", "de", "re"});
	private JComboBox cb_alligator = new JComboBox(new String[] {">", "<"});

	private JCheckBox chb_useMADSUM = new JCheckBox("Use MADSUM Threshold Value");
	
	/**
	 * Sets up the initial layout of the DemoFrame
	 */
	public DemoFrame() 
	{
		this.setLayout(null);

		l_info = new JLabel("Enter Stock Symbol:");
		l_info.setBounds(5, 30, 200, 20);
		this.add(l_info);
		
		t_queryfield = new JTextField();
		t_queryfield.setBounds(5, 50, 200, 20);
		t_queryfield.setText("IBM");
		this.add(t_queryfield);

		l_blindInfo = new JLabel("Select Blind Profile:");
		l_blindInfo.setBounds(5, 335, 150, 20);
		this.add(l_blindInfo);
		this.cb_blind.setSelectedItem("standard");
		this.cb_blind.setBounds(5, 355, 150, 20);
		this.cb_blind.setActionCommand("blind");
		this.cb_blind.addActionListener(this);
		this.add(cb_blind);

		l_grabberInfo = new JLabel("Select Grabber Profile:");
		l_grabberInfo.setBounds(160, 335, 150, 20);
		this.add(l_grabberInfo);
		this.cb_grabber.setSelectedItem("standard");
		this.cb_grabber.setBounds(160, 355, 150, 20);
		this.cb_grabber.setActionCommand("grabber");
		this.cb_grabber.addActionListener(this);
		this.add(cb_grabber);

		l_informedInfo = new JLabel("Select Informed Profile:");
		l_informedInfo.setBounds(315, 335, 150, 20);
		this.add(l_informedInfo);
		this.cb_informed.setSelectedItem("standard");
		this.cb_informed.setBounds(315, 355, 150, 20);
		this.cb_informed.setActionCommand("informed");
		this.cb_informed.addActionListener(this);
		this.add(cb_informed);

		l_accessInfo = new JLabel("Select Access Threshold:");
		l_accessInfo.setBounds(470, 335, 150, 20);
		this.add(l_accessInfo);
		this.cb_access.setSelectedItem("Day");
		this.cb_access.setBounds(470, 355, 150, 20);
		this.add(cb_access);

		chb_useMADSUM.setBounds(415, 30, 200, 20);
		this.add(chb_useMADSUM);
		cb_vals.setBounds(415, 50, 50, 20);
		this.add(cb_vals);
		cb_alligator.setBounds(481, 50, 50, 20);
		this.add(cb_alligator);
		t_madsum = new JTextField();
		t_madsum.setBounds(547, 50, 70, 20);
		this.add(t_madsum);

		b_submit = new JButton("Submit");
		b_submit.setBounds(210, 50, 100, 20);
		b_submit.setActionCommand("Submit");
		b_submit.addActionListener(this);
		this.add(b_submit);

		b_train = new JButton("Train top NYSE data");
		b_train.setBounds((xSize-155), 0, 150, 20);
		b_train.setActionCommand("Train");
		b_train.addActionListener(this);
		this.add(b_train);

		t_result = new JTextArea();
		//t_result.setLineWrap(true);
		t_result.setEditable(false);
		JScrollPane jsp = new JScrollPane(t_result);
		jsp.setBounds(5, 75, xSize-15, 255);
		this.add(jsp);

		t_profileView = new JTextArea();
		t_profileView.setEditable(false);
		t_profileView.setBounds(5, 380, xSize-15, 110);
		this.add(t_profileView);
		
		this.setTitle("Informed Analysis Module (IAM) Demo, Evan Steinberger 2007");
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.setSize(xSize,ySize);
		this.setLocationRelativeTo(null);
		this.setResizable(false);
		this.setVisible(true);
	}
	
	/**
	 * Handles all button clicks and drop down menu changes
	 * 
	 * @param e the action event happening
	 */
	public void actionPerformed(ActionEvent e) 
	{
		if("Submit".equals(e.getActionCommand())) 
		{
			
			InformedAnalyzer IA = this.setData();
			if (IA != null)
			{
				IA.analyze();
				t_result.setText( IA.displayAllInfo() );
			}

		}
		if("Train".equals(e.getActionCommand()))
		{

			if (JOptionPane.showConfirmDialog(null, "Are you sure you wish to do this?\r\nAll history files will be wiped and IAM will begin training on the top 50 NYSE entries.\r\nEnsure all profiles are correct, the process will take some time.", "Check and Confirm", JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION)
			{
				File workingDir = new File(".");
				File[] files = workingDir.listFiles();
				String s="";
				for (File f : files)
				{
					String name = f.getName();
					if (name.endsWith("Grabber.txt") ||
						name.endsWith("Analysis.txt") ||
						name.endsWith("Analyzer.txt") ||
						name.endsWith("Results.txt"))
					{
						System.out.println("Deleting " + name);
						f.delete();
					}
				}
				InformedAnalyzer IA = this.setData();
				if (IA != null)
				{
					IA.automatedDataCollection();
					t_result.setText( "done");
				}
			}
		}
		if("blind".equals(e.getActionCommand()))
		{
			String name = cb_blind.getSelectedItem().toString() + ".blindprofile";
			String s = "--Viewing " + name + "--\r\n";
			s += ReadData.fromFile(name);
			t_profileView.setText(s);
		}
		if("grabber".equals(e.getActionCommand()))
		{
			String name = cb_grabber.getSelectedItem().toString() + ".grabberprofile";
			String s = "--Viewing " + name + "--\r\n";
			s += ReadData.fromFile(name);
			t_profileView.setText(s);
		}
		if("informed".equals(e.getActionCommand()))
		{
			String name = cb_informed.getSelectedItem().toString() + ".informedprofile";
			String s = "--Viewing " + name + "--\r\n";
			s += ReadData.fromFile(name);
			t_profileView.setText(s);
		}
	}
	
	/**
	 * Sets the data for IAM
	 * 
	 * @return a copy of the Informed Analyzer
	 */
	private InformedAnalyzer setData()
	{
		String s = cb_access.getSelectedItem().toString();
		StockInfoGrabber.AccessThreshold at;
		try { at = StockInfoGrabber.parseAccessThreshold(s); }
		catch (Exception ex) { System.out.println(ex.toString()); at = StockInfoGrabber.AccessThreshold.Day; }

		boolean useMadsum = false;
		boolean greaterThan = false;
		double madsumValue = 0;
		String madsumWhich = "";
		if (chb_useMADSUM.isSelected())
		{
			useMadsum = true;
			if (cb_alligator.getSelectedItem().toString().equals(">"))
				greaterThan = true;
			try { madsumValue = Double.parseDouble(t_madsum.getText()); }
			catch (Exception ex)
			{
				t_result.setText("Invalid Threshold Value Set.");
				return null;
			}
			madsumWhich = cb_vals.getSelectedItem().toString();
		}
			
		InformedAnalyzer IA = new InformedAnalyzer(t_queryfield.getText(),
			cb_blind.getSelectedItem().toString(),
			cb_grabber.getSelectedItem().toString(),
			cb_informed.getSelectedItem().toString(),
			at, useMadsum, madsumValue, greaterThan, madsumWhich);
		return IA;
	}
	
	/**
	 * Searches for profiles in the current directory
	 * 
	 * @param extension the extension of the profile file
	 * @return a string array of profiles found
	 */
	private String[] addToComboBox(String extension)
	{
		File workingDir = new File(".");
		File[] files = workingDir.listFiles();
		String s="";
		for (File f : files)
		{
			if (f.getName().endsWith(extension))
				s+= f.getName().substring(0, f.getName().indexOf(extension)) + "-,-";
		}
		s=s.substring(0,s.length()-3);
		return s.split("-,-");
	}
	
	/**
	 * Runs the program
	 */
	public static void main (String[] args)
	{
		DemoFrame wqf = new DemoFrame();
	}

}