import java.util.*;
import java.io.*;
/**
 * GenTask.java
 *
 * A superclass for DECAF Tasks in the PL ontology.
 *
 * Created: Tue Apr 13 21:58:42 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public abstract class GenTask {

    Hashtable ht; //points to Local agent's data hashtable
    String myEndTime; //time by which this task should have completed and returned results
    private static final boolean DEBUG1 = false;
    private static final boolean PRINT_TO_FILE = true;
    protected String job;
    private static final String loc = " in GenTask: ";
    private static final String wm = "Warning in GenTask: ";
    private static final String em = "ERROR in GenTask: ";
    //protected int simulateDelay = 3000; //milliseconds (now in AgentDef.java)

    Hashtable bidsAccepted;
    Hashtable bidsMade;

    public GenTask() {

    } // GenTask constructor

    protected String getMsgJob(KQMLmsg k){

       String msgJob = k.getContentFieldValue("job");
       if (msgJob.equals(""))
	   System.out.println(em + " job is empty in " + k);
       
       return msgJob;
    }

    /********************************************************************
    For handling opportunity (failing up); required by multiple PL_Executes
    *********************************************************************/
    protected boolean isOpportunity(KQMLmsg k){
	String s = k.getContentFieldValue("opportunity");
	if (s.equals("")) return false;
	else return true;
    }

    /*
     * Construct message for local.Execute to send to processAgent
     * when processing an opportunity (failing up).
     *
     * toExecute is the original msg from FailingUp to Execute task.
     */
    protected KQMLmsg oppMsgToProcess(String result, String failedResult, 
				      Envelope env, KQMLmsg toExecute, Agent Local){
	if ( ! isOpportunity(toExecute) ){
	    System.out.println(em + " is not an opportunity msg");
	    return null;
	}

	String finalDestAgent = toExecute.getContentFieldValue("finalDest");	

	String receiver;

	if (Local.isNumeric()){ 
	    if (finalDestAgent.equals(Local.getParent()))
		receiver = Local.getName();//use own process to send DataSet to parent
	    else
		receiver = Local.getParent();//use parent process to send tree
	}
	else 
	    receiver = Local.getName();

	KQMLmsg K = new KQMLmsg();
	K.addFieldValuePair("sender", Local.getName() );
	K.addFieldValuePair("performative", "achieve");
	K.addFieldValuePair("receiver", receiver );
	K.addFieldValuePair("ontology", "PL" );
	K.addFieldValuePair("language", "DECAF");
	MsgUtilities.copyGenContent(toExecute, K, Local);
	MsgUtilities.addContentFieldValuePair(K, "opportunity", "yes");
	MsgUtilities.addContentFieldValuePair(K, "finalDest", finalDestAgent);
	MsgUtilities.addContentFieldValuePair(K, "results", result);
	MsgUtilities.addContentFieldValuePair(K, "failedResults", failedResult);
	MsgUtilities.addContentFieldValuePair(K, "env", env.toString());
	MsgUtilities.addContentFieldValuePair(K, "task", "ProcessAgent");
	MsgUtilities.addContentFieldValuePair(K, "msg3", "opp");
	return K;
    }//oppMsgToProcess(String result, String failedResult, Envelope env, KQMLmsg toExecute, Agent Local)

//    protected KQMLmsg executeOppMsgToCount(KQMLmsg toExecute, Agent Local){
//	String receiver = toExecute.getContentFieldValue("finalDest");
//	KQMLmsg K = new KQMLmsg();
//	K.addFieldValuePair("sender", Local.getName() );
//	K.addFieldValuePair("performative", "achieve");
//	K.addFieldValuePair("receiver", receiver );
//	K.addFieldValuePair("ontology", "PL" );
//	K.addFieldValuePair("language", "DECAF");
//	MsgUtilities.copyGenContent(toExecute, K, Local);
//	MsgUtilities.addContentFieldValuePair(K, "opportunity", "yes");
//	MsgUtilities.addContentFieldValuePair(K, "task", "Count");
//	MsgUtilities.addContentFieldValuePair(K, "replies", "opp");
//	MsgUtilities.addContentFieldValuePair(K, "bidder", toExecute.getValue("sender"));
//	MsgUtilities.addContentFieldValuePair(K, "env", toExecute.getContentFieldValue("env"));
//	return K;
//    }

    /*
     * Construct message for processAgent to send to authorizing
     * agent's Count when processing an opportunity (failing up)
     *
     * toProcess is the original msg from execute to Process task.
     */
//    protected KQMLmsg oppMsgToCount(KQMLmsg toProcess, Agent Local){
//	String receiver = toProcess.getContentFieldValue("finalDest");
//	KQMLmsg K = new KQMLmsg();
//	K.addFieldValuePair("sender", Local.getName() );
//	K.addFieldValuePair("performative", "achieve");
//	K.addFieldValuePair("receiver", receiver );
//	K.addFieldValuePair("ontology", "PL" );
//	K.addFieldValuePair("language", "DECAF");
//	MsgUtilities.copyGenContent(toProcess, K, Local);
//	MsgUtilities.addContentFieldValuePair(K, "opportunity", "yes");
//	MsgUtilities.addContentFieldValuePair(K, "task", "Count");
//	MsgUtilities.addContentFieldValuePair(K, "replies", "opp");
//	MsgUtilities.addContentFieldValuePair(K, "bidder", toProcess.getValue("sender"));
//	MsgUtilities.addContentFieldValuePair(K, "env", toProcess.getContentFieldValue("env"));
//	return K;
//    }

    /**********************************************************************/
    /**
     * WaitingForSender is for keeping track of how many envelopes
     * have been sent to an Agent.
     *
     * WaitingForSender should be Agent_name/count hash. Don't want to
     * store whole bids, they are kept in bidsSentForExecution, where
     * they stay; if they were in Waiting, they would get deleted and
     * then could be repeated.
     **/

    /*
     * called in PL_Count, PL_FailingUp
     */
    protected void addWaitingForSender(String name, String job, String source, String loc){
	Hashtable waitingForSenders = (Hashtable)ht.get("waitingForSenders" + job);
	int count = 0;
	if (waitingForSenders == null){
	    System.out.println(" addWaitingForSender Error: waitingForSenders is null*********************************");
	    writeAgentMsgToFile(" TRIED adding " + name + " to waitingForSenders" + loc, source);
	}
	else
	    synchronized(waitingForSenders)
		{
		    if (waitingForSenders.containsKey(name)){
			count = Integer.parseInt((String)waitingForSenders.get(name));
			waitingForSenders.put(name, Integer.toString(count + 1));
			//System.out.println("TESTING: waitingForSender adding " + name + " from " + count + " to " + (count+1));
			writeAgentMsgToFile(" adding " + name + " to waitingForSenders (a), now " + (count + 1) + loc , source);
		    }
		    else{
			waitingForSenders.put(name, Integer.toString(1));
			writeAgentMsgToFile(" adding " + name + " to waitingForSenders (b), now 1" + loc , source);
		    }
		}
    }

    protected int getWaitingForSender(String name, String job){
	Hashtable waitingForSenders = (Hashtable)ht.get("waitingForSenders" + job);
	int count;
	if (waitingForSenders == null)
	    System.out.println("getWaitingForSender Error: waitingForSenders is null*********************************");
	if (waitingForSenders.containsKey(name)){
	    count = Integer.parseInt((String)waitingForSenders.get(name));
	    //System.out.println("TESTING: waitingForSender getting " + name + " with " + count);
	    return count;
	}
	else{
	    System.out.println("getWaitingForSender: no such sender, " + name + ". returning zero");
	    return 0;
	}
    }

    protected int getNumberOfWaits(String job){
	Hashtable waitingForSenders = (Hashtable)ht.get("waitingForSenders" + job);
	Iterator i = waitingForSenders.values().iterator();
	int next, sum = 0;
	while (i.hasNext()){
	    next = Integer.parseInt((String)i.next());
	    if (next > 0)
		sum += next;
	}
	return sum;
    }

    /** not currently behaving as boolean! **/
    protected boolean removeWaitingForSender(String name, String job, String source, String loc){
	Hashtable waitingForSenders = (Hashtable)ht.get("waitingForSenders" + job);
	int count;
	if (waitingForSenders == null){
	    System.out.println("removeWaitingForSender Error: waitingForSenders is null*********************************");
	    writeAgentMsgToFile(" TRIED removing " + name + " from waitingForSenders " + loc, source);
	    return false;
	}
	else if (waitingForSenders.containsKey(name)){
	    synchronized(waitingForSenders)
		{
		count = Integer.parseInt((String)waitingForSenders.get(name));
		//if (count == 0){
		//	if (DEBUG1) System.out.println("removeWaitingForSender: ERROR " + name + ".has count==0");
		//	return false;
		//}
		//else if (count > 0){
		waitingForSenders.put(name, Integer.toString(count - 1));
		//if ((count-1) == 0)
		//    waitingForSenders.remove(name);
		writeAgentMsgToFile(" removing " + name + " from waitingForSenders (a), now " + (count - 1) + loc, source);
	    }
	    return true;
	}
	//else return false;
	
	else{
	    waitingForSenders.put(name, Integer.toString( -1 ));
	    System.out.println("removeWaitingForSender: no such sender, " + name + ", entering -1");
	    writeAgentMsgToFile(" removing " + name + " from waitingForSenders (b), now -1 " + loc, source);
	    return true;
	}

    }

    protected boolean waitingForSenders(String job){
    	//Hashtable waitingForSenders = (Hashtable)ht.get("waitingForSenders" + job);
	return getNumberOfWaits(job) != 0; //! waitingForSenders.isEmpty();
    }
    protected String showWaitingForSenders(String job){
    	Hashtable waitingForSenders = (Hashtable)ht.get("waitingForSenders" + job);
	if (waitingForSenders != null && ! waitingForSenders.isEmpty())
	    return waitingForSenders.toString();
	else return " empty ";
    }
    /**********************************************************************/
    ///Returns null if object is not present
    protected Object localGet(Object key, String job){
	if (ht == null){
	    System.out.println("Get Error: " + key + " ht is null for job "+ job +"*********************************");
	    return null;
	}
	else
	    synchronized(ht){
		if (ht.containsKey(key.toString() + job))
		    return ht.get(key.toString() + job);
		else return null;
	    }
    }

    protected Object localPut(Object key, String job, Object value){
	if (ht == null){
	    System.out.println("Put Error:  " + key + " ht is null********************************");
	    return null;
	}
	else {
	    synchronized(ht){
		HashSet keys = (HashSet)ht.get("jobStorageKeys" + job);
		if (keys != null){
		    keys.add(key.toString() + job);
		}
		return ht.put(key.toString() + job, value);
	    }
	}
    }

    /********************* Bidding fcns ******************************/

    public Set getBidsMade(){
	if (bidsMade == null){
	    System.out.println(em + "getBidsMade() NO HASHTABLE bidsMade");
	    return null;
	}
	else synchronized(bidsMade){
	    return bidsMade.keySet();
	}
    }

    public boolean madeBidP(Bid b){
	if (bidsMade == null){
	    System.out.println(em + "madeBidP NO HASHTABLE bidsMade");
	    return false;
	}
	else synchronized(bidsMade){
	    return bidsMade.containsKey(b);
	}
    }    
    public BidSet getBidSetMade(String agentName){
	if (bidsMade == null){
	    System.out.println(em + "getBidSetMade NO HASHTABLE bidsMade");
	    return null;
	}
	else synchronized(bidsMade){
	    return new BidSet(agentName,(Collection)bidsMade.keySet());}
    }

    public void setBidsMade(BidSet bs){
	if (bidsMade == null)
	    System.out.println(em + "setBidsMade NO HASHTABLE bidsMade");
	
	else synchronized(bidsMade){
	    Iterator i = bs.iterator();
	    Bid b;
	    while (i.hasNext()){
		b = (Bid)i.next();
		bidsMade.put(b, b.getBidSetSource());
		if (b.getBidSetSource() == null)
		    System.out.println(wm + "bid " + b + " has null source");
	    }
	}
    }

    public BidSet getBidComponents(Bid b){
	if (bidsMade == null)
	    System.out.println(em + "getBidComponents NO HASHTABLE bidsMade");

	else synchronized(bidsMade){
	    Iterator i = getBidsMade().iterator();
	    Bid stored;
	    while(i.hasNext()){
		stored = (Bid)i.next();
		//System.out.println(stored + "EQUALS: " + b.equals(stored));
		if (b.equals(stored)){
		    if (DEBUG1) System.out.println(loc + 
						   "getBidComponents found bid " + b + "set " + bidsMade.get(stored));
		    return (BidSet)bidsMade.get(stored);
		}

	    }
	    //return in while failed to fire
	    System.out.println(wm + "getBidComponents sources not found for " + b);
	}
	return null;
    }

    public void addAcceptedBid(Bid b, Envelope e) { 
	if (bidsAccepted == null) 
	    System.out.println(em + " bidsAccepted NO HASHTABLE bidsMade");
	else synchronized(bidsAccepted){
	    bidsAccepted.put(b, e);
	}
    }



    /**********************************************************************/

    /**
     * TWO versions of this, one with endtime and Agent parameter
     **/
    protected void writeAgentMsgToFile(String msg, String endtime, Agent Local){
	//System.out.println("WRITEAGENTMSG TO FILE ################### " + msg + "," + endtime + "," + Local.getName());

	if (PRINT_TO_FILE){
	    boolean repeat = true;
	    while (repeat){
		try {
		    String outputFileName =  "debuglog" + Local.getName();
		    FileWriter fw = new FileWriter(outputFileName, true);
		    
		    fw.write("job: " + job + " ");
		    fw.write(msg);
		    if (! endtime.equals("")){
			double time = (Local.getTimeRemaining(endtime))/1000.0;
			fw.write("\ntime remaining: " + time);
		    }
		    fw.write("\n______________________________________________________________\n");
		    fw.flush();
		    fw.close();
		    repeat = false;
		} 
		catch (Exception e) {
		    System.out.println(e);
		}
	    }   
	}
    }
    
    protected void writeAgentMsgToFile(String msg, String source){

	if (PRINT_TO_FILE){
	    boolean repeat = true;
	    while (repeat){
		try {
		    String outputFileName =  "debuglog" + source;
		    FileWriter fw = new FileWriter(outputFileName, true);
		    
		    fw.write("job: " + job + " ");
		    fw.write(msg);
		    fw.write("\n______________________________________________________________\n");
		    fw.flush();
		    fw.close();
		    repeat = false;
		} 
		catch (Exception e) {
		    System.out.println(e);
		}
	    }   
	}
    }
    

    

    public static void main(String[] a){
	KQMLmsg k = new KQMLmsg("(achieve :sender Coord :receiver Coord :ontology PL :language DECAF :content (:task DistrEnv :newEnv Env0 :symbol IBM :price 126 :quantity 100 :endtime 1082351207261 :levels 4 :job job0))");
	String job = "spam";
	PL_DistrEnv de = new PL_DistrEnv();
//	de.ht = new Hashtable();
//	de.ht.put("waitingForSenders" + job, new Hashtable() );
//	de.addWaitingForSender("risk", job);
//	System.out.println(de.ht.get("waitingForSenders" + job));
//	de.addWaitingForSender("risk", job);
//	de.addWaitingForSender("spam", job);
//	System.out.println(de.ht.get("waitingForSenders" + job));
//
//	de.removeWaitingForSender("risk", job);
//	System.out.println(de.getWaitingForSender("risk" ,job));
//	System.out.println(de.ht.get("waitingForSenders" + job));
//	de.removeWaitingForSender("risk", job);
//	System.out.println(de.getWaitingForSender("risk" ,job));
//	System.out.println(de.ht.get("waitingForSenders" + job));
//	de.removeWaitingForSender("risk", job);
//	System.out.println(de.getWaitingForSender("risk" ,job));
//	System.out.println(de.ht.get("waitingForSenders" + job));
//	System.out.println("waitingForSenders:" + de.waitingForSenders(job));
//	System.out.println("waitingForSenders:" + de.showWaitingForSenders(job));
//	System.out.println("getNumberOfWaits:" + de.getNumberOfWaits(job));
	
     }
    
} // GenTask
