import java.util.*;
/**
 * Integrator.java
 *
 *
 * Created: Tue Jan 20 12:39:08 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class Integrator {

    static RuleSet rules;

    static boolean DEBUG = false;
    static final String loc = "in Integrator ";
    static final String em = "ERROR in Integrator: ";
    BidUtility bidU = null;

    Hashtable treeComponents; //for keeping track of which trees are
			      //inside other trees so that they don't
			      //get combined multiple times.

    public Integrator() {
	if (rules == null){
	    Vector rv = new Vector();
	    rv.add(new Integer(2));
	    rv.add(new Integer(3));
	    rv.add(new Integer(1));
	    rv.add(new Integer(5));
	    rv.add(new Integer(6));
	    //rv.add(new Integer(4));
	    rules = new RuleSet(rv);
	}
	treeComponents = new Hashtable();
    } // Integrator constructor
    
    public Integrator(BidUtility bu) {
	bidU = bu;//right now there is no sorting during integration; 
	          //if there is, use bu.dsOnlyClone() NOTE: fcn is called with dsOnlyClone of bu
	if (rules == null){
	    Vector rv = new Vector();
	    rv.add(new Integer(2));
	    rv.add(new Integer(3));
	    rv.add(new Integer(1));
	    rv.add(new Integer(5));
	    rv.add(new Integer(6));
	    //rv.add(new Integer(4));
	    rules = new RuleSet(rv);
	}
	treeComponents = new Hashtable();
    } // Integrator constructor
    
    /**
     * Take two top trees, try to integrate; then take result and try
     * to add to it; If can't, then find two new trees to integrate,
     * recurse. Put result in vector parameter.
     */
    public Vector addNewTrees(Vector trees, String maker){
	int index1, index2, laterIndex;
	Vector newTrees = new Vector(10);
	//hash new trees with vector containing tree components
	Hashtable associated = new Hashtable(10);
	
	for(index1 = 0; index1 < trees.size() - 1; index1++){
	    for(index2 = index1 + 1; index2 < trees.size(); index2++){
		synchronized(trees){
			applyRuleSet((RSTTree)trees.get(index1), (RSTTree)trees.get(index2), newTrees, maker, associated);
		}
	    }
	}

	//trees.addAll(newTrees);
	trees = subAddNewTrees(trees, newTrees, maker, associated);
	return trees;
    }//    public Vector addNewTrees(Vector trees)

    private Vector subAddNewTrees(Vector trees, Vector newTrees, String maker, Hashtable assoc){
	int index1, index2;
	Vector moreNewTrees = new Vector(10);
	for(index1 = 0; index1 < newTrees.size() - 1; index1++){
	    for(index2 = 0; index2 < trees.size(); index2++){
		if ( ! sharesComponentTrees((RSTTree)newTrees.get(index1), 
					    (RSTTree)trees.get(index2)) )
		    synchronized(trees){
			applyRuleSet((RSTTree)newTrees.get(index1), 
				   (RSTTree)trees.get(index2), moreNewTrees, maker, assoc);
		    }
	    }
	}
	for(int i = 0; i < newTrees.size(); i++)
	    ((RSTTree)newTrees.get(i)).updateUtility(bidU);
	trees.addAll(newTrees);
	if (moreNewTrees.isEmpty())
	    return trees;
	else 
	    return subAddNewTrees(trees, moreNewTrees, maker, assoc);
    }

    public RSTTree joinAllTrees(Vector trees, String maker){
	int index;
	if (trees.isEmpty()) return null;

	RSTTree bigTree, newTree = null;
	RSTTree next;
	synchronized(trees){
	    bigTree = (RSTTree)trees.get(0);
	    newTree = bigTree;
	    if (DEBUG) System.out.println("size = " + trees.size());
	    
 	    for(index = 1; index < trees.size(); index++){
		next = (RSTTree)trees.get(index);
		if (DEBUG) System.out.println("Integrator considering making new joint tree.");
		if (! (sharesComponentTrees(bigTree, next))){

		    if (DEBUG) System.out.println("bigTree " + bigTree + bigTree.getRoot() +  " components: " + 
			       treeComponents.get(bigTree));
		    //bigTree.print();
		    if (DEBUG) System.out.println("next " + next + next.getRoot() + " components: " + 
				       treeComponents.get(next));
		    //next.print();
		    Rule8 r8 = new Rule8();
		    newTree = r8.apply(bigTree, next, maker);
		    addComponentTree(newTree, bigTree);
		    addComponentTree(newTree, next);
		    if (DEBUG) System.out.println("Integrator making new joint tree:");
		    if (DEBUG) System.out.println(newTree.toString() + " components: " + 
				       treeComponents.get(newTree));
		    bigTree = newTree;
		}
		else
		    if (DEBUG) System.out.println("not joined, share components: " + bigTree + bigTree.getRoot() + next + next.getRoot());
		
	    }
	}

	return newTree;
    }// joinAllTrees(Vector trees, String maker)

    public RSTTree joinRandomTrees(Vector trees, String maker){
	int index;
	if (trees.isEmpty()) return null;

	RSTTree bigTree, newTree = null;
	RSTTree next;
	synchronized(trees){
	    bigTree = (RSTTree)trees.get(0);
	    if (DEBUG) System.out.println("joinRandom size = " + trees.size());
	    
 	    for(index = 1; index < trees.size(); index++){
		next = (RSTTree)trees.get(index);
		if (DEBUG) System.out.println("Integrator considering making random joint tree.");
		if (DEBUG) next.print();
		Rule9 r9 = new Rule9();
		newTree = r9.apply(bigTree, next, maker);
		if (DEBUG) System.out.println("Integrator making new joint tree:");
		bigTree = newTree;
	    }
	}

	return newTree;
    }// joinRandomTrees(Vector trees, String maker)

    //adds new trees derived from RuleSet to Vector v
    private void applyRuleSet(RSTTree t1, RSTTree t2, Vector v, String maker, Hashtable assoc){

	Vector vNew = rules.applyRules(t1, t2, maker);
	for(int i = 0; i < vNew.size(); i++)
	    storeNewTree((RSTTree)vNew.elementAt(i), v, t1, t2);
	
    }

    private void storeNewTree(RSTTree newTree, Vector v, RSTTree t1, RSTTree t2){
	synchronized(newTree){
	    v.add(newTree);
	    addComponentTree(newTree, t1);
	    addComponentTree(newTree, t2);
	    if (DEBUG) System.out.println("storeNewTree new integrated tree: ");
	    if (DEBUG) newTree.print();	
	    if (DEBUG) System.out.println( "t1 " + t1 + t1.getRoot() + " components: " + 
			       treeComponents.get(t1));
	    if (DEBUG) t1.print();
	    if (DEBUG) System.out.println( "t2 " + t2 + t2.getRoot() + " components: " + 
			       treeComponents.get(t2));
	    if (DEBUG) t2.print();
	    if (DEBUG) System.out.println("newTree " + newTree + newTree.getRoot() + " components: " + 
			       treeComponents.get(newTree));
	}
	//newTree = null;
    }

    //self is always a component
    public void addComponentTree(RSTTree parent, RSTTree part){
	HashSet hParent = null;
	if (DEBUG) System.out.println("addComponentTree: " + parent + " " + part);
	
	if (treeComponents.containsKey(parent))
	    hParent = (HashSet)treeComponents.get(parent);
	else {
	    hParent = new HashSet(10);
	    hParent.add(parent);
	    treeComponents.put(parent, hParent);
	}

	hParent.add(part);
	
	if (treeComponents.containsKey(part)){
	    HashSet hPart = (HashSet)treeComponents.get(part);
	    hParent.addAll(hPart);
	}
    }

    public boolean isAComponentTree(RSTTree parent, RSTTree part){
	if (treeComponents.containsKey(parent))
	    return ((HashSet)treeComponents.get(parent)).contains(part);
	else return false;
    }

    public boolean sharesComponentTrees(RSTTree t1, RSTTree t2){
	if (DEBUG) System.out.println("inside Shares");
	HashSet hT1, hT2;
	if (treeComponents.containsKey(t1)){
	    hT1 = (HashSet)treeComponents.get(t1);
	    if (treeComponents.containsKey(t2)){
		hT2 = (HashSet)treeComponents.get(t2);
		HashSet intersection = new HashSet(hT1);
		intersection.retainAll(hT2);
		if ( ! intersection.isEmpty()) return true;
		else return false;
	    }
	    else return hT1.contains(t2);
	}
	else if (treeComponents.containsKey(t2)){
	    hT2 = (HashSet)treeComponents.get(t2);
	    return hT2.contains(t1);
	}
	else
	    return t1 == t2;
    }



    public static void main(String[] args){

	Vector trees = new Vector();
	trees.add(RSTTree.GoalDo("test1", DSType.RISK, 3, new Bid()));
	trees.add(RSTTree.GoalDo("Short term debt is very low", DSType.RISK, 3, new Bid()));
	trees.add(RSTTree.GoalDoNot("Short term debt is very high", DSType.RISK, -3, new Bid()));
       ///sort trees!!
       
       if (DEBUG) {
	   System.out.println(loc + ".all trees before sort: ");
	   for(int i = 0; i < trees.size(); i++)
	       ((RSTTree)trees.get(i)).print();
       }

       BidUtility bu, bu2;
       
       bu = BidUtility.readBidUtilityFromString("{length_coeff=10, DSg_coeff=1, DSv_coeff=5, DSv_fcn=linear, length_target=85, DSg_target=5, RV_coeff=1, DSr_target=5, DSg_fcn=linear, DSr_fcn=linear, cost_coeff=10, cost_fcn=zeroNormal, DSv_target=5, cost_target=15, length_fcn=zeroNormal, DSr_coeff=3}");
       System.out.println(bu);
       BidUtility DSbu = bu.dsOnlyClone();
       System.out.println(DSbu);

       Integrator i1 = new Integrator(DSbu);

       RSTTreeComp rc = new RSTTreeComp(DSbu); //need to put in a bidUtility
       Collections.sort(trees,rc);
       if (true) {
	   System.out.println(loc + ". all trees after sort: ");
	   for(int i = 0; i < trees.size(); i++)
	       ((RSTTree)trees.get(i)).print();
       }
       Vector trees2 = i1.addNewTrees(trees, "spam");
       Collections.sort(trees2,rc);
       System.out.println(loc + " all trees after rules: ");
       for(int i = 0; i < trees2.size(); i++)
	   ((RSTTree)trees2.get(i)).print();
       RSTTree big = i1.joinAllTrees(trees2, "spam");
       System.out.println("big:");
       
       big.print();
  
    }

} // Integrator

/*
	RSTTree newTree = null;		
	
	newTree = rules.Rule2(t1,t2, maker);
	if (newTree == null)
	    newTree = rules.Rule2(t2,t1, maker);
	if (newTree != null){//This is not an else!!
	    storeNewTree(newTree, v, t1, t2);
	}
	
	newTree = rules.Rule3(t1,t2, maker);
	if (newTree == null)
	    newTree = rules.Rule2(t2,t1, maker);
	if (newTree != null){//This is not an else!!
	    storeNewTree(newTree, v, t1, t2);
	}
	
	newTree = rules.Rule4(t1,t2, maker);
	if (newTree == null)
	    newTree = rules.Rule2(t2,t1, maker);
	if (newTree != null){//This is not an else!!
	    storeNewTree(newTree, v, t1, t2);
	}
	
	newTree = rules.Rule1(t1,t2, maker);
	if (newTree == null)
	    newTree = rules.Rule2(t2,t1, maker);
	if (newTree != null){//This is not an else!!
	    storeNewTree(newTree, v, t1, t2);
	}
	
	newTree = rules.Rule5(t1,t2, maker);
	if (newTree == null)
	    newTree = rules.Rule2(t2,t1, maker);
	if (newTree != null){//This is not an else!!
	    storeNewTree(newTree, v, t1, t2);
	}
	newTree = rules.Rule6(t1,t2, maker);
	if (newTree == null)
	    newTree = rules.Rule2(t2,t1, maker);
	if (newTree != null){//This is not an else!!
	    storeNewTree(newTree, v, t1, t2);
	}
*/
