/*
DECAF (Distributed, Environment-Centered, Agent Framework) is a toolkit 
which allows a well-defined software engineering approach to building 
multi-agent systems. It is essentially, the internal "operating system" 
of a software agent, to which application programmers have strictly 
limited access.

This material is based upon work supported by the National Science 
Foundation under Grant No. IIS-9812764. Any opinions, findings and 
conclusions or recomendations expressed in this material are those of 
the authors and do not necessarily reflect the views of the National 
Science Foundation (NSF).

Copyright (C) 2002-3 University of Delaware

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

For more information, visit http://www.eecis.udel.edu/~decaf/, 
or contat by e-mail at decaf@eecis.udel.edu
*/

import java.util.*;
import java.text.*;

public class KQMLmsg {
	final boolean DEBUG = false;

    //TH new for counting msgs during failure testing
    static Integer lastExternalMsgNum = new Integer(0);
    int externalMsgNum = 0;
    static Integer count = new Integer(0);
    int myNum = 0;

    public void resetExternalMsgNum(){lastExternalMsgNum = new Integer(0);}
    public void resetCount(){count = new Integer(0);}

    //TH new for counting msgs during failure testing
    //Incremented in Agent.sendNow()
    public void incrementExtCount(){
	synchronized(lastExternalMsgNum){

	    externalMsgNum = lastExternalMsgNum.intValue() + 1;
	    lastExternalMsgNum = new Integer( externalMsgNum );

	}
    }

    //TH new for counting msgs during failure testing
    //Incremented in this.constructors
    public void incrementCount(){
	synchronized(count){

	    myNum = count.intValue() + 1;
	    count = new Integer( myNum );

	}
    }

    public int getExternalMsgNum(){ return lastExternalMsgNum.intValue(); }
    public int getCount(){ return count.intValue(); }

	/**
	* Empty Constructor
	*/
	public KQMLmsg() {
		if (DEBUG) System.out.print("Constructor1 - Started and");
		values[4] = "DECAF";
		if (DEBUG) System.out.println(" done.");

		incrementCount();
	}

	/**
	* Constructor with the KQML-fields as parameters.
	* @param performative Performative in KQML-message
	* @param sender Sender in KQML-message
	* @param receiver Receiver in KQML-message
	* @param ontology Ontology in KQML-message
	* @param language Language in KQML-message
	* @param inReplyTo In-reply-to in KQML-message
	* @param replyWith Reply-with in KQML-message
	* @param content Content in KQML-message
	*/

	public KQMLmsg (String performative, String sender, String receiver,
					String ontology, String language, String inReplyTo,
					String replyWith, String content) {

		values[0] = performative;
		values[1] = sender;
		values[2] = receiver;
		values[3] = ontology;
		values[4] = language;
		values[5] = inReplyTo;
		values[6] = replyWith;
		values[7] = content;
		if (DEBUG) System.out.println("Constructor2 - Made new KQML Messatge from parts.");

		incrementCount();

	}

	/**
	* Constructor that creates a new KQML-message instance based on
	* KQML-message string passed as parameter.
	* @param message A String representation of the KQML-message.
	*/
	public KQMLmsg (String message) {
		if (message.equals("")) {
			if (DEBUG) System.out.println("Constructor3 - Given blank message.  Nothing to do but return.");
			values[4] = "DECAF";
		} else {
			if (DEBUG) System.out.println("Constructor3 - Received new message.  Parsing ...\n" + message);
			parseKQMLString(message);
			//System.out.println("Constructor3 - Error trying to create message.\n" + ke.toString());
			if (DEBUG) System.out.println("Constructor3 - Message parsed and created.\n\n");
		}

		incrementCount();

	}

	/**
	* Access method to add a new field into the KQML-message.
	* @param field The field name.
	* @param value The value of the field.
	* @return 0 if all went well, -1 for unknown field, -2 for error 
	* in content field.
	*/ 
	public int addFieldValuePair (String field, String value) {
		if (DEBUG) System.out.println("AFVP1 - Attempting to add key: " + field + " and value: " + value);
		int response = 0;
		if (field.equals("performative")) {
			values[0] = value;
		} else if (field.equals("sender")) {
			values[1] = value;
		} else if (field.equals("receiver")) {
			values[2] = value;
		} else if (field.equals("ontology")) {
			values[3] = value;
		} else if (field.equals("language")) {
			values[4] = value;
		} else if (field.equals("in-reply-to")) {
			values[5] = value;
		} else if (field.equals("reply-with")) {
			values[6] = value;
		} else if (field.equals("delimiter")) {
			contentStartDelim = value;
			contentStopDelim = value;
		} else if (field.equals("delimiters")) {
			value = value.trim();
			contentStartDelim = value.substring(0, value.indexOf(" "));
			contentStopDelim = value.substring(value.indexOf(" "), value.length());
		} else if (field.equals("content")) {
			resetContentParser();
			values[7] = value;
			getContentKeyValue("");
		} else {
			response = -1;
			if (DEBUG) System.out.println("AFVP3 - Unknown field.");
		}
		if (DEBUG && response == 0) System.out.println("AFVP4 - Key and value added.");
		return response;
	}

	/**
	* This method is used for exchanging the sender and receiver field
	* values. It also adds the content field passed as parameter to the
	* KQML-message.  If the content param is null then the old value of
	* the content field will be kept.
	* @param content The new content to be added to the KQML-message.
	* @return A String representation of the new KQML-message.
	*/    
	public String constructSendString (String content) {
		String temp = values[1];
		values[1] = values[2];
		values[2] = temp;
		if (content != null) {
			addFieldValuePair("content", content);
		}
		return this.toString();
	}

	/**
	* Access method to get the a field specified as a parameter from the 
	* KQML-message.
	* @param field The name of the field.
	* @return Value of the field as a String.
	*/
	public String getValue (String field) {
		return getValue(field, new String(""), true);
	}

	/**
	* Access method to get the a field specified as a parameter from the 
	* KQML-message.
	* @param field The name of the field.
	* @param defaultIfEmpty The default value of the field.
	* @return Value of the field as a String.
	*/
	public String getValue (String field, String defaultIfEmpty) {
		return getValue(field, defaultIfEmpty, false);
	}

	/**
	* Provides functionality for the other two getValue functions.
	* @param field The name of field
	* @param defaultIfEmpty The default value of the given field.
	* @param returnNull Wether it should return an empty String or the default given.
	* @return The field value, or if the field is empty, the default if returnNull is false.
	*/
	public String getValue (String field, String defaultIfEmpty, boolean returnNull) {
		String response;
		if (field == "performative") {
			response = values[0];
		} else if (field == "sender") {
			response = values[1];
		} else if (field == "receiver") {
			response = values[2];
		} else if (field == "ontology") {
			response = values[3];
		} else if (field == "language") {
			response = values[4];
		} else if (field == "in-reply-to") {
			response = values[5];
		} else if (field == "reply-with") {
			response = values[6];
		} else if (field == "content") {
			response = values[7];
		} else {
			response = null;
		}

		// see what we got
		if (response == null) {
			if (returnNull) {
				return new String("");
			} else {
				return defaultIfEmpty;
			}
		} else {
			return response;
		}
	}

	/**
	* Access method to get the whole KQML-message as a Hashtable.
	* @return A Hashtable containing the field names as keys.
	*/
	public Hashtable getKQML() {
		Hashtable response = new Hashtable();
		response.put("performative", values[0]);
		response.put("sender", values[1]);
		response.put("receiver", values[2]);
		response.put("ontology", values[3]);
		response.put("language", values[4]);
		response.put("in-reply-to", values[5]);
		response.put("reply-with", values[6]);
		response.put("content", values[7]);
		return response;
	}

	/**
	* Gets the KQML Message Content as a hashtable.
	* @return The Hashtable with the key value pairs separated and the values 
	* stripped of their delimiters.
	*/
	public synchronized Hashtable getHashContent() {
		if (isFullyCached) {
			if (DEBUG) System.out.println("GHC1 - Contents already addressed, De-referencing.");
			return completeHashtable();
		} else {
			try {
				if (DEBUG) System.out.println("GHC2 - Parsing the content field for addresses.");
				getContentKeyValue("");
			} catch (KQMLException ke) {
				System.out.println("GHC3 - Error parsing content field for hash content.\n" + ke.toString());
			}
			if (DEBUG) System.out.println("GHC4 - Done parsing content field.  Dereferencing addresses.");
			return completeHashtable();
		}
	}  

	/**
	* Access method to get a field specified as a parameter from the 
	* content field of the KQML-message.
	* @param field The name of the field.
	* @return Value of the field as a String.
	*/
	public String getContentFieldValue (String field) {
		return getContentFieldValue(field, "", true);
	}

	/**
	* Access method to get a field specified as a parameter from the 
	* content field of the KQML-message.
	* @param field The name of the field.
	* @param defaultIfEmpty The default value of the field.
	* @return Value of the field as a String.
	*/
	public String getContentFieldValue (String field, String defaultIfEmpty) {
		return getContentFieldValue(field, defaultIfEmpty, false);
	}

	/**
	* Access method to get a field specified as a parameter from the 
	* content field of the KQML-message.
	* @param field The name of the field.
	* @param defaultIfEmpty The default value of the field.
	* @param returnNull Return the default given or an empty string.
	* @return Value of the field as a String.
	*/
	public String getContentFieldValue (String field, String defaultIfEmpty, boolean returnNull) {
		if (DEBUG) System.out.println("GCFV1 - Looking for content field value: " + field);
		String response = null;

		// Check in the hastable first
		int[] responseLocation = (int[])cachedContents.get(field);
		if (responseLocation == null) {
			if (DEBUG) System.out.println("GCFV2 - Content field has not been cached.  Looking in string.");
			responseLocation = getContentKeyValue(field);
		}
		// Process info
		if (responseLocation != null  && responseLocation[0] > -1 && responseLocation[1] > 0) {
			if (DEBUG) System.out.println("GCFV3 - Content field found.");
			try {
				response = values[7].substring(responseLocation[0], responseLocation[1]).trim();
			} catch (Exception e) {
				KQMLException ke = new KQMLException(e.toString(), values[7]);
				throw ke;
			}
		} else if (DEBUG) System.out.println("GCFV4 - Content field not found.");
		// See what we got from the content field
		if (response != null) {
			return response;
		} else {
			if (returnNull) {
				return new String("");
			} else {
				return defaultIfEmpty;
			}
		}
	}
    
    /**
     * Get the KQML message as a string.  Uses the toString() function.
     * Should probably just use the toString method directly.
     */
	public String getKQMLString() {
		return this.toString();
	} 

    /**
     * Returns the object as a valid KQML String.
     */
	public String toString() {
		StringBuffer message = new StringBuffer();
		message.append("(" + values[0] + 
						" :sender " + values[1] +
						" :receiver " + values[2]);
		if (values[3] != null && !values[3].equals("")) {
			message.append(" :ontology " + values[3]);
		}
		if (values[4] != null && !values[4].equals("")) {
			message.append(" :language " + values[4]);
		}
		if (values[5] != null && !values[5].equals("")) {
			message.append(" :in-reply-to " + values[5]);
		}
		if (values[6] != null && !values[6].equals("")) {
			message.append(" :reply-with " + values[6]);
		}
		if (!contentStartDelim.equals("(") || !contentStopDelim.equals(")")) {
			if (contentStartDelim.equals(contentStopDelim)) {
				message.append(" :delimiter " + contentStartDelim);
			} else {
				message.append(" :delimiters " + contentStartDelim + " " + contentStopDelim);
			}
		}
		if (values[7] != null && !values[7].equals("")) {
			message.append(" :content " + contentStartDelim  + values[7] + contentStopDelim);
		}
		message.append(")");
		if (DEBUG) System.out.println("I made myself into a message.  Lookie:\n" + message.toString());
		return message.toString();
	}

    /**
     * Top level function to parse the performative and all the key and
     * value pairs in a KQML message.  For info on formatting requirements
     * see the class description.
     * @param message The KQML message to process.
     */
	private void parseKQMLString(String message) throws KQMLException {
		if (DEBUG) System.out.println("PKS1 - Parsing new KQML Message");

		StringCharacterIterator sci = new StringCharacterIterator(message);
		parsePerformative(sci, message);
		if (DEBUG) System.out.println("PKS2 - Parsed the performative and got " + values[0]);

		char test = sci.current();
		boolean done = false;

		while (!done) {
			if (test == CharacterIterator.DONE) {
				throw new KQMLException("PKS3 - Message must contain ending delimiter", message);
			} else if (test == ')') {
				if (DEBUG) System.out.println("PKS4 - Found ending parenthesis of message.");
				done = true;
			} else if (!done) {
				parseNextKeyValue(sci, message);
				if (DEBUG) System.out.println("PKS5 - Parsed a key value");
				test = sci.current();
			}
		}
		if (DEBUG) System.out.println("PKS6 - Finished parsing new message.");
	}
    
	/**
	* Parses the start of a KQML message and the performative.  Stores the
	* performative in the message, i.e. no return value.
	* @param sci The current iterator in the message string.
	* @param message The message currently being parsed.
	*/
	private void parsePerformative(StringCharacterIterator sci, String message) throws KQMLException {
		int startIndex, stopIndex;
		char test = sci.current();
		while (Character.isWhitespace(test)) {
			test = sci.next();
			if (DEBUG) System.out.print("*");
		}
		if (test != '(') {
			throw new KQMLException("PP1 - Messages must start with an open paranethesis.", message);
		}
		test = sci.next();
	
		while (Character.isWhitespace(test)) {
			test = sci.next();
			if (DEBUG) System.out.print("*");
		}
		startIndex = sci.getIndex();
		while (!Character.isWhitespace(test) && test != CharacterIterator.DONE) {
			test = sci.next();
			if (DEBUG) System.out.print("#");
		}
		if (test == CharacterIterator.DONE) {
			throw new KQMLException("PP2 - Found end of message early", message);
		}
		stopIndex = sci.getIndex();
	
		addFieldValuePair("performative", message.substring(startIndex, stopIndex));
		if (DEBUG) System.out.println("PP3 - Found a performative in the message.");
		return;
	}

    /**
     * Reads from the curent point looking for a key and subsequently its value.
     * Delimniter values are handled internally, as is the value of the content 
     * field.  Delimiters are stored dirtectly into memory for later use.  The content
     * field is saved differently to avoid validation.  Any other key values are added 
     * to the message through addFieldValuePair(key, value).
     * @param sci The current iterator for the message string.
     * @param message The current message being processed.
     */
	private void parseNextKeyValue(StringCharacterIterator sci, String message) throws KQMLException {
		int keyStart, keyStop;
		int valueStart, valueStop;
		String key;
		char test = sci.current();

		while (Character.isWhitespace(test) && test != CharacterIterator.DONE) {
			test = sci.next();
			if (DEBUG) System.out.print("*");
		}
		if (test == ')') {
			if (DEBUG) System.out.println("PNKV1 - Wonder how we got here.  Oh well, I can survive.");
			return;
		} else if (test == CharacterIterator.DONE) {
			if (DEBUG) System.out.println("PNKV2 - Wonder how we got here.  Oh well, I can survive.");
			return;
		} else if (test != ':') {
			throw new KQMLException("PNKV3 - Found value without key at index " + 
									sci.getIndex() + " character '" + sci.current() + "'", message);
		}

		test = sci.next();
		keyStart = sci.getIndex();
		while (!Character.isWhitespace(test)) {
			test = sci.next();
			if (DEBUG) System.out.print("#");
		}
		keyStop = sci.getIndex();
		key = message.substring(keyStart, keyStop);
		if (DEBUG) System.out.println("PNKV4 - Found a key: " + key);

		while (Character.isWhitespace(test)) {
			test = sci.next();
			if (DEBUG) System.out.print("*");
		}
		valueStart = sci.getIndex();

		if (key.equals("delimiter")) {
			if (DEBUG) System.out.println("PNKV5 - Looking for one new delimiter");
			while (!Character.isWhitespace(test) && test != CharacterIterator.DONE) {
				test = sci.next();
				if (DEBUG) System.out.print("#");
			}
			if (test == CharacterIterator.DONE) {
				throw new KQMLException("PNKV6 - Found end of message early", message);
			}
			valueStop = sci.getIndex();

			contentStartDelim = message.substring(valueStart, valueStop);
			contentStopDelim = contentStartDelim;
			if (DEBUG) System.out.println("PNKV7 - Found new delimiter: " + contentStartDelim);
	    
		} else if (key.equals("delimiters")) {
			if (DEBUG) System.out.println("PNKV8 - Looking for new left delimiter");
			while (!Character.isWhitespace(test) && test != CharacterIterator.DONE) {
				test = sci.next();
				if (DEBUG) System.out.print("#");
			}
			if (test == CharacterIterator.DONE) {
				throw new KQMLException("PNKV9 - Found end of message early", message);
			}
			valueStop = sci.getIndex();

			contentStartDelim = message.substring(valueStart, valueStop);
			if (DEBUG) System.out.println("PNKV10 - Found new left delimiter: " + contentStartDelim);

			while (Character.isWhitespace(test)) {
				test = sci.next();
				if (DEBUG) System.out.print("*");
			}
			valueStart = sci.getIndex();
			if (DEBUG) System.out.println("PNKV11 - Looking for new right delimiter");
			while (!Character.isWhitespace(test) && test != CharacterIterator.DONE) {
				test = sci.next();
				if (DEBUG) System.out.print("#");
			}
			if (test == CharacterIterator.DONE) {
				throw new KQMLException("PNKV12 - Found end of message early", message);
			}
			valueStop = sci.getIndex();

			contentStopDelim = message.substring(valueStart, valueStop);
			if (DEBUG) System.out.println("PNKV13 - Found new right delimiter: " + contentStopDelim);
	    
		} else if (key.equals("content")) {
			int[] response = getKeysValue(message, sci.getIndex(), contentStartDelim, contentStopDelim);
			if (DEBUG) System.out.println("PNKV14 - Found the span of the content field." + 
											response[0] + " - " + response[1]);
			values[7] = message.substring(response[0], response[1]);
			if (DEBUG) System.out.println("PNKV15 - Content field set: \n" + values[7]);
			sci.setIndex(response[1] + contentStopDelim.length());
		} else {
			while (!Character.isWhitespace(test) && test != CharacterIterator.DONE && test != ')') {
				test = sci.next();
				if (DEBUG) System.out.print("#");
			}
			if (test == CharacterIterator.DONE) {
				throw new KQMLException("PNKV16 - Found end of message early", message);
			}
			valueStop = sci.getIndex();

			if (DEBUG) System.out.println("PNKV17 - Adding new values: " + key + " " + 
											message.substring(valueStart, valueStop) + "\n" + message);
			int accepted = addFieldValuePair(key, message.substring(valueStart, valueStop));
			if (accepted < 0) {
				if (DEBUG) System.out.println("PNKV18 - Error adding key and value");
				throw new KQMLException("PNKV19 - Found unknown key: " + key + " with value: " + 
										message.substring(valueStart, valueStop), message);
			}
		}
	}

    /**
     * Looks for a key's value in the content field.  Unlike in a message, a space 
     * is not neccesarily a delimiter.  This handles that.
     * @param keySought The key whose value is sought.  If passed an empty string then
     * the function will recursively parse to the end of the content string.  This will
     * cache all the data.
     * @return An integer array of length 2 which has the starting and ending index of
     * value in the content string.
     */
	private int[] getContentKeyValue(String keySought) throws KQMLException {
		if (DEBUG) System.out.println("GCKV1 - Looking for a key value.");
		if (values[7] == null) {
			if (DEBUG) System.out.println("There isn't a content field.  Why am I checking it?");
			isFullyCached = true;
			contentStringLocation = -1;
			return null;
		}
		if (contentStringLocation < 0 || contentStringLocation >= values[7].length()) {
			isFullyCached = true;
			if (DEBUG) System.out.println("GCKV2 - Oops, already at end. Guess it is not here.");
			return null;
		}
		StringCharacterIterator sci = new StringCharacterIterator(values[7], contentStringLocation);
		char test = sci.current();
		int keyStart, keyStop;

		while (Character.isWhitespace(test)) {
			test = sci.next();
			if (DEBUG) System.out.print("*");
		}
		if (test == CharacterIterator.DONE) {
			isFullyCached = true;
			contentStringLocation = -1;
			if (DEBUG) System.out.println("GCKV3 - Looked for it, but got to the end first.");
			return null;
		} else if (test != ':') {
			throw new KQMLException("GCKV4 - Found value without key at " + sci.getIndex() + " ", values[7]);
		}
	
		test = sci.next();
		keyStart = sci.getIndex();
		if (DEBUG) System.out.println("GCKV5 - Reading in key starting at " + keyStart +".");
		while( !Character.isWhitespace(test) && test != CharacterIterator.DONE) {
			test = sci.next();
			if (DEBUG) System.out.print("#");
		}
		if (test == CharacterIterator.DONE) {
			throw new KQMLException("GCKV6 - Reached an unexpected end.");
		}
		keyStop = sci.getIndex();
	
		while (Character.isWhitespace(test)) {
			test = sci.next();
			if (DEBUG) System.out.print("*");
		}

		String key = values[7].substring(keyStart, keyStop);
		if (DEBUG) System.out.println("GCKV7 - Found key in content: " + key);

		int[] response = getKeysValue(values[7], sci.getIndex(), currentStartDelim, currentStopDelim);
		if (response[2] == 0) {
			contentStringLocation  = response[1] + currentStopDelim.length();
			test = sci.setIndex(response[1] + currentStopDelim.length());
		} else {
			contentStringLocation  = response[1];
			test = sci.setIndex(response[1]);
		}

		if (DEBUG) System.out.println("Found a value for the key from " + response[0] + " to " + response[1]);

		if (key.equals("delimiter")) {
			currentStartDelim = values[7].substring(response[0], response[1]);
			currentStopDelim = currentStartDelim;

			if (DEBUG) System.out.println("GCKV8 - Found new delimiter: " + currentStartDelim);
			return getContentKeyValue(keySought);

		} else if (key.equals("delimiters")){
			if (DEBUG) System.out.println("GCKV9 - Looking for new delimiters.");
			test = sci.setIndex(response[0]);

			while (!Character.isWhitespace(test) && test != CharacterIterator.DONE) {
				test = sci.next();
				if (DEBUG) System.out.print("#");
			}
			if (test == CharacterIterator.DONE) {
				throw new KQMLException("GCKV10 - Found end of content early", values[7]);
			}
			currentStartDelim = values[7].substring(response[0], sci.getIndex());
			if (DEBUG) System.out.println("GCKV11 - Found new start delimiter: " + currentStartDelim);

			while (Character.isWhitespace(test)) {
				test = sci.next();
				if (DEBUG) System.out.print("*");
			}
			currentStopDelim = values[7].substring(sci.getIndex(), response[1]);
			if (DEBUG) System.out.println("GCKV12 - Found new end delimiter: " + currentStopDelim);

			contentStringLocation = response[1];
			return getContentKeyValue(keySought);
		} else {
			cachedContents.put(key, response);
			if (DEBUG) System.out.println("GCKV13 - Hashing key: " + key + " Address: " + response[0] + "-" + response[1]);
			if (keySought.equals("")) {
				if (DEBUG) System.out.println("GCKV14 - Parsing whole content field.  Looping back.");
				return getContentKeyValue(keySought);
			} else if (keySought != null && key.equals(keySought)) {
				if (DEBUG) System.out.println("GCKV15 - Found your value.  Thank you.");
				return response;
			} else {
				if (DEBUG) System.out.println("GCKV16 - Didn't find your key.  Looping Back");
				return getContentKeyValue(keySought);
			}
		}
	}

    /**
     * Utility function that can parse to the end of a value without having to analyze
     * the content of the value.
     * @param message The message to search
     * @param startingFrom The index into message to start from
     * @param startDelim The opening delimiter, can be of any length.
     * @param stopDelim the closing delimiter, can be the same as startDelim, and of an 
     * arbitrary length.
     * @return An int array containing the start and end index into the message for the
     * value.
     */
    private int[] getKeysValue(String message, int startingFrom, String startDelim, String stopDelim ) throws KQMLException {
	int counter = 0;
	int valueStop;
	int valueStart = startingFrom;

	if (! message.regionMatches(startingFrom, startDelim, 0, startDelim.length())) {
	    // never found start, content must not have specified delimiter
	    // so we get one word
	    StringCharacterIterator sci = new StringCharacterIterator(message, startingFrom);
	    char test = sci.current();
	    valueStart = startingFrom;
	    while (test != ':' && test != CharacterIterator.DONE) {
		test = sci.next();
		if (DEBUG) System.out.print("#");
	    }
	    test = sci.previous();
	    while (Character.isWhitespace(test)) {
		test = sci.previous();
		if (DEBUG) System.out.print("*");
	    }
	    test = sci.next();
	    valueStop = sci.getIndex();

	    int[] response = {valueStart, valueStop, -1};
	    return response;
	}
	
	valueStart += startDelim.length();
	valueStop = valueStart;
	counter++;

	if (DEBUG) System.out.println("GKV1 - Starting search for end delimiter. Counter: " + counter);

	boolean didSearch = true;
	if (stopDelim.equals(startDelim)) {
	    if (DEBUG) System.out.println("GKV2 - Never mind, only one delimiter to look for.");
	    valueStop = message.indexOf(stopDelim, valueStart);
	    counter--;
	    if (valueStop < 0) {
		throw new KQMLException("GKV3 - Found end before closing delimiter.", message);
	    }
		didSearch = false;
	}
	while (counter != 0) {
	    int nextOpen = message.indexOf(startDelim, valueStop);
	    int nextClose = message.indexOf(stopDelim, valueStop);

	    if ( nextOpen > -1 && (nextOpen < nextClose || nextClose < 0)) {
		counter++;
		if (DEBUG) System.out.println("GKV4 - Found another open.  New Depth: " + counter);
		valueStop = nextOpen + startDelim.length();
	    } else if ( nextClose > -1 && (nextClose < nextOpen || nextOpen < 0)) {
		counter--;
		if (DEBUG) System.out.println("GKV5 - Found another close. New Depth: " + counter);
		valueStop = nextClose + stopDelim.length();
	    } else {
		throw new KQMLException("GKV6 - Serious error parsing string for end.\n" +
					"Depth: " + counter + "nextOpen = " + nextOpen + 
					" nextClose = " + nextClose, message);
	    }
	}
	
	if (didSearch) {
		int[] response = {valueStart, valueStop - stopDelim.length(), 0};
		return response;
	} else {
		int[] response = {valueStart, valueStop, 0};
		valueStop += stopDelim.length();
		return response;
	}
    }

    /**
     * Utility function to take the cache hashtable and create a new
     * hashtable with fully expanded values, not addresses.  For speed,
     * programmers should consider just asking for the key's value.
     */
    private Hashtable completeHashtable() {
	if (DEBUG) System.out.println("CH1 - Attempting to de-reference address in the hashtable");
	Hashtable response = new Hashtable();
	for (Enumeration k = cachedContents.keys(); k.hasMoreElements();) {
	    Object key = k.nextElement();
	    int[] address = (int[])cachedContents.get(key);
	    response.put(key, values[7].substring(address[0], address[1]).trim());
	    if (DEBUG) System.out.println("CH2 - Found key " + key + " and its value in new hashtable.");
	}
	if (DEBUG) System.out.println("CH3 - New hashtable with full values created.");
	return response;
    }

    private void resetContentParser() {
		contentStringLocation = 0;
		isFullyCached = false;
		cachedContents = new Hashtable();
		currentStartDelim = "(";
		currentStopDelim = ")";
	}

    /**
     * 0 - performative
     * 1 - sender
     * 2 - receiver
     * 3 - ontology
     * 4 - language
     * 5 - in-reply-to
     * 6 - reply-with
     * 7 - content
     */
    private String[] values = new String[8];

    private String contentStartDelim = "(";
    private String contentStopDelim = ")";

    private Hashtable cachedContents = new Hashtable();
    /**
     * 0 is at the begining of the string, and -1 is at the end.
     * Any other numbers are points in between.
     */
    private int contentStringLocation = 0;
    private boolean isFullyCached = false;

    private String currentStartDelim = "(";
    private String currentStopDelim = ")";
}
