import java.util.*;

/**
 * Class for implementing Kruskal's algorithm - in this case for
 * generating a predetermined number of trees. Since K's starts with a
 * forest and slowly consolidates it using a minimum path, we can stop
 * the process at any point and have a number of clusters of (closely
 * related) trees.
 *
 * Take in a Vector of KruskalCells. Each KruskalCell is a unary
 * forest (what is? hashset). Make all possible KruskalCell pairs (one
 * order only) (what is? 2 row array). Store the distance between
 * members of each pair in a distance array, and sort, smallest
 * distance pairs first. Start consolidating forests by Union-ing the
 * forests associated with the first (smallest) distance pair;
 * continue until the desired number of forests is reached.
 *
 * Need to put Bids inside KCells. Have distance between
 * Bids. Distance will be cosine distance of vector of attribute values
 * projected on unit sphere.
 *
 * Kruskal cells should have an interface or superclass in common so
 * that I could use a shared method for determining distance between
 * objects; I could also use Method as a parameter, but the Reflection
 * instructions advised against that, and I don't want to do it until
 * I see the reasons, so this hasn't been done yet.
 */

public class Kruskal{

    //    private Vector weights;
    private int numberOfClusters;
    //    private HashSet trees;
    public Vector trees;
    private Vector inputs;
    private KruskalCell[][] pairs; //second index coincides with distances
    private double[] distances;
    private int numCombos;
    private final int MAX_INPUT_SIZE = 100;
    public static boolean DEBUG = false;
    public static boolean DEBUG2 = true;

    /**
     * numClusters is number of desired "forests" or clusters, at end
     * of running Kruskal's algorithm. V is a vector of objects that
     * contain whatever it is we are making clusters of.
     **/
    public Kruskal(int numClusters, Vector v){

        Object o;
	int size = v.size() < MAX_INPUT_SIZE ? v.size() : MAX_INPUT_SIZE; 
	inputs = new Vector(size);

	Enumeration ev = v.elements();
	if (v.size() < MAX_INPUT_SIZE){
	    while (ev.hasMoreElements()){
		o = ev.nextElement();
		if (o instanceof KruskalCell)
		    inputs.add(o);
		else
		    inputs.add( new KruskalCell(o) );
	    }
	}
	else {
	    Collections.shuffle(v);
	    for (int i = 0; i < MAX_INPUT_SIZE && ev.hasMoreElements(); i++){
		o = ev.nextElement();
		if (o instanceof KruskalCell)
		    inputs.add(o);
		else
		    inputs.add( new KruskalCell(o) );
	    }
	}

        numberOfClusters = numClusters;
        initialize();
    }/** constructor Kruskal(int numClusters, Vector v) **/

    public Kruskal(int numClusters, BidSet bs){

	int size = bs.size() < MAX_INPUT_SIZE ? bs.size() : MAX_INPUT_SIZE; 
	inputs = new Vector(size);
	//inputs = new Vector(bs.size());
	Bid b;
	Iterator bsIter = bs.iterator();
	//for (int i = 0; i < size && bsIter.hasNext(); i++){
	//    //while (bsIter.hasNext()){
	//    b = (Bid)bsIter.next();
	//    inputs.add( new KruskalCell(b) );
	//}

	if (bs.size() < MAX_INPUT_SIZE)
	    for (int i = 0; i < MAX_INPUT_SIZE && bsIter.hasNext(); i++){
		b = (Bid)bsIter.next();
		inputs.add( new KruskalCell(b) );
	    }
	else {
	    Vector v = new Vector(bs);
	    Collections.shuffle(v);
	    for (int i = 0; i < MAX_INPUT_SIZE; i++){
		b = (Bid)v.get(i);
		inputs.add( new KruskalCell(b) );
	    }
	}

        numberOfClusters = numClusters;

        initialize();
    }/**constructor Kruskal(int numClusters, BidSet bs)**/

    /**
     * Returns Vector of HashSets of original objects (instead of
     * KruskalCells, as from consolidate().
     **/
    public Vector getHashSetsOfObjects() throws Exception{
	Vector results = consolidate();
	Vector objs = new Vector(results.size());
	for (int i = 0; i < results.size(); i++){
	    HashSet tree = (HashSet)results.get(i);
	    HashSet newTree = new HashSet(tree.size());
	    Iterator treeIter = tree.iterator();
	    KruskalCell kc;
	    while( treeIter.hasNext() ){
		kc = (KruskalCell)treeIter.next();
		newTree.add( kc.getObject() );
	    }
	    objs.add(newTree);
	}
	return objs;
    }

    /**
     * Get number of trees down to number of desired clusters by
     * unioning trees. Don't run this until arrays have been sorted
     * by distance/weight.
     */
    public Vector consolidate() throws Exception{
        
        int nextPair = 0;
        HashSet firstTree = null, secondTree = null;
	Iterator treeIter;

        while (trees.size() > numberOfClusters){
            if (DEBUG) System.out.println("in Kruskal.consolidate: trees.size() : " + trees.size());
            //if members of pair aren't in same tree, union trees
            firstTree = getTree( pairs[0][nextPair] );
            secondTree = getTree( pairs[1][nextPair] );
            if ( ! firstTree.equals( secondTree )){
                //if (DEBUG) System.out.println("inside if, nextpair = " + nextPair);
                
                firstTree.addAll( secondTree );

                //if (DEBUG) System.out.println("contains: " + trees.contains(secondTree));
                
		treeIter = secondTree.iterator();
		while (treeIter.hasNext()){
		    setTree( treeIter.next(), firstTree );
		}

                if ( ! trees.remove( secondTree ))
                    throw new Exception("Kruskal remove tree failed for " + secondTree);

		//if (DEBUG) for (int i = 0; i < trees.size(); i++)
		//    System.out.println("trees1     " + trees.get(i));

            }
            nextPair++;
        }
        if (DEBUG) System.out.println("final number of trees is: " + trees.size());
	return trees;
    }

    private HashSet getTree( Object o ){
        if (o instanceof KruskalCell){
            if (DEBUG) System.out.println("inside getTree of " + o + " is " + ((KruskalCell)o).getTree());
            return ((KruskalCell)o).getTree();
        }
        else return null;
    }
    private void setTree( Object o, HashSet tree ){
        //if (o instanceof KruskalCell){
        if (DEBUG) System.out.println("inside setTree of " + o + " new tree is " + tree);
        ((KruskalCell)o).setTree(tree);
            //}
            //else return;
    }

    /**
     * Sets up arrays of pairs and distances. Sorts all three by distances.
     **/
    private void initialize(){
        
        // n! over k!(n - k)!
	if (DEBUG2) System.out.println("Entering K.initalize " + inputs.size());
	
	int n = inputs.size();
        numCombos =  n * (n - 1) / 2;
	    //factorial(inputs.size()) / (2 * factorial(inputs.size() - 2));
        if (DEBUG2) System.out.println("numCombos is " + numCombos);
                                                
        pairs = new KruskalCell[2][numCombos];
        distances = new double[numCombos];
        
        //make each KruskalCell a tree of one; trees is a forest of every tree
        trees = new Vector();//HashSet((int)(1.3 * inputs.size()));
        HashSet h;
        KruskalCell kc;
        for (int i = 0; i < inputs.size(); i++){
            h = new HashSet(5);
            kc =  (KruskalCell)inputs.get(i);
            //inputs.set(i, kc); is already a kc
            kc.setTree(h);
            h.add(kc);
            trees.add(h);
        }

	if (DEBUG) for (int i = 0; i < trees.size(); i++)
	    System.out.println("     " + trees.get(i));


        int i = 0;
        for(int j = 0; j < (inputs.size() - 1); j++) //init all pairs containing j
            for(int k = j + 1; k < inputs.size(); k++){ //init all pairs containing
                pairs[0][i] = (KruskalCell)inputs.get(j);
                pairs[1][i] = (KruskalCell)inputs.get(k);
                //System.out.println("injk: " + i + n + j + k);
                //System.out.println("next pair: " + pairs[0][i] + pairs[1][i]);
                //System.out.println("pairs[0]: " + Arrays.asList(pairs[0]));
                //System.out.println("pairs[1]: " + Arrays.asList(pairs[1]));
                distances[i] = calcDistance(pairs[0][i], pairs[1][i]);
                i++;
	    }
        if (DEBUG) System.out.println("pairs[0]: " + Arrays.asList(pairs[0]));
        if (DEBUG) System.out.println("pairs[1]: " + Arrays.asList(pairs[1]));
        if (DEBUG) {
	    System.out.print("distance: ");
            for (i=0; i<numCombos; i++)
                System.out.print(distances[i] + " ");
        }
        if (DEBUG) System.out.println();

        selectionSort(); //do mergesort instead

        if (DEBUG) System.out.println("pairs[0]: " + Arrays.asList(pairs[0]));
        if (DEBUG) System.out.println("pairs[1]: " + Arrays.asList(pairs[1]));
        if (DEBUG) {System.out.print("distance: ");
            for (i=0; i<numCombos; i++)
                System.out.print(distances[i] + " ");
        }
        if (DEBUG) System.out.println();
    }
    
    double calcDistance(KruskalCell a, KruskalCell b){


        if (a.getObject() instanceof Bid && b.getObject() instanceof Bid)
            return ((Bid)a.getObject()).cosineDistance((Bid)b.getObject());
        else if (a.getObject() instanceof Integer)
            return Math.abs(((Integer)((KruskalCell)a).getObject()).intValue() - 
                            ((Integer)((KruskalCell)b).getObject()).intValue());
        else return 0;  
        
    }

    private void selectionSort(){

        int size = distances.length;
        int minIndex;

        for (int start = 0; start < size; start++){
            minIndex = start;
            for (int i = start; i < size; i++)
                if (distances[minIndex] > distances[i])
                    minIndex = i;

            swap(start, minIndex);
        }
    }

    /**
     * Swaps elements of three parallel arrays.
     **/
    private void swap(int a, int b){

	KruskalCell tempKC;
        double tempD;

        tempKC = pairs[0][a];
        pairs[0][a] = pairs[0][b];
        pairs[0][b] = tempKC;

        tempKC = pairs[1][a];
        pairs[1][a] = pairs[1][b];
        pairs[1][b] = tempKC;

        tempD = distances[a];
        distances[a] = distances[b];
        distances[b] = tempD;

    }

    public int factorial(int n){
        int result = 1;
        while (n > 0)
            result *= n--;
        return result;
    }

    public static void main(String[] args){

        Vector v = new Vector();
        v.add(new KruskalCell(new Integer(43)));
        v.add(new KruskalCell(new Integer(44)));
        v.add(new KruskalCell(new Integer(45)));
        v.add(new KruskalCell(new Integer(45)));
        v.add(new KruskalCell(new Integer(46)));
        v.add(new KruskalCell(new Integer(46)));
        v.add(new KruskalCell(new Integer(13)));
        v.add(new KruskalCell(new Integer(13)));
        v.add(new KruskalCell(new Integer(13)));
        Kruskal k = new Kruskal(2, v);  

	Vector v1 = null;

        try {
            v1 = k.getHashSetsOfObjects();
        } catch (Exception e) {
	    e.printStackTrace();
            System.out.println(e);
        }

	if (v1 != null)
	    System.out.println(v1);

	
    }

}


//
//	Enumeration ev = v.elements(); //RANDOMIZE THIS when MAX_INPUT_SIZE is used FIX
//	for (int i = 0; i < size && ev.hasMoreElements(); i++){
//		 //while (ev.hasMoreElements()){
//	    o = ev.nextElement();
//	    if (o instanceof KruskalCell)
//		inputs.add(o);
//	    else
//		inputs.add( new KruskalCell(o) );
//	}
