// Evan Steinberger
// CISC 466
// Spring 2007

/**
 * MSNGrabber pulls data from MSN's financial web site(s).
 * 
 * @author Evan Steinberger
 */

public class MSNGrabber extends StockInfoGrabber
{
	
	// Contains Current Ratio and Debt/Equity Ratio
	private static final String address =
		"http://moneycentral.msn.com/investor/invsub/results/compare.asp?Page=FinancialCondition&Symbol=";

	// Contains P/E Ratio
	private static final String address2 =
		"http://moneycentral.msn.com/investor/invsub/results/compare.asp?Page=PriceRatios&Symbol=";

	// Contains info about Return on Equity
	private static final String address3 =
		"http://moneycentral.msn.com/investor/invsub/results/compare.asp?Page=InvestmentReturns&Symbol=";

	// To get Industry
	private static final String address4 =
		"http://moneycentral.msn.com/companyreport?Symbol=";

	private String crSearch = "Current Ratio";
	private String peSearch = "Current P/E Ratio";
	private String deTotalSearch = "Debt/Equity Ratio";
	private String reTTMSearch = "Return On Equity";
	private String endingDelim = "</td>";
	
	/**
	 * Constructor
	 * 
	 * @param stockSymbol	the stock ticker symbol
	 * @param at	the Access Threshold for pulling data
	 */
	public MSNGrabber(String stockSymbol, AccessThreshold at)
	{
		super.currentAccessThreshold = at;
		super.init(stockSymbol);
		this.validateSymbol();
	}

	/**
	 * Ensures that the selected symbol is a valid.
	 * 
	 */
	protected void validateSymbol()
	{
		super.downloadHTML(this.buildURL(address));
		if (super.pageHTML.startsWith(WebQuery.errorMessageStart))
		{
			super.mineErrorMsg = "Error: URL is not valid.";
			super.validSymbol = false;
		}
		else
		{
			if (super.pageHTML.contains("Please try again with a modified request"))
				super.validSymbol = false;
			else
				super.validSymbol = true;
		}
	}

	/**
	 * Builds a URL based on a prefix, stock symbol, and any
	 * other additional information.
	 * 
	 * @param s additional information placed into the URL
	 * 
	 */
	protected String buildURL(String s)
	{
		return s + super.stockSymbol;
	}

	/**
	 * Mines all relevant information.
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	public boolean mineAllInformation()
	{
		if (super.hasHistoricalContent)
			return true;
		if (!super.validSymbol)
		{
			if (super.mineErrorMsg == "")
				super.mineErrorMsg = "Error: Invalid Symbol Provided to Data Miner.";
			return false;
		}
		return this.mineCurrentRatio() && this.mineDeTotal() && this.minePeTTM() &&
			this.mineReTTM() && this.mineIndustry();
	}

	/**
	 * Mines the Current Ratio
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineCurrentRatio()
	{
		String temp = this.mineSingleData(this.crSearch + this.endingDelim);
		if (temp.equals("NA"))
			return true;
		try { super.currentRatio = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining Current Ratio: " + ex.toString();
			return false;
		}
		return true;
	}

	/**
	 * Mines Price/Earnings ratio
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean minePeTTM()
	{
		super.downloadHTML(this.buildURL(address2));
		String temp = this.mineSingleData(this.peSearch + this.endingDelim);
		if (temp.equals("NA"))
			return true;
		try { super.peTTM = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining P/E TTM: " + ex.toString();
			return false;
		}
		return true;
	}

	/**
	 * Mines the Debt/Equity Ratio.
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineDeTotal()
	{
		String temp = this.mineSingleData(this.deTotalSearch + this.endingDelim);
		if (temp.equals("NA"))
			return true;
		try { super.deTotal = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining DE Total: " + ex.toString();
			return false;
		}
		return true;
	}

	/**
	 * Mines Return on Equity
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineReTTM()
	{
		super.downloadHTML(this.buildURL(address3));
		String temp = this.mineSingleData(this.reTTMSearch + this.endingDelim);
		if (temp.equals("NA"))
			return true;
		try { super.reTTM = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining RE TTM: " + ex.toString();
			return false;
		}
		return true;
	}
	
	/**
	 * Mines the Industry.
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineIndustry()
	{
		super.downloadHTML(this.buildURL(address4));
		String temp = this.mineSingleData("Industry : ");
		if (temp == "")
		{
			super.mineErrorMsg = "Error: Problem mining Industry";
			return false;
		}
		else
		{
			super.industry = temp;
			return true;
		}
	}
	
	/**
	 * Mines piece of single data
	 * 
	 * @param searchFor	a string to search for
	 * @param beginParser a string where the parsing begins
	 * 
	 * @return	a string of parsed data
	 * 
	 */
	private String mineSingleData(String searchFor)
	{
		try
		{
			String temp = "";
			int start = super.pageHTML.indexOf(searchFor);
			int end;
			if (start < 0)
				return "";
			temp = pageHTML.substring(start);
			temp = temp.substring(searchFor.length(), temp.length()).trim();
			start = temp.indexOf(">");
			end = temp.indexOf("</");
			if (start < 0 || end < 0)
				return "";
			temp = temp.substring(start + 1, end);
			// New to the Yahoo Parser
			if (temp.endsWith("%"))
				temp = temp.substring(0, temp.length() -1);
			return temp;
		}
		catch (Exception ex) { return ""; }
	}
}