import java.util.*;
/**
 * MsgUtilities.java
 *
 *
 * Created: Tue Apr  6 13:10:48 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class MsgUtilities {
    public MsgUtilities() {
	
    } // MsgUtilities constructor
    

    //writes over existing content if it uses same key
    public static void copyGenContent(KQMLmsg from, KQMLmsg to, Agent local){
	Iterator copyFields = local.getMsgCopyFields();
	String value, field;
	while(copyFields.hasNext()){
	    field = (String)copyFields.next();
	    value = from.getContentFieldValue(field);
	    if ( ! value.equals("") )
		addContentFieldValuePair(to, field, value);
	}
    }

    //adds a content field and corresponding value to existing
    //content. Does not check for prior existence of field in msg.
    public static void addContentFieldValuePair(KQMLmsg k, String field, String value) {
	String currContent = k.getValue("content");
	String newStuff = " :" + field + " " + value + " ";
	if (field.equals("task"))
	    k.addFieldValuePair("content", newStuff.concat(currContent) );
	else
	    k.addFieldValuePair("content", currContent.concat(newStuff) );
    }

    /*
     * Get a constraint value from  within a content field
     */
    public static String getConstraintValue(KQMLmsg K, String cname){
	String cstring = K.getContentFieldValue("constraints");
	ConstraintSet cs = ConstraintSet.readCSFromString(cstring);
	return (String)cs.getValue(cname);
    }

    /*
     * Replace a constraint value within a content field
     */
    public static void replaceConstraint(KQMLmsg K, String cname, String value){

	String cstring = K.getContentFieldValue("constraints");
	ConstraintSet cs = ConstraintSet.readCSFromString(cstring);
	cs.put(cname, value);
	replaceContent(K, "constraints", cs.toString());
    }

    //replaces a content field value if one already exists. If not,
    //adds the field value pair with addContentFieldValuePair()
    public static void replaceContent(KQMLmsg K, String field, String value){
	String content = K.getValue("content");
	int start = content.indexOf(field);
	//find start of next pair

	if (start >= 0){
	    int end = content.indexOf(':' , start);
	    //System.out.println("content " + content);
	    //System.out.println("end " + end);
	 
	    StringBuffer newContent = new StringBuffer(content.substring(0, start));
	    newContent = newContent.append(field + " " + value + " ");
	    if (end != -1)//no more pairs starting with ':'
		newContent = newContent.append(content.substring(end));

	    K.addFieldValuePair("content", newContent.toString());
	}
	else addContentFieldValuePair(K, field, value);
    }

    public static boolean removeContentField(KQMLmsg K, String field){
	String content = K.getValue("content");
	int start = content.indexOf(field);
	//find start of next pair

	if (start >= 0){
	    int end = content.indexOf(':' , start);
	    //System.out.println("content " + content);
	    //System.out.println("end " + end);
	 
	    StringBuffer newContent = new StringBuffer(content.substring(0, start - 1)); //-1 for colon

	    if (end != -1)//no more pairs starting with ':'
		newContent = newContent.append(content.substring(end));
	    //	    else newContent = newContent.append(")"); no, KQMLmsg does this
	    
	    //System.out.println("new: " + newContent.toString());
	    
	    K.addFieldValuePair("content", newContent.toString());
	    return true;
	}
	else return false;
    }


    /*
     * Removes any more than one copy of the word "empty" from a
     * StringBuffer
     */
    public static String removeDuplEmpties(StringBuffer s){
	int loc = s.indexOf("empty");
	while(s != null && loc > -1){
	    s.delete(loc, loc + 5);
	    loc = s.indexOf("empty");
	}

	String temp = s.toString();
	temp = temp.trim();

	if (temp.length() == 0)
	    return "empty";	
	else
	    return temp;
    }


    public static void main(String[] args){
	
	   KQMLmsg K = new KQMLmsg();
	   KQMLmsg K2 = new KQMLmsg();
	   K2.addFieldValuePair("sender", "spam" );
	   K2.addFieldValuePair("performative", "achieve");
	   K2.addFieldValuePair("reply-with", "z");
	   K2.addFieldValuePair("receiver", "LocgetName" );
	   K2.addFieldValuePair("ontology", "PL" );
	   K2.addFieldValuePair("content", ":task acceptRFB :in msg");
	   System.out.println(K2.toString());
	   MsgUtilities.addContentFieldValuePair(K2, "task", "span");
	   //System.out.println(K2.toString() + Util.addTimeout(10));
	   MsgUtilities.replaceContent(K2, "on", "re");	  
	   //MsgUtilities.replaceContent(K2, "on", "rap");
	   System.out.println(K2.toString());
	   MsgUtilities.removeContentField(K2, "on");
	   System.out.println(K2.toString());

	   //MsgUtilities.addContentFieldValuePair(K2, "spic", "span");


    }

} // MsgUtilities
