import java.io.*;
import java.util.*;
/**
 * NumericAgent.java
 *
 *
 * Created: Wed Jan 14 17:10:58 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public abstract class NumericAgent extends GenAgent{

    static String dataFileName = "stockDataUM05"; //default; gets changed by update msg from PA (PL_Start)
    String originator; //sender of execution request
    String originatorReplyWith;
    static final boolean DEBUG = false;

    static Hashtable data = null; //copied to Local.userHash in start

    public NumericAgent() {
	System.out.println("NumericAgent zero constructor");
	
	if (data == null){
	    data = new Hashtable();

	    try {
		readDataFileToHT(dataFileName, data);
	    } catch (IOException ioe) {
		System.out.println("NumericAgent exception reading data; " + ioe);
	    }
	}

    } // NumericAgent constructor
    
    static public void setDataFileName(String s){dataFileName = new String(s);}

    void start(Agent Local, String ds, String executeLspFile, String executeTaskName){
       System.out.println("**********starting NumericAgent: " + Local.getName() + " **************");

       Local.userHash.putAll(data);
       Local.setIsNumeric();
       Local.setMyTaskObject(this);

       Local.setExecuteTaskName(executeTaskName);
       Local.setDsType(ds);
       System.out.println("original tasks: " + Local.taskTemplates);
       Local.addTaskTemplate("acceptRFB.lsp");
       Local.addTaskTemplate("acceptEnv.lsp");
       Local.addTaskTemplate("BidAgent.lsp");
       Local.addTaskTemplate("failingUp.lsp");
       Local.addTaskTemplate("update.lsp");
       Local.addTaskTemplate(executeLspFile);
       //System.out.println("all tasks: " + Local.taskTemplates);

       Local.setLocalBids(new BidSet(Local.getName()));
       return;
    }

    /**********************************************************************/
    /**********************************************************************/
    /* The following functions are for operating MADSUM by itself,
     * before the Informed Analysis module is added. These methods
     * (and data files) are maintained for historic reasons. No new
     * data is written to these files. All data histories are now
     * contained in the IA module.
     */
    /**********************************************************************/
    /**********************************************************************/

    /**
     * Removes old data previously read in from file so that new data can be used for testing.
     **/
    public void eraseOldData(Agent Local){
	if (data != null){
	    Hashtable h = Local.userHash;
	    Enumeration e = data.keys();
	    while(e.hasMoreElements()){
		Object o = e.nextElement();
		if ( null != Local.userHash.remove( o ))
		    if (DEBUG) System.out.println("Successfully removed: " + o);
		
	    }
	    data.clear();
	}
    }

    public void replaceData(String filename, Agent Local){
	eraseOldData(Local);
	try{
	    readDataFileToHT(filename, data);
	} catch (IOException e) {
	    System.out.println("in replaceData(): " + e);
	    e.printStackTrace();
	}
	Local.userHash.putAll(data);
	System.out.println("All data replaced." + Local.userHash);
    }

    /*
     * Like above, but doesn't erase all existing data first
     */
    public void replaceSomeData(String filename, Agent Local){
	
	try{
	    readDataFileToHT(filename, data);
	} catch (IOException e) {
	    System.out.println("in replaceSomeData(): " + e);
	    e.printStackTrace();
	}
	Local.userHash.putAll(data);
	System.out.println("Some data replaced." + Local.userHash);
    }

    /*
     * Like above, but takes a String and doesn't erase all existing data first
     */
    public void replaceSomeDataString(String dataString, Agent Local){
	
	try{
	    readDataStringToHT(dataString, data);//data is the name of DECAF hashtable
	} catch (IOException e) {
	    System.out.println("in replaceSomeDataString(): " + e);
	    e.printStackTrace();
	}
	Local.userHash.putAll(data);
	System.out.println("Some data replaced from String." + Local.userHash);
    }

    static public String get(Hashtable ht, String key, String cat){

		String k = key + cat;
		if (ht.containsKey(k))
			return (String)ht.get(k);
		else return null;
	
	
    }

    /*for getting raw data from triplets in a file. example:
     *
     *ibm cr 2
     *ibm industry tech
     *
     *Stores data in hashtable arg.
     */
    static void readDataFileToHT(String filename, Hashtable h)
    throws IOException
    {

	String key = null, cat = null, value = null;
	int count = 0;
	char c;

	BufferedReader in = new BufferedReader(new FileReader(filename));
	StreamTokenizer sT = new StreamTokenizer(in);
	sT.wordChars(63,63); //allows '?' as part of token
	sT.wordChars(95,95); //allows '_' as part of token
	sT.wordChars('.','.'); //allows '.' as part of token
	sT.wordChars(':',':'); //allows '.' as part of token
	sT.slashStarComments(true);

	try{
	    while (sT.nextToken() != StreamTokenizer.TT_EOF){

		switch(sT.ttype){
		case StreamTokenizer.TT_WORD:
		    if (count == 0){
			key = sT.sval;
			//System.out.println("0count,key,cat,value: " + count + key + cat + value);
			count++;
			
		    }
		    else if (count == 1){
			cat = sT.sval;
			//System.out.println("1count,key,cat,value: " + count + key + cat + value);
			count++;
		    }
		    else if (count == 2){ 
			value = sT.sval;
			System.out.println("storing key,cat,value: " + count + key + cat + value);
			h.put(key + cat, value);
			count = 0;
		    }
		    else throw new IOException("DataSet count should be 0, 1 or 2, not: " + 
					       count + "; key is: " + key + " " + sT.sval);
		    break;
		case StreamTokenizer.TT_EOL:
		    break;
		    //need to take in value, plus string after? no, must be numeric for now
		case StreamTokenizer.TT_NUMBER:
		    if (count == 2) {
			value = Double.toString(sT.nval);
			System.out.println("storing count,key,cat,value: " + count + key + cat + value);
			h.put(key + cat, value);
			count = 0;		
		    }	
		    else throw new IOException("DataSet count should be 2, not: " + count + 
					       "; value is: " + value);
		    break;
		default: //single character in ttype
		    c = (char)sT.ttype;
		    value = String.valueOf(c);
		}
	    
		if (count > 2)
		    throw new IOException("DataSet count should be < 3; key is: " + 
					  key + "; value is: " + value);
	    }
	}
	catch (IOException e){
	    e.printStackTrace();
	    throw e;
	}
	return;
    }

//string looks like: "ibm industry tech" no quotes
    static void readDataStringToHT(String dataString, Hashtable h)
    throws IOException
    {

	String key = null, cat = null, value = null;
	int count = 0;
	char c;

	StringReader in = new StringReader(dataString);
	StreamTokenizer sT = new StreamTokenizer(in);
	sT.wordChars(63,63); //allows '?' as part of token
	sT.wordChars(95,95); //allows '_' as part of token
	sT.wordChars('.','.'); //allows '.' as part of token
	sT.wordChars(':',':'); //allows '.' as part of token
	sT.slashStarComments(true);

	try{
	    while (sT.nextToken() != StreamTokenizer.TT_EOF){

		switch(sT.ttype){
		case StreamTokenizer.TT_WORD:
		    if (count == 0){
			key = sT.sval;
			//System.out.println("0count,key,cat,value: " + count + key + cat + value);
			count++;
			
		    }
		    else if (count == 1){
			cat = sT.sval;
			//System.out.println("1count,key,cat,value: " + count + key + cat + value);
			count++;
		    }
		    else if (count == 2){ 
			value = sT.sval;
			System.out.println("storing key,cat,value: " + count + key + cat + value);
			h.put(key + cat, value);
			count = 0;
		    }
		    else throw new IOException("DataSet count should be 0, 1 or 2, not: " + 
					       count + "; key is: " + key + " " + sT.sval);
		    break;
		case StreamTokenizer.TT_EOL:
		    break;
		    //need to take in value, plus string after? no, must be numeric for now
		case StreamTokenizer.TT_NUMBER:
		    if (count == 2) {
			value = Double.toString(sT.nval);
			System.out.println("storing count,key,cat,value: " + count + key + cat + value);
			h.put(key + cat, value);
			count = 0;		
		    }	
		    else throw new IOException("DataSet count should be 2, not: " + count + 
					       "; value is: " + value);
		    break;
		default: //single character in ttype
		    c = (char)sT.ttype;
		    value = String.valueOf(c);
		}
	    
		if (count > 2)
		    throw new IOException("DataSet count should be < 3; key is: " + 
					  key + "; value is: " + value);
	    }
	}
	catch (IOException e){
	    e.printStackTrace();
	    throw e;
	}
	return;
    }

} // NumericAgent
