import java.util.*;
/**
 * PL_DistrEnv.java
 *
 *
 * Created: Sun Apr 18 13:30:16 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 *
 * When an Envelope is received from above, separate into Envelopes to
 * distribute to children of this agent. Distribution of Env is
 * according to bids submitted by children, which are components of
 * the this agent's big selected by this agent's parent.
 *
 * To enable failure recovery, save a mapping of Env to child bids, so
 * can determine, when a child result is returned, which Env it was
 * supposed to address (because once failures are allowed, then can
 * have multiple outstanding envs and multiple outstanding bids from
 * single child.)
 *
 */
public class PL_DistrEnv extends GenTask {

    static String loc = " in PL_DistrEnv ";
    static String em = " Error in PL_DistrEnv ";
    static final boolean DEBUG = false;
    static final boolean DEBUGenv = true;
    KQMLmsg startMsg; //message that started this task
    int numMsgsCollected = 0;//see collectMsgs
    static double minEnvelopePercent = .99;//max acceptable diff between required U and actual 

    public PL_DistrEnv() {
	
	System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
	System.out.println("% PL_DistrEnv() Zero constructor % <==============");
	System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
    } // PL_DistrEnv constructor
    
    /**
     * Take an envelope and a chosen bid, find the agents that
     * contributed to the bid, and send a portion of the envelope to
     * each contributor, i.e. distribute the envelope among the agents
     * whose bids are components of the chosen bid.  
     *
     * Is there any
     * reason not to have a new "envelopes" hashtable each time
     * through? Why does it need to be static, stored in Agent?
     **/
    public ProvisionCell distributeEnv(LinkedListQ Plist, Agent Local) {

    ht = Local.userHash; //This init is for GenTask, not to use directly

       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));//(Plist, "in5"));//
       startMsg = K;
       boolean isRecovery; //tells whether msg is from PL_Count.manageFailure()
       job = getMsgJob(K);
       if (DEBUG) System.out.println("INSIDE distributeEnv for job: " + job + " " + K);

       Local.checkMsgTime(K, Local.dataFile1, job);

       //inits
       bidsMade = (Hashtable)localGet("bidsMade", job);
       bidsAccepted = (Hashtable)localGet("bidsAccepted", job);
       
       String myEndTime = K.getContentFieldValue("endtime");
       String envName = K.getContentFieldValue("newEnv");
       String incomingParentEnv = K.getContentFieldValue("parentEnv");

       if (DEBUGenv) System.out.println(Local.getName() + " distributeEnv env: " + envName + " " + localGet(envName, job));
       if (DEBUGenv) System.out.println(Local.getName() + " distributeEnv parentEnv: " + incomingParentEnv);

       isRecovery = ((String)K.getContentFieldValue("recovery")).equals("yes");

       StringBuffer msgs = new StringBuffer();

	Bid best;
	Envelope env = null;
	if (envName.equals("empty"))
	    best = null;
	else{
	    env = (Envelope)localGet(envName, job);
	    best = env.getBid();
	}
	ProvisionCell P1 = null;
	Vector rejects = new Vector(5);
	rejects.addAll(Local.getChildrenVector());//start with all, then remove successful bidders

	if (best != null){

	    //should be set false in PL_BidAgent (s/b acceptEnv?),
	    //and set true in Count.myTimeout or collectResults
	    Boolean finished = (Boolean)localGet("finished", job); 
	    synchronized(finished){
		if (finished.booleanValue() == true){
		    localPut("finished", job, new Boolean(false)); 
		    System.out.println("DistEnv setting finished to false");
		}
		addAcceptedBid(best, env);
		BidSet sourceBids = getBidComponents(best);
		if (DEBUG )System.out.println(Local.getName() + " distributeEnv sourceBids: " + sourceBids);
	    
		Hashtable envelopes = new Hashtable();//FIX? (Hashtable)localGet("envelopes", job);
		if (sourceBids != null){
		    Iterator i = sourceBids.iterator();
		    if (DEBUG)
			while(i.hasNext())
			    System.out.println("source bidder: " + ((Bid)i.next()).getBidder());
		    //fill Hashtable envelopes with bid/env pairs for children.
		    distributeEnvelopeAmongBids(envelopes, sourceBids, env, Local);
		}

		if (Local.isLocalBid(best)){
		    //check to see if this is a bid made by this agent
		    if (DEBUG) System.out.println("THIS IS A LOCAL BID (execute): " + best + " " + env);
		    KQMLmsg temp = msgToExecute(Local, "results", env);
		    MsgUtilities.addContentFieldValuePair(temp, "parentEnv", Local.getName() + envName);
		    P1 = new ProvisionCell( temp.toString(), "OK");
		}
		else {//best is not local, represents basket of bids

		    Enumeration bids = envelopes.keys(); //get bidder from source bids
		    Bid b = null;

		    HashSet bidsSentForExecution = (HashSet)localGet("bidsSentForExecution", job);
		    int bidsSentThisTime = envelopes.size();

		    Collection envBidMapping = new Vector();

		    while (bids.hasMoreElements()){
			b = (Bid)bids.nextElement();
			//System.out.println("********************waitingForSenders adding: " + b.getBidder());	

			//if recovery, do not add source of failure again (never got removed)

			if (isRecovery && b.getBidder().equals(K.getContentFieldValue("failedSender")))
			    System.out.println("********************waitingForSenders not adding recovery for: " + 
					       b.getBidder());//do not add, was never removed
			else{
			    addWaitingForSender(b.getBidder(), job, Local.getName(), " distrEnv1");
			    System.out.println("********************waitingForSenders1 adding: " + 
					       b.getBidder());	
			}
			if (rejects.contains(b.getBidder())) rejects.removeElement(b.getBidder());
		   
			//if bidsSentForExecution contains limitedEq b, continue
			if (alreadySentForExec(b)) {
			    if (DEBUG) System.out.println("not resending bid to child for execution: " + b);
			    if (! removeWaitingForSender(b.getBidder(), job, Local.getName() , " distrEnv-1 already sent"))
				System.out.println("Warning: unable to remove sender " + b.getBidder());
			    bidsSentThisTime--;
			}
			//if a local bid
			else if (Local.getName().equals(b.getBidder())){
			    //then bid came from this agent and can be executed? yes, must also be local
			    Envelope e = (Envelope)envelopes.get(b);
			    localPut(e.getName(), job, e);
			    bidsSentForExecution.add(b);
			    System.out.println(Local.getName() + " Sending for local execution: " + e.toString() + 
					       " with timeout: " + Local.lowerMsgTimeout(myEndTime));
			    KQMLmsg temp = msgToExecute(Local, "results", e);
			    MsgUtilities.addContentFieldValuePair(temp, "parentEnv", Local.getName() + envName);
			    msgs.append(temp.toString());
			    msgs.append(Util.addTimeout(Local.lowerMsgTimeout(myEndTime)));
			}
			//Non-local, send b out for execution
			else {
			    bidsSentForExecution.add(b);
			    envBidMapping.add(b);

			    KQMLmsg K2 = new KQMLmsg();
			    K2.addFieldValuePair("sender", Local.getName() );
			    K2.addFieldValuePair("performative", "achieve");
			    //K2.addFieldValuePair("in-reply-to", 
			    //		 (String)ht.get(b.getBidder() + "EnvReplyWith", job));//see countMsgs in acceptRFB
			    //(String)localGet(b.getBidder() + "EnvReplyWith", job));//see countMsgs in acceptRFB
			    K2.addFieldValuePair("reply-with", "results");//see countMsgs2 in PL_Count
			    K2.addFieldValuePair("receiver", b.getBidder());
			    K2.addFieldValuePair("ontology", "PL" );
			    K2.addFieldValuePair("content", ":task ExecAgent :msg2 spoo :envelope " + 
						 ((Envelope)envelopes.get(b)).toString());
			    MsgUtilities.copyGenContent(startMsg, K2, Local);
			    MsgUtilities.replaceContent(K2, "parentEnv", Local.getName() + envName); //after copy is important
			    if (DEBUG) System.out.println(Local.getName() + " distributeEnv writing: " + K2);
			    if (DEBUG) System.out.println("name of outgoing envelope is: " + 
							  ((Envelope)envelopes.get(b)).getName() + " with timeout: " + Local.lowerMsgTimeout(myEndTime));
			    msgs.append(K2.getKQMLString());
			    msgs.append(Util.addTimeout(Local.lowerMsgTimeout(myEndTime)));
			}
		    }//end while for making envelopes for child agents

		    //envBidMapping is collection of bids assoc with an envelope -- needed for failure processing
		    localPut(Local.getName() + envName + "BidMap", job, envBidMapping); //store env to child bid mapping
		    if (DEBUGenv) System.out.println("DistrEnv stored " + Local.getName() + envName + "BidMap" + "=" + envBidMapping);
		    
		    //store this agent Env name w parent env name
		    localPut(Local.getName() + envName + "ParentEnvMap", job, incomingParentEnv); 
		    if (DEBUGenv) System.out.println("DistrEnv stored " + Local.getName() + envName + "ParentEnvMap" + "=" + incomingParentEnv);

		    System.out.println("DistrEnv: sent " + bidsSentThisTime + " msgs: " + msgs.toString());
		    writeAgentMsgToFile("\nDistrEnv: sent " + bidsSentThisTime + " msgs: " + msgs.toString(), startMsg.getContentFieldValue("endtime"), Local);
		    if (msgs.length() == 0) P1 =  new ProvisionCell("empty","OK");
		    else P1 =  new ProvisionCell(msgs.toString(),"OK");
		}
	    }//synch on finished
	}
	
	rejectLetters(Local, rejects);

	if (P1 == null) P1 =  new ProvisionCell("empty","OK");
	if ( ! waitingForSenders(job) ){
	    System.out.println("DistrEnv waitingForSenders is empty");
	    KQMLmsg K3 = new KQMLmsg();
	    K3.addFieldValuePair("sender", Local.getName() );
	    K3.addFieldValuePair("performative", "achieve");
	    K3.addFieldValuePair("receiver", Local.getName() );
	    K3.addFieldValuePair("ontology", "PL" );
	    K3.addFieldValuePair("content", ":task Count :replies x2" +
	    			 " :result " + "empty" +
	    			 " :status " + "DistrEnvDone");
	    MsgUtilities.copyGenContent(startMsg, K3, Local);
	    MsgUtilities.replaceContent(K3, "parentEnv", Local.getName() + envName); //after copy is important
	    System.out.println("DistrEnv.distributeEnv: " + K3);
	    Local.send(K3);
	}
	if (DEBUG) System.out.println("distributeEnv finished.");
	
	return P1;
       
    }// distributeEnv(LinkedListQ Plist, Agent Local)

    /** 
     * Need this action so can require system timeout msgs. Send each
     * message out to PL_Count as it comes in. 
     **/
    public ProvisionCell collectMsgs(LinkedListQ Plist, Agent Local) {
	
       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
       System.out.println("inside DistrEnv.collectMsgs: " + K);

       Local.checkMsgTime(K, Local.dataFile1, job);

       //FIX send a msg to this task after finished is true so that it terminates
       Boolean finished = (Boolean)localGet("finished", job); //should be set in PL_acceptEnv
       if (finished.booleanValue() == true){
	   System.out.println("inside DistEnv.collectMsgs: finished == true");
	   return new ProvisionCell("postFinished", "jobDone");
       }

       KQMLmsg K2 = new KQMLmsg();
       K2.addFieldValuePair("sender", Local.getName());//K.getValue("sender")
       K2.addFieldValuePair("performative", "achieve");
       K2.addFieldValuePair("receiver", Local.getName());//K.getValue("receiver")
       K2.addFieldValuePair("ontology", "PL" );
       
       if (K.getValue("performative").equals("timeout")){
	   K2.addFieldValuePair("content", ":task Count :replies x :status TIMEOUT" + 
				" :oldPerformative " + K.getValue("performative") +
				" :oldSender " + K.getValue("receiver") +
				" :oldTask " + K.getContentFieldValue("task"));
       }
       else if (K.getContentFieldValue("result").equals("")){
	   System.out.println("collectMsgs ignoring msg with no result: " + K);
	   return new ProvisionCell("ignoring", "counting");
       }
       else{
	   String status = K.getContentFieldValue("status");
	   if (status.equals("")) status = "fine";

	   K2.addFieldValuePair("content", ":task Count :replies x" +
				" :result " + K.getContentFieldValue("result") +
				" :status " + status +
				" :oldSender " + K.getValue("sender"));
       }
       MsgUtilities.copyGenContent(K, K2, Local);
       //MsgUtilities.addContentFieldValuePair(K2, "parentEnv", K.getContentFieldValue("parentEnv")); //not here, came from kid
       System.out.println("DistrEnv.collectMsgs sending: " + K2);
   
       /******** Send all msgs to PL_Count ***********/
       Local.send(K2);
       /**********************************************/

       //int numExpected = waitingForSenders.size();
       //numMsgsCollected++;
       //if (numMsgsCollected >= numExpected){
       //	   KQMLmsg K3 = new KQMLmsg();
       //	   K3.addFieldValuePair("sender", Local.getName());//K.getValue("sender")
       //	   K3.addFieldValuePair("performative", "achieve");
       //	   K3.addFieldValuePair("receiver", Local.getName());//K.getValue("receiver")
       //	   K3.addFieldValuePair("ontology", "PL" );
       //	   K3.addFieldValuePair("content", ":task Count :replies x :status collectDONE");
       //	   return new ProvisionCell(K3.toString(), "OK");
       //}
       //System.out.println("collectMsgs counting:" + numMsgsCollected);
       
       return new ProvisionCell("NOTDONE", "counting");

    }//collectMsgs(LinkedListQ Plist, Agent Local) 

    //fills a Hashtable with bid+env pairs for children. Each env
    //contains portion of downward allocated U based on bid's
    //contribution to the total U of the bidset. If that is higher
    //than bid's expected U, then the expected U is used for the
    //envelope (otherwise huge numbers of failures happen and slow
    //things down; the higher U expected by the agent above is
    //probably due to synergies of the combined bids and non-linear
    //aspects of the U fcn.)
    private void distributeEnvelopeAmongBids(Hashtable envelopes, BidSet sourceBids, Envelope env, Agent Local)
    {
	synchronized(envelopes){
	synchronized(env){

	    if (DEBUG) System.out.println("env: " + env);
	    if (DEBUG) System.out.println("env.containsKey('utility_min'): " + env.containsKey("utility_min"));
	    if (DEBUG) System.out.println("sourceBids: " + sourceBids);
	
	    Bid temp = sourceBids.makeOneBidFromBidSet(Local.getName() + "-temp");
	    double totalU = temp.getUtility(Local.getUtilityFcn()).doubleValue();
	    if (Double.parseDouble((String)env.get("utility_min")) > totalU)
		System.out.println(Local.getName() + "Warning: expected utility does not cover envelope. Env: " 
				   + env);
	
	    Iterator i = sourceBids.iterator();
	    double envUMin, tempUMin, childBidU;
	    Bid b;
	    Envelope e;
	    String childEnvUMin;
	    while (i.hasNext()){
		b = (Bid)i.next();
		e = new Envelope(b);
		if (DEBUG) System.out.println("distribute's incoming env name is " 
					      + env.getName() + " new env name is " + e.getName());
		if (env.containsKey("utility_min")){
		    envUMin = Double.parseDouble((String)env.get("utility_min"));
		    childBidU = b.getUtility(Local.getUtilityFcn()).doubleValue();
		    tempUMin = minEnvelopePercent * envUMin * (childBidU / totalU);
		    childEnvUMin = Double.toString( Math.min(childBidU, tempUMin) );
		    e.put("utility_min",  childEnvUMin);
		    envelopes.put(b, e);
		}
		else
		    System.out.println(Local.getName()+ em + "distributeEnvelopeAmongBids: env does not contain utility_min " + env);
	    
	    }
	    System.out.println("distributeEnvelopeAmongBids envelopes: " + envelopes);
	    return;
	}}//synch env, envelopes
    }//distributeEnvelopeAmongBids(BidSet sourceBids, Envelope env, Agent Local){


    private boolean alreadySentForExec(Bid b){
	HashSet bidsSentForExecution = (HashSet)localGet("bidsSentForExecution", job);
	if (DEBUG) System.out.println("DistrEnv.alreadySentForExec bidsSentForExecution: " + bidsSentForExecution);
	//System.out.println("DistrEnv.alreadySentForExec job: " + job);

	synchronized(bidsSentForExecution){
	    Iterator sent = bidsSentForExecution.iterator();
	    while (sent.hasNext()){
		if (b.limitedEquals((Bid)sent.next())){
		    System.out.println("bid already sent below for execution: " + b);
		    return true;
		}
	    }
	}
	return false;
    }

    /**
     * Get the name of the DECAF Task which has execution code so we
     * can send a message to it telling it to execute. Compose the
     * message including the name of the Envelope so that the
     * execution task can look it up and see what it is supposed to
     * do. For executing local bids only
     **/
    private KQMLmsg msgToExecute(Agent Local, String replyWith, Envelope env){

		   KQMLmsg K = new KQMLmsg();
		   K.addFieldValuePair("sender", Local.getName() );
		   K.addFieldValuePair("performative", "achieve");
		   K.addFieldValuePair("reply-with", replyWith);
		   K.addFieldValuePair("receiver", Local.getName());
		   K.addFieldValuePair("ontology", "PL" );
		   K.addFieldValuePair("content", ":task " + Local.getExecuteTaskName() + " :env1 " + env.getName()); //"Execute"
		   MsgUtilities.copyGenContent(startMsg, K, Local);
		   if (DEBUG) System.out.println(Local.getName() + " PL_DistrEnv.sendToExecute writing: " + K);
		   return K; //.getKQMLString();
    }//msgToExecute(Agent Local, String replyWith, Envelope env)

    /** not currently needed, since no task waits on envelope**/
    private void rejectLetters(Agent Local, Vector rejects){

	StringBuffer msgs = new StringBuffer();
	KQMLmsg K;
	String agent;

	Iterator i = rejects.iterator();
	while (i.hasNext()){
	    agent = (String)i.next();
	    K = new KQMLmsg();
	    K.addFieldValuePair("sender", Local.getName() );
	    K.addFieldValuePair("performative", "achieve");
	    //K.addFieldValuePair("in-reply-to", (String)localGet(agent + "EnvReplyWith", job));//see countMsgs in acceptRFB
	    K.addFieldValuePair("receiver", agent);
	    K.addFieldValuePair("ontology", "PL" );
	    K.addFieldValuePair("content", ":task ExecAgent :msg2 derl :envelope empty ");
	    MsgUtilities.copyGenContent(startMsg, K, Local);
	    System.out.println(Local.getName() + " rejectLetters sending : " + K);
	    Local.sendNow(K);
	}
    }//rejectLetters(Agent Local, Vector rejects)



} // PL_DistrEnv

    
