import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Numeric agent has own execute. See PL_CurrentRatio.java
 *
 */

public class PL_ExecuteCR extends GenTask
{
    //Hashtable ht;
    String taskReplyWith;
    static boolean DEBUG = false;
    static final String loc = "in PL_ExecuteCR ";
    static final String em = "ERROR in PL_ExecuteCR: ";
    static final String wm = "Warning in PL_ExecuteCR: ";

    static final boolean RANDOM = false; //set true for random generation

    String parentEnv = "unset"; //parent uses this to identify when result is returned

    KQMLmsg startMsg = null;

   public PL_ExecuteCR()
   {
	  System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
      System.out.println("% PL_ExecuteCR Zero constructor % <==============");
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
   }

   public ProvisionCell receiveEnv1(LinkedListQ Plist, Agent Local)
   {
       if (DEBUG) System.out.println(Local.getName() + " inside PL_ExecuteCR.receiveEnv1");

       ht = Local.userHash;

       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
       System.out.println(Local.getName() + " PL_ExecuteCR.receiveEnv1 receiving: " + K);
       startMsg = new KQMLmsg(K.toString());
             
       job = getMsgJob(K);

       bidsMade = (Hashtable)localGet("bidsMade", job);
       bidsAccepted = (Hashtable)localGet("bidsAccepted", job);       

       parentEnv = K.getContentFieldValue("parentEnv");
       String envelopeName = null;

       if (K.getContentFieldValue("opportunity").equals("yes")){
	   Envelope newEnv = Envelope.readEnvFromString(K.getContentFieldValue("env1"));
	   if (newEnv == null) 
	       envelopeName = "empty";
	   else {
	       envelopeName = newEnv.getName();
	       localPut(envelopeName, job, newEnv);
	   }
       }
       else
	   envelopeName = (String)K.getContentFieldValue("env1");

       if (DEBUG){
	   Envelope env = (Envelope)localGet(envelopeName, job);
	   System.out.println(Local.getName() + "PL_ExecuteCR.receiveEnv1 Envelope: " + env);
       }

       taskReplyWith = (String)K.getValue("reply-with");
       
       return new ProvisionCell(envelopeName, "OK");  
   }

   public ProvisionCell execute(LinkedListQ Plist, Agent Local)
    {
	   try{
	   	   Thread.sleep(Local.getSimulatedDelay());
       }
       catch(Exception e){
	   System.out.println("Delay problem in " + Local.getName() + ": " + e);
       }

	ProvisionCell P1 = null;
	String envName = new String(Util.getValue(Plist, "env3"));

	if (DEBUG) System.out.println(Local.getName() + loc + " entering execute");
	KQMLmsg K = new KQMLmsg();
	K.addFieldValuePair("sender", Local.getName() );
	K.addFieldValuePair("performative", "achieve");
	
	if ( ! isOpportunity(startMsg) ) 
	    K.addFieldValuePair("in-reply-to", taskReplyWith);
	
	K.addFieldValuePair("receiver", Local.getName() );
	K.addFieldValuePair("ontology", "PL" );

	Envelope env = null;
	if ( envName.equals("empty") ) env = null;
	
	String result, outcome;
	env = (Envelope)localGet(envName, job);
	if (env != null && ! env.isEmpty()) {
	    
	    System.out.println(Local.getName() + loc + " entering execute: " + env);
	    System.out.println(Local.getName() + loc + " bids made: " + getBidsMade());

	    //change DS; how to get DataSet? or make new? make new.
	    //get bid from env
	    //update DSr, cost in Bid;
	    //make new DataSet with bid, symbol, source

	    // (:result {bid={cost_exp=1, RV=[[] 0], DSv_exp=[[0 0 0][0 0 0]], length_exp=10, DSr_exp=[[10 10 1][0 0 0]], DSg_exp=[[0 0 0][0 0 0]], U=null, bidder=CR}, symbol=, source=CR}))

	    Bid b = (Bid)env.getBid().clone();
	    double price = Double.parseDouble((String)b.get("cost_exp"));
	    String symbol = b.getSymbol();
	    double currentRatio, industryCurrentRatio = -1;
	    if (RANDOM)
		currentRatio = Math.random() * 6;
	    else {
		//if (DEBUG && symbol == null) System.out.println(wm + " symbol is null for bid " + b);
		//if (DEBUG) System.out.println(loc + " symbol is " + symbol);

		// Informed Analysis Module Add-on
		String blindprofile, informedprofile, grabberprofile, accessthresh;
		StockInfoGrabber.AccessThreshold at;
		try
		{
			blindprofile = NumericAgent.get(ht, "cr", "blindprofile");
			informedprofile = NumericAgent.get(ht, "cr", "informedprofile");
			grabberprofile = NumericAgent.get(ht, "cr", "grabberprofile");
			accessthresh = NumericAgent.get(ht, "cr", "accessthresh");

			at = StockInfoGrabber.parseAccessThreshold(accessthresh);

			if (blindprofile==null || informedprofile==null || grabberprofile==null || accessthresh==null)
			{
				blindprofile = informedprofile = grabberprofile = "standard";
				at = StockInfoGrabber.AccessThreshold.Day;
			}
		}
		catch (Exception ex)
		{
			blindprofile = informedprofile = grabberprofile = "standard";
			at = StockInfoGrabber.AccessThreshold.Day;
		}
		InformedAnalyzer IA = new InformedAnalyzer(symbol,
			blindprofile,
			grabberprofile,
			informedprofile,
			at, true, 1.5, true, "cr");
		IA.analyze();

		String newResult = symbol + " cr " + IA.cr; 
		try { NumericAgent.readDataStringToHT(newResult, ht); }
		catch (Exception e){ System.out.println("WRONG"); }

		currentRatio = Double.parseDouble(NumericAgent.get(ht, symbol, "cr"));

		if (DEBUG) System.out.println(loc + " currentRatio for " + symbol + " is " + currentRatio);
	    }

	    if (price >= 3){
		if (DEBUG) System.out.println(loc + " symbol is " + symbol);
		String industry = NumericAgent.get(ht, symbol, "industry");
		industryCurrentRatio = Double.parseDouble(NumericAgent.get(ht, industry, "cr"));
		if (DEBUG) System.out.println(loc + " industryCurrentRatio for " + 
					      industry + " is " + industryCurrentRatio);
	    }

	    //Strength newDS = new Strength( calcDSfromCR(currentRatio) );
	    //if (DEBUG) System.out.println(loc + " new DSr is " + newDS);
	    //b.put("DS" + DSType.RISK + "_exp", newDS);
	    DataSet d = new DataSet(Local.getName());
	    d.put("symbol", symbol);
	    d.put("bid", b);
	    d.put("currentRatio", Double.toString(currentRatio));
	    d.put("industryCurrentRatio", Double.toString(industryCurrentRatio));

	    //K.addFieldValuePair("content", ":result " + d.toString());
	    //System.out.println(Local.getName() + " execute writing: " + K);

	    result = d.toString();
	    outcome = "OK";
	    //P1 = new ProvisionCell(K.getKQMLString(),"OK");
	}
	
	else {
	    result = "empty";
	    outcome = "FAIL";
	    //K.addFieldValuePair("content", ":result empty");
	    //return new ProvisionCell(Local.getName() + ".execute","FAIL");
	}
	
	System.out.println(loc + " checking opp: " + isOpportunity(startMsg));
	
	if ( isOpportunity(startMsg) ){
	    MsgUtilities.copyGenContent(startMsg, K, Local);
	    MsgUtilities.addContentFieldValuePair(K, "result", result);
	    MsgUtilities.addContentFieldValuePair(K, "opportunity", "first");
	    MsgUtilities.addContentFieldValuePair(K, "oppEnv", env.toString());
	    MsgUtilities.addContentFieldValuePair(K, "bidder", Local.getName());
	    MsgUtilities.addContentFieldValuePair(K, "replies", "opp12");
	    MsgUtilities.addContentFieldValuePair(K, "task", "Count");
	    P1 = new ProvisionCell(K.toString(), outcome);
	}
	    //P1 = new ProvisionCell(oppMsgToProcess(result, "empty", env, startMsg, Local).toString(), outcome);
	else { 
	    K.addFieldValuePair("content", ":result " + result);
	    MsgUtilities.copyGenContent(startMsg, K, Local);
	    P1 = new ProvisionCell(K.toString(), outcome);
	}
	System.out.println(loc + " returning " + P1.getData());
	return P1; 
  }
   
    /*
     * calculate DS from current ratio
     */
    public static double calcDSfromCR(double cr){
	double zeroVal = cr - 1.5;
	if (zeroVal >= 0)
	    return Math.pow(zeroVal, 2) + 2 * zeroVal;
	else
	    return  zeroVal - Math.pow(zeroVal, 2) - 1.5;
    }
 
    public static void main(String[] args){
	System.out.println(PL_ExecuteCR.calcDSfromCR(0.25));
	System.out.println(PL_ExecuteCR.calcDSfromCR(0.5));
	System.out.println(PL_ExecuteCR.calcDSfromCR(0.75));
	System.out.println(PL_ExecuteCR.calcDSfromCR(1));
	System.out.println(PL_ExecuteCR.calcDSfromCR(1.25));
	System.out.println(PL_ExecuteCR.calcDSfromCR(1.5));
	System.out.println(PL_ExecuteCR.calcDSfromCR(1.2));
	//System.out.println(PL_ExecuteCR.calcDSfromCR(3));
	
    }

	
}//PL_ExecuteCR
