import java.io.*;
import java.net.*;
import java.util.*;

/**
 *
 * This file is for processing the results from the children of the
 * DebtRisk agent, CurrentRisk and DebtEquity.
 *
 */

public class PL_ExecuteDE extends GenTask
{
    String taskReplyWith;
    static boolean DEBUG = true;
    static final String loc = "in PL_ExecuteDE ";
    static final String em = "ERROR in PL_ExecuteDE: ";
    static final String wm = "Warning in PL_ExecuteDE: ";

    KQMLmsg startMsg = null;
    static final boolean RANDOM = false; //set true for random generation
    String parentEnv = "unset"; //parent uses this to identify when result is returned

   public PL_ExecuteDE()
   {
	  System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
      System.out.println("% PL_ExecuteDE Zero constructor % <==============");
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");

   }

   public ProvisionCell receiveEnv1(LinkedListQ Plist, Agent Local)
   {
       System.out.println(Local.getName() + " inside PL_ExecuteDE.receiveEnv1");

       ht = Local.userHash;

       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
       startMsg = K;
       System.out.println(Local.getName() + " PL_ExecuteDE.receiveEnv1 receiving: " + K);
       
       
       job = getMsgJob(K);

       bidsMade = (Hashtable)localGet("bidsMade", job);
       bidsAccepted = (Hashtable)localGet("bidsAccepted", job);       


       parentEnv = K.getContentFieldValue("parentEnv");
       String envelopeName = (String)K.getContentFieldValue("env1");
       Envelope env = (Envelope)localGet(envelopeName, job);

       taskReplyWith = (String)K.getValue("reply-with");
       
       System.out.println(Local.getName() + "PL_ExecuteDE.receiveEnv1 Envelope: " + env);

       return new ProvisionCell(env.getName(), "OK");  

   }

   public ProvisionCell execute(LinkedListQ Plist, Agent Local)
    {
       try{
	   	   Thread.sleep(Local.getSimulatedDelay());
       }
       catch(Exception e){
	   System.out.println("Delay problem in " + Local.getName() + ": " + e);
       }

	ProvisionCell P1 = null;
	String envName = new String(Util.getValue(Plist, "env3"));
	Envelope env = (Envelope)localGet(envName, job);
	System.out.println(Local.getName() + loc + " entering execute: " + env);
	System.out.println(Local.getName() + loc + " bids made: " + getBidsMade());
	
	double price, dsNum, debtEquityRatio, industryDebtEquityRatio = -1;

	Bid b = (Bid)env.getBid().clone();
	price = Double.parseDouble((String)b.get("cost_exp"));
	String symbol = b.getSymbol();
	if (DEBUG && symbol == null) System.out.println(wm + " symbol is null for bid " + b);

	if (RANDOM){ //range is zero to 2
	    double temp = Math.random();
	    if (temp < 0.75)
		debtEquityRatio = temp;
	    else 
		debtEquityRatio = 0.75 + (1.25 * Math.random());
	}
	else {
	    if (DEBUG) System.out.println(loc + " symbol is " + symbol);

		// Informed Analysis Module Add-on
		String blindprofile, informedprofile, grabberprofile, accessthresh;
		StockInfoGrabber.AccessThreshold at;
		try{
			blindprofile = NumericAgent.get(ht, "de", "blindprofile");
			informedprofile = NumericAgent.get(ht, "de", "informedprofile");
			grabberprofile = NumericAgent.get(ht, "de", "grabberprofile");
			accessthresh = NumericAgent.get(ht, "de", "accessthresh");

			at = StockInfoGrabber.parseAccessThreshold(accessthresh);

			if (blindprofile==null || informedprofile==null || grabberprofile==null || accessthresh==null)
			{
				blindprofile = informedprofile = grabberprofile = "standard";
				at = StockInfoGrabber.AccessThreshold.Day;
			}
		}
		catch (Exception ex)
		{
			blindprofile = informedprofile = grabberprofile = "standard";
			at = StockInfoGrabber.AccessThreshold.Day;
		}
		InformedAnalyzer IA = new InformedAnalyzer(symbol,
				blindprofile,
				grabberprofile,
				informedprofile,
				at, false, 0, false, "de");
		IA.analyze();

		String newResult = symbol + " de " + IA.de;
		try { NumericAgent.readDataStringToHT(newResult, ht); }
		catch (Exception e) { System.out.println("WRONG"); }
		
	    debtEquityRatio = Double.parseDouble(NumericAgent.get(ht, symbol, "de"));
	    if (DEBUG) System.out.println(loc + " debtEquityRatio for " + 
					  symbol + " is " + debtEquityRatio);
	}

	if (price >= 3){
	    if (DEBUG) System.out.println(loc + " symbol is " + symbol);
	    String industry = NumericAgent.get(ht, symbol, "industry");
	    industryDebtEquityRatio = Double.parseDouble(NumericAgent.get(ht, industry, "de"));
	    if (DEBUG) System.out.println(loc + " industryDebtEquityRatio for " + 
					  industry + " is " + industryDebtEquityRatio);
	}

	DataSet d = new DataSet(Local.getName());
	d.put("symbol", symbol);
	d.put("bid", b);
	d.put("debtEquityRatio", Double.toString(debtEquityRatio));
	d.put("industryDebtEquityRatio", Double.toString(industryDebtEquityRatio));

	KQMLmsg K = new KQMLmsg();
	K.addFieldValuePair("sender", Local.getName() );
	K.addFieldValuePair("performative", "tell");
	K.addFieldValuePair("in-reply-to", taskReplyWith);
	K.addFieldValuePair("receiver", Local.getName() );
	K.addFieldValuePair("ontology", "PL" );
	K.addFieldValuePair("content", ":result " + d.toString()); // + Local.generate(env)
	MsgUtilities.copyGenContent(startMsg, K, Local);
	System.out.println(Local.getName() + " execute writing: " + K);

	if(true){
	    P1 = new ProvisionCell(K.getKQMLString(),"OK");
	}

	else if(true) {
	    return new ProvisionCell(Local.getName() + ".execute","FAIL");
	}
      return P1; 
  }
   
    ///*
    // * calculate DS from current ratio
    // */
    //public static double calcDSfromCR(double cr){
    //	double zeroVal = cr - 1.5;
    //	if (zeroVal >= 0)
    //	    return Math.pow(zeroVal, 2) + 2 * zeroVal;
    //	else
    //	    return  zeroVal - Math.pow(zeroVal, 2) - 1.5;
    //}
 
    //public static void main(String[] args){}

}//PL_ExecuteDE
