import java.io.*;
import java.net.*;
import java.util.*;

/**
 * The numeric Goal children that do not (so far) interact with
 * external agents will perform some calculation to determine actual
 * data before bidding. The justification for this is that they are
 * internal, and their design is such (so far) that their results will
 * not vary based on the envelope, except within degrees that can be
 * represented by the bids. They will submit a subset of possible
 * bids, perhaps two.
 *
 */

public class PL_ExecutePA extends GenTask
{
    String taskReplyWith;
    static boolean DEBUG = true;
    static final String loc = "in PL_ExecutePA ";
    static final String em = "ERROR in PL_ExecutePA: ";
    static final String wm = "Warning in PL_ExecutePA: ";

    KQMLmsg startMsg = null;
    static final boolean RANDOM = false; //set true for random generation
    String parentEnv = "unset"; //parent uses this to identify when result is returned

   public PL_ExecutePA()
   {
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
      System.out.println("% PL_ExecutePA Zero constructor % <==============");
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");

   }

   public ProvisionCell receiveEnv1(LinkedListQ Plist, Agent Local)
   {
       System.out.println(Local.getName() + " inside PL_ExecutePA.receiveEnv1");

       ht = Local.userHash;

       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
       startMsg = K;
       System.out.println(Local.getName() + " PL_ExecutePA.receiveEnv1 receiving: " + K);
       
       
       job = getMsgJob(K);

       bidsMade = (Hashtable)localGet("bidsMade", job);
       bidsAccepted = (Hashtable)localGet("bidsAccepted", job);       

       parentEnv = K.getContentFieldValue("parentEnv");
 
       String envelopeName = (String)K.getContentFieldValue("env1");
       Envelope env = (Envelope)localGet(envelopeName, job);

       taskReplyWith = (String)K.getValue("reply-with");
       
       System.out.println(Local.getName() + "PL_ExecutePA.receiveEnv1 Envelope: " + env);

       return new ProvisionCell(env.getName(), "OK");  

   }

   public ProvisionCell execute(LinkedListQ Plist, Agent Local)
    {
       try{
	   	   Thread.sleep(Local.getSimulatedDelay());
       }
       catch(Exception e){
	   System.out.println("Delay problem in " + Local.getName() + ": " + e);
       }

	ProvisionCell P1 = null;
	String envName = new String(Util.getValue(Plist, "env3"));
	Envelope env = (Envelope)localGet(envName, job);
	System.out.println(Local.getName() + loc + " entering execute: " + env);
	System.out.println(Local.getName() + loc + " bids made: " + getBidsMade());
	
	int length; 
	double ds;
	String phrase, investmentType;
	double allocGoal, allocCurrent, allocWithPurchase, investmentTypeBalance, totalBalance,
	    annualConsumption, share_price, num_shares, newAllocStatus;

	//KQMLmsg startMsg = (KQMLmsg)localGet("startMsg", job);
	String price = startMsg.getContentFieldValue("price");
	share_price = Double.parseDouble(price);
	String quantity = startMsg.getContentFieldValue("quantity");
	num_shares = Double.parseDouble(quantity);

	Bid b = (Bid)env.getBid().clone();
	//	price = Double.parseDouble((String)b.get("cost_exp"));
	String symbol = b.getSymbol();
	if (DEBUG && symbol == null) System.out.println(wm + " symbol is null for bid " + b);

	investmentType = (String)NumericAgent.get(ht, symbol, "type");
	investmentTypeBalance = Double.parseDouble(NumericAgent.get(ht, investmentType, "balance"));
	totalBalance = Double.parseDouble(NumericAgent.get(ht, "total", "balance"));
	allocGoal = Double.parseDouble(NumericAgent.get(ht, investmentType, "goal_percent"));
	//yearsToRetire = Double.parseDouble(NumericAgent.get(ht, "year", "to_retire"));
	annualConsumption = Double.parseDouble(NumericAgent.get(ht, "year", "consumption"));
	allocCurrent = 100 * investmentTypeBalance / totalBalance;
	allocWithPurchase  = 100 * (investmentTypeBalance + share_price * num_shares)/ totalBalance;

	if (true || DEBUG){
	    System.out.println("investmentType " + investmentType);
	    System.out.println("investmentTypeBalance" + investmentTypeBalance);
	    System.out.println("totalBalance " + totalBalance);
	    System.out.println("allocGoal " + allocGoal);
	    System.out.println("allocCurrent " + allocCurrent);
	    System.out.println("allocWithPurchase " + allocWithPurchase );
	}

	//order is important	
	if(allocCurrent > allocGoal){ 
	    phrase = "you have already exceeded your stated allocation goal for " + 
		investmentType +" "; 
	    length = 9; ds = -2.5;
	}
	else if (allocWithPurchase < (0.9 * allocGoal)){//allocWithPurchase > allocCurrent && 
	    phrase = "this investment is consistent with your stated allocation goal for " + 
		investmentType +" "; 
	    length = 10; ds = 0.1;
	}
	else if (allocWithPurchase < allocGoal){//allocWithPurchase > allocCurrent && 
	    phrase = "this investment puts you close to your stated allocation goal for " + 
		investmentType +" "; 
	    length = 12; ds = 0.5;
	}
	else if (allocWithPurchase < (1.03 * allocGoal)) { //allocWithPurchase > allocCurrent && 
	    phrase = "this investment achieves your stated allocation goal for " + 
		investmentType +" "; 
	    length = 8; ds = 2;
	}
	else if (allocWithPurchase < (1.10 * allocGoal)) { //allocWithPurchase > allocCurrent && 
	    phrase = "this investment exceeds your stated allocation goal for " + 
		investmentType +" "; 
	    //	    phrase = "this investment does not have a significant impact on your allocation goals;";
	    length = 8; ds = -1.5;
	}
	else if (allocWithPurchase > allocGoal) { //allocWithPurchase > allocCurrent && 
	    phrase = "this investment substantially exceeds your stated allocation goal for " + 
		investmentType +" "; 
	    //	    phrase = "this investment does not have a significant impact on your allocation goals;";
	    length = 9; ds = -2;
	}
	else  {//never reached
	    phrase = em + " awp " + allocWithPurchase + " ac " + allocCurrent + 
		" ag " + allocGoal + " it " + investmentType; 
	    length = 5; ds = -12;
	}
	RSTTree resultTree = null;
	//divide by two for free internal data agent
	Strength newDS = new Strength( ds/2 );
	if (DEBUG) System.out.println(loc + " new DSg is " + newDS);
	b.put("DS" + DSType.GOAL + "_exp", newDS);
	b.put("length_exp", Double.toString(length));
	b.put("topic", DSType.GOAL );
	if (ds > 0)
	    resultTree = RSTTree.GoalDo(phrase, DSType.GOAL, b);
	else
	    resultTree = RSTTree.GoalDoNot(phrase, DSType.GOAL, b);

	DataSet d = new DataSet(Local.getName());
	d.put("symbol", symbol);
	d.put("bid", b);
	d.put("tree", resultTree.getString());

	KQMLmsg K = new KQMLmsg();
	K.addFieldValuePair("sender", Local.getName() );
	K.addFieldValuePair("performative", "tell");
	K.addFieldValuePair("in-reply-to", taskReplyWith);
	K.addFieldValuePair("receiver", Local.getName() );
	K.addFieldValuePair("ontology", "PL" );
	K.addFieldValuePair("content", ":result " + d.toString()); // + Local.generate(env)

	MsgUtilities.copyGenContent(startMsg, K, Local);

	System.out.println(Local.getName() + " execute writing: " + K);

	if(true){
	    P1 = new ProvisionCell(K.getKQMLString(),"OK");
	}

	else if(true) {
	    return new ProvisionCell(Local.getName() + ".execute","FAIL");
	}
      return P1; 
  }
   

}//PL_ExecutePA
