import java.io.*;
import java.net.*;
import java.util.*;

/**
 *
 * This file is for processing the results from the children of the
 * PebtRisk agent, CurrentRisk and PriceEarnings.
 *
 */

public class PL_ExecutePE extends GenTask
{
    String taskReplyWith;
    static boolean DEBUG = true;
    static final String loc = "in PL_ExecutePE ";
    static final String em = "ERROR in PL_ExecutePE: ";
    static final String wm = "Warning in PL_ExecutePE: ";

    static final boolean RANDOM = false; //set true for random generation
    String parentEnv = "unset"; //parent uses this to identify when result is returned

    KQMLmsg startMsg = null;

   public PL_ExecutePE()
   {
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
      System.out.println("% PL_ExecutePE Zero constructor % <==============");
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");

   }

   public ProvisionCell receiveEnv1(LinkedListQ Plist, Agent Local)
   {
       System.out.println(Local.getName() + " inside PL_ExecutePE.receiveEnv1");

       ht = Local.userHash;

       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
       startMsg = K;
       System.out.println(Local.getName() + " PL_ExecutePE.receiveEnv1 receiving: " + K);
       
       
       job = getMsgJob(K);

       bidsMade = (Hashtable)localGet("bidsMade", job);
       bidsAccepted = (Hashtable)localGet("bidsAccepted", job);       

       parentEnv = K.getContentFieldValue("parentEnv");
 
       String envelopeName = (String)K.getContentFieldValue("env1");
       Envelope env = (Envelope)localGet(envelopeName, job);

       taskReplyWith = (String)K.getValue("reply-with");
       
       System.out.println(Local.getName() + "PL_ExecutePE.receiveEnv1 Envelope: " + env);

       return new ProvisionCell(env.getName(), "OK");  

   }

   public ProvisionCell execute(LinkedListQ Plist, Agent Local)
    {

       try{
	   Thread.sleep(Local.getSimulatedDelay());
       }
       catch(Exception e){
	   System.out.println("Delay problem in " + Local.getName() + ": " + e);
       }

	ProvisionCell P1 = null;
	String envName = new String(Util.getValue(Plist, "env3"));
	Envelope env = (Envelope)localGet(envName, job);
	System.out.println(Local.getName() + loc + " entering execute: " + env);
	System.out.println(Local.getName() + loc + " bids made: " + getBidsMade());
	
	double price, dsNum, priceEarningsRatio, industryPriceEarningsRatio = -1;
	String industry = null;
	Bid b = (Bid)env.getBid().clone();
	price = Double.parseDouble((String)b.get("cost_exp"));
	String symbol = b.getSymbol();
	if (DEBUG && symbol == null) System.out.println(wm + " symbol is null for bid " + b);

	if (RANDOM){ //range is 1 to 100
	    priceEarningsRatio = 100 * Math.random();
	}
	else {
	    if (DEBUG) System.out.println(loc + " symbol is " + symbol);

		
		// Informed Analysis Module Add-on
		String blindprofile, informedprofile, grabberprofile, accessthresh;
		StockInfoGrabber.AccessThreshold at;
		try
		{
			blindprofile = NumericAgent.get(ht, "pe", "blindprofile");
			informedprofile = NumericAgent.get(ht, "pe", "informedprofile");
			grabberprofile = NumericAgent.get(ht, "pe", "grabberprofile");
			accessthresh = NumericAgent.get(ht, "pe", "accessthresh");

			at = StockInfoGrabber.parseAccessThreshold(accessthresh);

			if (blindprofile==null || informedprofile==null || grabberprofile==null || accessthresh==null)
			{
				blindprofile = informedprofile = grabberprofile = "standard";
				at = StockInfoGrabber.AccessThreshold.Day;
			}
		}
		catch (Exception ex)
		{
			blindprofile = informedprofile = grabberprofile = "standard";
			at = StockInfoGrabber.AccessThreshold.Day;
		}
		InformedAnalyzer IA = new InformedAnalyzer(symbol,
			blindprofile,
			grabberprofile,
			informedprofile,
			at, false, 0, false, "pe");
		IA.analyze();


		String newResult = symbol + " pe " + IA.pe;
		try { NumericAgent.readDataStringToHT(newResult, ht); }
		catch (Exception e) { System.out.println("WRONG"); }

	    priceEarningsRatio = Double.parseDouble(NumericAgent.get(ht, symbol, "pe"));
	    if (DEBUG) System.out.println(loc + " priceEarningsRatio for " + 
					  symbol + " is " + priceEarningsRatio);
	}

	DataSet d = new DataSet(Local.getName());
	d.put("symbol", symbol);
	d.put("bid", b);
	d.put("priceEarningsRatio", Double.toString(priceEarningsRatio));

	if (price >= 3){
	    if (DEBUG) System.out.println(loc + " symbol is " + symbol);
	    industry = NumericAgent.get(ht, symbol, "industry");
	    if (industry != null) {
		industryPriceEarningsRatio = Double.parseDouble(NumericAgent.get(ht, industry, "pe"));
		if (DEBUG) System.out.println(loc + " industryPriceEarningsRatio for " + 
					      industry + " is " + industryPriceEarningsRatio);
		d.put("industry", industry);
		d.put("industryPriceEarningsRatio", Double.toString(industryPriceEarningsRatio));
	    }
	}
	
	KQMLmsg K = new KQMLmsg();
	K.addFieldValuePair("sender", Local.getName() );
	K.addFieldValuePair("performative", "tell");
	K.addFieldValuePair("in-reply-to", taskReplyWith);
	K.addFieldValuePair("receiver", Local.getName() );
	K.addFieldValuePair("ontology", "PL" );
	K.addFieldValuePair("content", ":result " + d.toString()); // + Local.generate(env)
	MsgUtilities.copyGenContent(startMsg, K, Local);

	System.out.println(Local.getName() + " execute writing: " + K);

	if(true){
	    P1 = new ProvisionCell(K.getKQMLString(),"OK");
	}

	else if(true) {
	    return new ProvisionCell(Local.getName() + ".execute","FAIL");
	}
      return P1; 
  }
   

}//PL_ExecutePE
