import java.io.*;
import java.net.*;
import java.util.*;

/*
 * Failing up: happens when a bid does not fit within constraints, but
 * utility is exceptional. Agents send these kinds of opportunities to
 * agents higher up who may have the authority to accept the bid. If
 * so, execution takes place at bid source and result is forwarded all
 * the way to the authorizing agent.
 * 
 * Numeric agents always pass the decision up. 
 */

public class PL_FailingUp extends GenTask 
{

    static String loc = " in PL_FailingUp ";
    static String em = " Error in PL_FailingUp ";
    static String wm = " WARNING in PL_FailingUp ";
    static final boolean DEBUG = true;
    KQMLmsg startMsg;

   public PL_FailingUp()
   {
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
      System.out.println("% PL_FailingUp Zero constructor % <==============");
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");

   }

   public ProvisionCell checkTime(LinkedListQ Plist, Agent Local)
   {
       System.out.println(Local.getName() + " inside checkTime");
       ht = Local.userHash;
       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
       System.out.println(Local.getName() + " FUcheckTime receiving: " + K);

       startMsg = new KQMLmsg(K.toString());

       String bid = K.getContentFieldValue("opportunity");

       if (Local.isNumeric() || Local.hasNumericChildren()){//cannot approve beyond constraints
	   KQMLmsg K1 = new KQMLmsg();
	   K1.addFieldValuePair("sender", Local.getName() );
	   K1.addFieldValuePair("performative", "achieve");
	   K1.addFieldValuePair("receiver", Local.getParent() );
	   K1.addFieldValuePair("ontology", "PL" );
	   K1.addFieldValuePair("language", "DECAF");
	   MsgUtilities.copyGenContent(K, K1, Local);
	   MsgUtilities.addContentFieldValuePair(K1, "opportunity", bid);
	   MsgUtilities.addContentFieldValuePair(K1, "execTask", 
						 K.getContentFieldValue("execTask"));
	   MsgUtilities.addContentFieldValuePair(K1, "execProvName", 
						 K.getContentFieldValue("execProvName"));
	   MsgUtilities.addContentFieldValuePair(K1, "task", K.getContentFieldValue("task"));
	   System.out.println(loc + " insufficient authority, passing upward: " + K1.toString());
	   
	   return new ProvisionCell(K1.toString(), "insuff");
       }


       
       job = getMsgJob(K);
       localPut("finished", job, new Boolean(false)); //see PL_Count.myTimeout()

       myEndTime = K.getContentFieldValue("endtime");

      String message = new String(Util.getValue(Plist, "MESSAGE"));

      if(true)
      {return new ProvisionCell(bid,"suff");}



      else 
      {return new ProvisionCell("SOME STRING HERE","decline1");}
   }//checkTime

   public ProvisionCell checkMyLimits(LinkedListQ Plist, Agent Local)
   {

       String bid = Util.getValue(Plist, "opp2");

      if(false)
      {return new ProvisionCell("SOME STRING HERE","decline2");}

      else if(false)
      {return new ProvisionCell("SOME STRING HERE","insuff2");}

      else
      {return new ProvisionCell(bid, "suff2");}
   }

   public ProvisionCell decide(LinkedListQ Plist, Agent Local) {
       
       String bid = Util.getValue(Plist, "opp3");
       Bid b = Bid.readBidFromString(bid);

       System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
       System.out.println("% PL_FailingUp decide: " + bid);
       System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
 
       BidSet bs = new BidSet();
       bs.add(b);
       Envelope env = bs.getBestBidsEnvelope(Local.getUtilityFcn());
	//if (envName.equals("empty"))
	//    best = null;
	//else{
	//    env = (Envelope)localGet(envName, job);
	//    best = env.getBid();

       if (env == null){
	   System.out.println(wm + " rejecting null envelope for bid " + b);
	   return new ProvisionCell("null envelope","reject");
       }

      KQMLmsg K = new KQMLmsg();
      K.addFieldValuePair("sender", Local.getName() );
      K.addFieldValuePair("performative", "achieve");
      K.addFieldValuePair("receiver", b.getBidder()); //Local.getParent() );
      K.addFieldValuePair("ontology", "PL" );
      K.addFieldValuePair("language", "DECAF");
      MsgUtilities.copyGenContent(startMsg, K, Local);
      MsgUtilities.addContentFieldValuePair(K, "opportunity", "yes");
      MsgUtilities.addContentFieldValuePair(K, "finalDest", Local.getName());
      MsgUtilities.addContentFieldValuePair(K, startMsg.getContentFieldValue("execProvName"), 
					    env.toString());
      MsgUtilities.addContentFieldValuePair(K, "task", startMsg.getContentFieldValue("execTask"));

      if(false)
      {return new ProvisionCell("SOME STRING HERE","reject");}

      else if(true) {
	  //if asking result to be sent here, add bidder to wait list
	  //so that Count won't finish prematurely.
	  
	  //if (bidder is numeric, add parent to wait list, not CR; how?)

	  //addWaitingForSender(b.getBidder(), job);
	  //System.out.println(loc + " added to waitingForSenders: " + b.getBidder());

	  System.out.println(" decide sending: " + K.toString());
	  Local.sendNow(K);
	  return new ProvisionCell("accept","accept");
      }

      else
      {return new ProvisionCell("SOME STRING HERE","send_up2");}
   }


   public ProvisionCell redirect(LinkedListQ Plist, Agent Local) {

       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
       System.out.println(loc + " redirect receiving: " + K);
       return new ProvisionCell("asdf","OK");
   }

}
