import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Goal has no local bids because the agents below calculate real
 * numbers when they bid, so there is no need for the parent to
 * approximate based on history.
 **/

public class PL_Goal extends GenAgent
{

    static final boolean DEBUG = true;
    static final String loc = "in PL_Goal ";
    static final String em = "ERROR in PL_Goal ";
    Hashtable ht;

    public final boolean PA_FAILS = false;
    public final boolean RA_FAILS = false;
        
   public PL_Goal()
   {
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
      System.out.println("% PL_Goal Zero constructor % <==============");
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");

   }

    /**
     *Initialize a Goal agent.
    **/
   public ProvisionCell start(LinkedListQ Plist, Agent Local)
   {
       super.start(Local, DSType.GOAL);
       System.out.println("**********starting Goal agent: " + Local.getName() + " **************");

       //ht = Local.userHash;

       Local.setLevel(2);
       Local.defaultString = "Goal output";

       Local.setParent("Coord");
       Local.setMyTaskObject(this);
       Local.setHasNumericChildren();
       Local.addChild("RA");
       Local.addChild("PA");

       System.out.println("PL_Goal: children are: " + Local.children);
       
       //Bid b;
       //DataSet d;

       System.out.println("local bids are: " + Local.getLocalBids());

	return new ProvisionCell("started Goal", "OK");
   }

    public RSTTree makeTreeFromData(DataSet d){
	if (DEBUG) System.out.println(loc + "makeTreeFromData d: " + d); 
	
	Bid b = (Bid)d.get("bid");
	RSTTree result = null;
	if (d.getSource().equals("Goal"))
	    result = makeLocalTree(d);
	else if (d.getSource().equals("PA"))
	    result = makePATree(d);
	else if (d.getSource().equals("RA"))
	    result = makeRATree(d);
	else{
	    result = RSTTree.GoalDo("Default phrase " + b.getSymbol() + 
				    " in Goal", DSType.GOAL, b);
	}

	TreeGenerate tg = new TreeGenerate( b.getSymbol() );
	result.updateBid("length_exp", Integer.toString(tg.wordCount( tg.generate(result)) ));

	if (DEBUG) System.out.println(loc + "makeTreeFromData tree: ");
	if (DEBUG && result != null) result.print();


	if (PA_FAILS && d.getSource().equals("PA"))
	return null;

	if (RA_FAILS && d.getSource().equals("RA"))
	 return null;

	return result;
    }// makeTreeFromData(DataSet d)

    private RSTTree makeLocalTree(DataSet d){

	double ds = 0, length = 0;
	String phrase;
	RSTTree result;
	Strength newDS = null;;
	Bid b = (Bid)d.get("bid");

	Strength dsg = (Strength)b.get("DSg_exp");
	if (dsg.direction() > 0)
	    ds = dsg.posCurr();
	else if (dsg.direction() < 0)
	    ds = dsg.negCurr();
	else
	    System.out.println(em + "DataSet has mixed or no Strength: " + d);

	if (ds < -2){ 
	    phrase = "this investment moves you away from your investment goals "; 
	    length = 9; ds = -1;
	}
	else if (ds < 2){ 
	    phrase = "this investment does not have a significant impact on your investment goals ";
	    length = 12; ds = 0;
	}
	else {//if (ds > 1){ 
	    phrase = "this investment moves you towards your investment goals ";
	    length = 8; ds = 1;
	}
	
	newDS = new Strength( ds );
	if (DEBUG) System.out.println(loc + " new DSg is " + newDS);
	b.put("DS" + DSType.GOAL + "_exp", newDS);
	b.put("length_exp", Double.toString(length));
	b.put("topic", DSType.GOAL );
	if (ds >= 0)
	    result = RSTTree.GoalDo(phrase, DSType.GOAL, b);
	else
	    result = RSTTree.GoalDoNot(phrase, DSType.GOAL, b);
	return result;
    }//makeLocalTree(DataSet d)

    //the PortfolioAllocation agent returns a whole tree.
    private RSTTree makePATree(DataSet d){

	RSTTree result = (RSTTree)d.get("tree");
	
	return result;
    }//makePATree(DataSet d)


    //the RetirementAlloc agent returns a whole tree.
    private RSTTree makeRATree(DataSet d){
	Object o = d.get("tree");
	if (o instanceof String && ((String)o).equals("empty"))
	    return null;
	//else
	RSTTree result = (RSTTree)o;
	
	return result;
    }//makeRATree(DataSet d)

}
