import java.io.*;
import java.net.*;
import java.util.*;

/*
 * This agent is a dummy GenAgent acting as parent to IndustryRisk;
 * IR, as a numeric agent, cannot be a direct child of Risk, so this
 * agent goes between. Has no local bids/trees.
 */

public class PL_IRParent extends GenAgent
{
    static final boolean DEBUG = true;
    static final String loc = "in PL_IRParent ";
    static final String em = "ERROR in PL_IRParent ";
    public final boolean TIMING_RUN = false;

   public PL_IRParent()
   {
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
      System.out.println("% PL_IRParent Zero constructor % <==============");
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
   }

    /**
     *Initialize a IRParent agent.
    **/
    public ProvisionCell start(LinkedListQ Plist, Agent Local)
    {
	super.start(Local, DSType.RISK);
	System.out.println("**********starting IRParent agent: " + Local.getName() + " **************");

	Local.setLevel(3);
	Local.setMyTaskObject(this);
	Local.setHasNumericChildren();

	Local.setParent("Risk");
	Local.addChild("IR");
	Local.defaultString = "IRParent output";

	//there are no local bids
	System.out.println("local bids are: " + Local.getLocalBids());

	return new ProvisionCell("started IRParent", "OK");
    }

    public RSTTree makeTreeFromData(DataSet d){
	if (DEBUG) System.out.println(loc + "makeTreeFromData d: " + d); 
	
	Bid b = (Bid)d.get("bid");
	RSTTree result = null;
	if (d.getSource().equals("IR"))
	    result = makeIRTree(d);
	else{
	    result = RSTTree.GoalDo("Default phrase " + b.getSymbol() + " in IRP", DSType.RISK, b);
	}

	if (TIMING_RUN &&
	    5 <= Double.parseDouble((String)b.get("cost_exp"))){
	    //Strength zeroDS = new Strength( 0 );
	    //result.updateBid("DS" + DSType.RISK + "_exp", zeroDS);
	    result.zeroDS(DSType.RISK);
	}

	TreeGenerate tg = new TreeGenerate( b.getSymbol() );
	result.updateBid("length_exp", Integer.toString(tg.wordCount( tg.generate(result)) ));

	if (DEBUG) System.out.println(loc + "leaving makeTreeFromData tree: ");
	if (DEBUG) result.print();
	return result;
    }//makeTreeFromData(DataSet d)


    private RSTTree makeIRTree(DataSet d){

	double ds = 0, length = 0, industryRisk;
	String phrase;
	RSTTree result;
	Strength newDS = null;;
	Bid b = (Bid)d.get("bid");

	industryRisk = Double.parseDouble((String)d.get("industryRisk"));
	Object o = d.get("industry");
	String industry;
	if (o != null) 
	    industry = "the " + (String)o + " industry";
	else
	    industry = "the industry";

	Strength dsr = (Strength)b.get("DSr_exp");
	
	ds = industryRisk;

	if (ds < -2){ 
	    phrase = industry + " has a high risk profile "; 
	    length = 8; ds = -1;
	}
	//else if (ds < -1){ 
	//    phrase = "company has a history of high debt "; 
	//    length = 7; ds = -1;
	//}
	else if (ds < 0){ 
	    phrase = industry + " has a somewhat aggressive risk profile "; 
	    length = 7; ds = -1;
	}
	else if (ds < 2){ 
	    phrase = industry + " has a moderate risk profile ";
	    length = 8;
	}
	else {//if (ds > 1){ 
	    phrase = industry + " has a conservative risk profile ";
	    length = 8;
	}
	
	ds = ds/2;

	newDS = new Strength( ds );
	if (DEBUG) System.out.println(loc + " new DSr is " + newDS);
	b.put("DS" + DSType.RISK + "_exp", newDS);
	b.put("length_exp", Double.toString(length));
	if (ds > 0)
	    result = RSTTree.GoalDo(phrase, DSType.RISK, b);
	else
	    result = RSTTree.GoalDoNot(phrase, DSType.RISK, b);
	return result;
    }

}
