import java.io.*;
import java.net.*;
import java.util.*;
/**
 * See DECAF plan pa.lsp
 * Also see PL_ExecutePA.java
 *
 * This is not strictly a true numeric agent, as PL_ExecutePA returns
 * a tree, not just a DataSet. See parent, PL_Goal.
 *
 **/

public class PL_PortfolioAlloc extends InternalNumericAgent
{
 
    public PL_PortfolioAlloc()
    {
	System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
	System.out.println("% PL_PortfolioAlloc Zero constructor % <==============");
	System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");

	setMakeBidMethod(this.getClass(), "prepareBids");
    }

    public ProvisionCell start(LinkedListQ Plist, Agent Local)
    {
	System.out.println("**********starting PortfolioAlloc agent: " + Local.getName() + " **************");
	super.start(Local, DSType.RISK, "execute-pa.lsp", "ExecutePA");
      
	Local.setLevel(3);
	Local.setParent("Goal");
	Local.defaultString = "PortfolioAlloc output";

	KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
	System.out.println(Local.getName() + " PL_PortfolioAlloc.start: " + K.getKQMLString());

	//no local bids, calculated at run time based on user info
	System.out.println("local bids are: " + Local.getLocalBids());
	return new ProvisionCell(" PA started ","OK");
    }//start(LinkedListQ Plist, Agent Local)



    public BidSet prepareBids(String msg, Agent Local)//
    {
	KQMLmsg K = new KQMLmsg(msg);
	String symbol = K.getContentFieldValue("symbol");
	String shares = K.getContentFieldValue("quantity");
	String price = K.getContentFieldValue("price");
	
	Hashtable ht = Local.userHash;

	//Hashtable ht = NumericAgent.data;

	String investmentType;
	double allocGoal, allocCurrent, allocWithPurchase, investmentTypeBalance, totalBalance,
	    annualConsumption, share_price, num_shares, newAllocStatus, length, ds;
	
	share_price = Double.parseDouble(price);
	num_shares = Double.parseDouble(shares);
	investmentType = (String)NumericAgent.get(ht, symbol, "type");
	investmentTypeBalance = Double.parseDouble(NumericAgent.get(ht, investmentType, "balance"));
	totalBalance = Double.parseDouble(NumericAgent.get(ht, "total", "balance"));
	allocGoal = Double.parseDouble(NumericAgent.get(ht, investmentType, "goal_percent"));
	////yearsToRetire = Double.parseDouble(NumericAgent.get(ht, "year", "to_retire"));
	//annualConsumption = Double.parseDouble(NumericAgent.get(ht, "year", "consumption"));
	allocCurrent = 100 * investmentTypeBalance / totalBalance;
	allocWithPurchase  = 100 * (investmentTypeBalance + share_price * num_shares)/ totalBalance;
	
	//order is important	
	if(allocCurrent > allocGoal){ 
	    length = 9; ds = -2.5;
	}
	else if (allocWithPurchase < (0.9 * allocGoal)){//allocWithPurchase > allocCurrent && 
	    length = 12; ds = 0.1;
	}
	else if (allocWithPurchase < allocGoal){//allocWithPurchase > allocCurrent && 
	    length = 10; ds = 0.5;
	}
	else if (allocWithPurchase < (1.03 * allocGoal)) { //allocWithPurchase > allocCurrent && 
	    length = 8; ds = 2;
	}
	else if (allocWithPurchase < (1.10 * allocGoal)) { //allocWithPurchase > allocCurrent && 
	    length = 8; ds = -1.5;
	}
	else if (allocWithPurchase > allocGoal) { //allocWithPurchase > allocCurrent && 
	    length = 9; ds = -2;
	}
	else  {//never reached
	    length = 5; ds = -12;
	}
	//divide by two for free internal data agent
	Strength newDS = new Strength( ds/2 );
	if (DEBUG) System.out.println(" new DSg is " + newDS);
	Bid b = new Bid(Local.getName());
	b.put("DS" + DSType.GOAL + "_exp", newDS);
	b.put("length_exp", Double.toString(5 + length)); //add for "from a portfolio perspective"
	b.put("cost_exp", "1");
	b.put("symbol", symbol);
	b.put("topic", DSType.GOAL );

	BidSet bs = new BidSet(Local.getName());
	bs.add(b);
	return bs; 
    }//prepareBids


    //for testing purposes only
    public BidSet prepareBids(String msg)//
    {
	KQMLmsg K = new KQMLmsg(msg);
	String symbol = K.getContentFieldValue("symbol");
	String shares = K.getContentFieldValue("quantity");
	String price = K.getContentFieldValue("price");
	
	//Hashtable ht = Local.userHash;

	Hashtable ht = NumericAgent.data;

	String investmentType;
	double allocGoal, allocCurrent, allocWithPurchase, investmentTypeBalance, totalBalance,
	    annualConsumption, share_price, num_shares, newAllocStatus, length, ds;
	
	share_price = Double.parseDouble(price);
	num_shares = Double.parseDouble(shares);
	investmentType = (String)NumericAgent.get(ht, symbol, "type");
	investmentTypeBalance = Double.parseDouble(NumericAgent.get(ht, investmentType, "balance"));
	totalBalance = Double.parseDouble(NumericAgent.get(ht, "total", "balance"));
	allocGoal = Double.parseDouble(NumericAgent.get(ht, investmentType, "goal_percent"));
	////yearsToRetire = Double.parseDouble(NumericAgent.get(ht, "year", "to_retire"));
	//annualConsumption = Double.parseDouble(NumericAgent.get(ht, "year", "consumption"));
	allocCurrent = 100 * investmentTypeBalance / totalBalance;
	allocWithPurchase  = 100 * (investmentTypeBalance + share_price * num_shares)/ totalBalance;
	System.out.println("allocCurrent " + allocCurrent + " allocWithPurchase " + allocWithPurchase + " allocGoal " + allocGoal);
	
	//order is important	
	if(allocCurrent > allocGoal){ 
	    length = 9; ds = -3;
	}
	else if (allocWithPurchase < allocGoal){//allocWithPurchase > allocCurrent && 
	    length = 10; ds = 1;
	}
	else if (allocWithPurchase < (1.03 * allocGoal)) { //allocWithPurchase > allocCurrent && 
	    length = 8; ds = 2;
	}
	else if (allocWithPurchase < (1.10 * allocGoal)) { //allocWithPurchase > allocCurrent && 
	    length = 8; ds = -2;
	}
	else if (allocWithPurchase > allocGoal) { //allocWithPurchase > allocCurrent && 
	    length = 9; ds = -4;
	}
	else  {//never reached
	    length = 5; ds = -12;
	}
	Strength newDS = new Strength( ds );
	 System.out.println(" new DSg is " + newDS);
	Bid b = new Bid("testing");
	b.put("DS" + DSType.GOAL + "_exp", newDS);
	b.put("length_exp", Double.toString(length));
	b.put("cost_exp", "1");
	b.put("symbol", symbol);
	b.put("topic", DSType.GOAL );

	BidSet bs = new BidSet("testing");
	bs.add(b);
	return bs; 
    }//prepareBids2


    static public void main(String[] args){

	KQMLmsg K = new KQMLmsg();

	K.addFieldValuePair("sender", "xxx");
	K.addFieldValuePair("performative", "achieve");
	K.addFieldValuePair("receiver", "xxx");
	K.addFieldValuePair("reply-with", "a");
	K.addFieldValuePair("language", "DECAF");
	K.addFieldValuePair("ontology", "PL");	
	K.addFieldValuePair("content", ":task Coord :msg start " +
			    " :symbol IBM :price 126 :quantity 100 :constraints " + 
			    "{cost_max=30.0, length_max=150, length_min=0 } " + 
			    ":utilityFcn {cost_coeff=1, cost_target=15, cost_fcn=startPlateauNormal, length_coeff=2, length_target=50, length_fcn=startPlateauNormal, DSr_coeff=3, DSr_target=5, DSv_coeff=1, DSv_target=5, DSg_target=5, DSg_coeff=1, RV_coeff=1, DSr_fcn=linear, DSv_fcn=normPlateauEnd, DSg_fcn=normPlateauEnd}");
	
	//PL_DebtRisk dr = new PL_DebtRisk();
	PL_PortfolioAlloc pa = new PL_PortfolioAlloc();
	
	BidSet bs = pa.prepareBids(K.toString());
	System.out.println(bs);
	
    }


}
  
