import java.io.*;
import java.net.*;
import java.util.*;
/**
 * See DECAF plan ra.lsp
 * Also see PL_ExecuteRA.java
 *
 * This is not strictly a true numeric agent, as PL_ExecuteRA returns
 * a tree, not just a DataSet. See parent, PL_Goal.
 *
 **/

public class PL_RetirementAlloc extends InternalNumericAgent
{
 
    public PL_RetirementAlloc()
    {
	System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
	System.out.println("% PL_RetirementAlloc Zero constructor % <==============");
	System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");

	setMakeBidMethod(this.getClass(), "prepareBids");
    }

    public ProvisionCell start(LinkedListQ Plist, Agent Local)
    {
	System.out.println("**********starting RetirementAlloc agent: " + Local.getName() + " **************");
	super.start(Local, DSType.RISK, "execute-ra.lsp", "ExecuteRA");
      
	Local.setLevel(3);
	Local.setParent("Goal");
	Local.defaultString = "RetirementAlloc output";

	KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
	System.out.println(Local.getName() + " PL_RetirementAlloc.start: " + K.getKQMLString());

	//no local bids, calculated at run time based on user info
	System.out.println("local bids are: " + Local.getLocalBids());
	return new ProvisionCell(" RA started ","OK");
    }//start(LinkedListQ Plist, Agent Local)


    public BidSet prepareBids(String msg, Agent Local) {

	KQMLmsg K = new KQMLmsg(msg);
	String symbol = K.getContentFieldValue("symbol");
	String shares = K.getContentFieldValue("quantity");
	String price = K.getContentFieldValue("price");
	
	Hashtable ht = Local.userHash;
	//Hashtable ht = NumericAgent.data;

	int length = 20;//default, is +/- 1 for this agent
	double ds;
	double bondAllocGoal, allocCurrent, allocWithPurchase, investmentTypeBalance, totalBalance,
	    yearsToRetire, annualConsumption, share_price, num_shares, newAllocStatus;
	String investmentType;

	share_price = Double.parseDouble(price);
	num_shares = Double.parseDouble(shares);

	investmentType = (String)NumericAgent.get(ht, symbol, "type");
	investmentTypeBalance = Double.parseDouble(NumericAgent.get(ht, investmentType, "balance"));
	totalBalance = Double.parseDouble(NumericAgent.get(ht, "total", "balance"));
	//allocGoal = Double.parseDouble(NumericAgent.get(ht, investmentType, "goal_percent"));
	yearsToRetire = Double.parseDouble(NumericAgent.get(ht, "year", "to_retire"));
	annualConsumption = Double.parseDouble(NumericAgent.get(ht, "year", "consumption"));
	//allocCurrent = 100 * investmentTypeBalance / totalBalance;
	//allocWithPurchase  = 100 * (investmentTypeBalance + share_price * num_shares)/ totalBalance;
	bondAllocGoal = Double.parseDouble(NumericAgent.get(ht, "bonds", "goal_percent"));

	RSTTree resultTree = null;
	if (yearsToRetire < 6 && 
	    ((5 - yearsToRetire) * annualConsumption) > ((bondAllocGoal/100) * totalBalance)){

	    ds = -((5 - yearsToRetire) * annualConsumption) / ((bondAllocGoal/100) * totalBalance);
	    //reconsider allocation to higher bond percentage
	}

	else  {
	    ds = 0;
	    System.out.println("No local bids, ds == 0");
	}

	BidSet bs = new BidSet(Local.getName());

	if (ds != 0){//only have bids when ds not zero

	    Strength newDS = new Strength( ds );
	    if (DEBUG) System.out.println(" new DSg is " + newDS);
	    Bid b = new Bid(Local.getName());
	    b.put("DS" + DSType.GOAL + "_exp", newDS);
	    b.put("length_exp", Double.toString(length));
	    b.put("topic", DSType.GOAL );
	    b.put("cost_exp", "1");
	    b.put("symbol", symbol);

	    bs.add(b);
	}

	return bs; 
  }

}
  
