import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Agent calling notifyKids() sends msg to all descendants. Recursive
 * calls stop when an agent has no children and send a message back up
 * to parent's catchMsg(). Each catchMsg() propagates up.
 *
 * Updates happen inside notifyKids(), when :update section of message
 * is parsed.
 *
 * Note: restrictive "agent" name lists affects all updates that come
 * after it in the update portion of the message. Otherwise, if no
 * agents list is present, defaults to all agents getting updated.
 *
 * Note: job number is used by Update, not changed.
 * Note: endtime is used by Update, not changed.
 **/

public class PL_Update extends GenTask
{

    static String loc = " in PL_Update ";
    static String em = " Error in PL_Update ";
    static final boolean DEBUG = false;
    KQMLmsg startMsg;
    Vector agentNames;

    public PL_Update()
    {
	System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
	System.out.println("% PL_Update Zero constructor % <==============");
	System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
	agentNames = readStringVector(
	     new StringTokenizer("[PE,RE,Value,PA,RA,Goal,IR,IRP,CR,DE,DR,Risk,Coord]", "[] ,", true));
    }
    
    public ProvisionCell notifyKids(LinkedListQ Plist, Agent Local)
    {
	ht = Local.userHash; //init hashtable
	
	KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
	startMsg = K;
	job = getMsgJob(K);

	String uNum = K.getContentFieldValue("updateNum");
	if (uNum.length() == 0)
	    uNum = "000";
	localPut("UpdateReplyWith", job + uNum, K.getValue("reply-with"));
	localPut("UpdateReplyTo", job + uNum, K.getValue("sender"));
	if (DEBUG) System.out.println("INSIDE Update notify for job: " + job + " " + K);
	
	parseUpdate(K.getContentFieldValue("update"), Local);

	StringBuffer msgs = new StringBuffer();
	ProvisionCell P1 = null;
	KQMLmsg K2;
	
	Iterator kids = Local.getChildren();

	///BASE CASE for messages going to bottom
	if ( ! kids.hasNext()){

	    String names = Local.getName();
	
	    if (DEBUG){
		System.out.println("**************************************************");
		System.out.println("**************************************************");
		int n = 0;
		StringTokenizer st = new StringTokenizer(names);
		while (st.hasMoreTokens()) {n++; st.nextToken();}
		System.out.println("Update contacted: " + n + " " + names);
		System.out.println("**************************************************");
		System.out.println("**************************************************");
	    }

	    K2 = new KQMLmsg();
	    K2.addFieldValuePair("sender", Local.getName() );
	    K2.addFieldValuePair("performative", "tell");
	    K2.addFieldValuePair("in-reply-to", (String)localGet("UpdateReplyWith", job +  uNum));
	    //Local.getParent()); does not necessarily go to parent
	    K2.addFieldValuePair("receiver", (String)localGet("UpdateReplyTo", job +  uNum));
	    K2.addFieldValuePair("ontology", "PL" );
	    K2.addFieldValuePair("language", "DECAF" );
	    //copyGen takes care of update part of msg
	    MsgUtilities.copyGenContent(startMsg, K2, Local);
	    MsgUtilities.addContentFieldValuePair(K2, "namesSoFar", names);
	    MsgUtilities.addContentFieldValuePair(K2, "responses", "someTag2");
	    MsgUtilities.addContentFieldValuePair(K2, "updateNum", uNum);

	    msgs.append(K2.getKQMLString());
	    if (DEBUG) System.out.println("Update bottom sending: " + K2);
	    
	    P1 =  new ProvisionCell(msgs.toString(),"noKids");
	}

	else {
	    int numMsgsSent = 0;
	    kids = Local.getChildren();
	    while(kids.hasNext()){

		K2 = new KQMLmsg();
		K2.addFieldValuePair("sender", Local.getName() );
		K2.addFieldValuePair("performative", "achieve");
		//K2.addFieldValuePair("in-reply-to", 
		// (String)localGet(b.getBidder() + "EnvReplyWith", job));//see countMsgs in acceptRFB
		K2.addFieldValuePair("reply-with", "upd"+Local.getName());//see countMsgs2 in PL_Count
		K2.addFieldValuePair("receiver", (String)kids.next());
		K2.addFieldValuePair("ontology", "PL" );
		K2.addFieldValuePair("language", "DECAF" );
		K2.addFieldValuePair("content", ":task Update :updateMsg data17 ");
		//copyGen takes care of update part of msg
		MsgUtilities.copyGenContent(startMsg, K2, Local);
		MsgUtilities.addContentFieldValuePair(K2, "updateNum", uNum);

		if (DEBUG) System.out.println(Local.getName() + "INSIDE Update notify writing: " + K2);
	    
		msgs.append(K2.getKQMLString());
		numMsgsSent++;
	    }
	    localPut("updateNumMsgsSent", job + uNum, new Integer(numMsgsSent));
	    P1 =  new ProvisionCell(msgs.toString(),"OK");
	}
	return  P1;
    }//notifyKids

    public ProvisionCell catchMsgs(LinkedListQ Plist, Agent Local)
    {

	KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
	KQMLmsg startMsg = K;
	if (DEBUG) System.out.println(Local.getName() + "INSIDE Update CATCH receiving: " + K);
	    
	String uNum = K.getContentFieldValue("updateNum");
	if (uNum.length() == 0)
	    uNum = "000";

	Integer numMsgsCollected = (Integer)localGet("updateNumMsgsCollected", job +  uNum);
	if (numMsgsCollected == null){
	    numMsgsCollected = new Integer(1);
	}
	else
	    numMsgsCollected = new Integer( 1 + numMsgsCollected.intValue() );
	localPut("updateNumMsgsCollected", job +  uNum, numMsgsCollected);

	String names = (String)localGet("updateNamesCollected", job +  uNum);
	if (names == null)
	    names = K.getContentFieldValue("namesSoFar");
	else
	    names = names + " " + K.getContentFieldValue("namesSoFar");
	localPut("updateNamesCollected", job +  uNum, names);
	
	if (Local.getLevel() == 1){ //is top level agent
	    System.out.println("**************************************************");
	    System.out.println("**************************************************");
	    int n = 1;
	    StringTokenizer st = new StringTokenizer(names);
	    while (st.hasMoreTokens()) {n++; st.nextToken();}
	    System.out.println("Update contacted: " + n + " " + names + " " + Local.getName());
    System.out.println("start: "+startMsg.toString());
	    System.out.println("**************************************************");
	    System.out.println("**************************************************");

	    String endtime;
	    if (startMsg != null)
		endtime = startMsg.getContentFieldValue("endtime");
	    else endtime = "";
	    writeAgentMsgToFile("Update contacted: " + n + " " + names + " " + 
				Local.getName(), endtime, Local);

	    localPut("updateTotalAgentCount", job +  uNum, new Integer(n));//for use outside?
	}
	else {
	    KQMLmsg K2 = new KQMLmsg();
	    K2.addFieldValuePair("sender", Local.getName() );
	    K2.addFieldValuePair("performative", "tell");
	    K2.addFieldValuePair("in-reply-to", (String)localGet("UpdateReplyWith", job +  uNum));
	    //K2.addFieldValuePair("reply-with", "updated");//see countMsgs2 in PL_Count
	    K2.addFieldValuePair("receiver", Local.getParent());
	    K2.addFieldValuePair("ontology", "PL" );
	    K2.addFieldValuePair("language", "DECAF" );
	    //copyGen takes care of update part of msg
	    MsgUtilities.copyGenContent(startMsg, K2, Local);
	    MsgUtilities.addContentFieldValuePair(K2, "namesSoFar", names + " " + Local.getName());
	    MsgUtilities.addContentFieldValuePair(K2, "responses", "someTag");
	    MsgUtilities.addContentFieldValuePair(K2, "updateNum", uNum);

	    int numMsgsSent = ((Integer)localGet("updateNumMsgsSent", job +  uNum)).intValue();
	    if(numMsgsCollected.intValue() == numMsgsSent) {
		if (DEBUG) System.out.println(Local.getName() + " Update CATCH writing: " + K2);
		return new ProvisionCell(K2.getKQMLString(), "OK");
	    }
	    else {
		System.out.println(Local.getName() + " waiting; so far have: " + names);
		return new ProvisionCell("waiting","waiting");
	    }

	}
	return new ProvisionCell("top", "OK");
    }//catchMs

    //Parses update information, then executes corresponding updating tasks
    private void parseUpdate(String input, Agent Local){

	StringTokenizer st = new StringTokenizer(input, "[], {}=", true); // "{}, =");

	parseUpdate(st, Local);
    }

    private void parseUpdate(StringTokenizer st, Agent Local){
	String key, eq = null, value = null;
	boolean errorFlag = false;
	if (DEBUG) System.out.println(loc + "inside parseUpdate");
	
	if (DEBUG && ! st.hasMoreTokens())
	    System.out.println("WARNING parsing update message - no content!");

	while (st.hasMoreTokens()){
	    key = st.nextToken("[], {}=");
	    if (DEBUG) System.out.println(loc + "inside parseUpdate, key=" + key);

	    if ( key.matches(",") || key.equals(" ") ) 
		continue;

	    else if ( key.startsWith("agent") ) {
		//System.out.println("inside agent, names are: " + agentNames);
		
		eq = st.nextToken();
		if (eq.equals("=")){
		    agentNames = readStringVector(st);
		    //System.out.println("inside agent 2, names are: " + agentNames);
		    //System.out.println("inside agent 2 of: " + Local.getName());
		    if ( ! agentNames.contains(Local.getName()) )
			return; //don't parse any more
		}
		else errorFlag = true;
	    }
	    else if ( key.startsWith("utility") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    BidUtility bu  = BidUtility.readBidUtility(st);
		    Local.setUtilityFcn(bu);
		    //if (DEBUG) 
	System.out.println("############################################################");
	System.out.println("####################Ufcn is now: " + Local.getUtilityFcn());
	System.out.println("############################################################");
		}
		else errorFlag = true;
	    }
	    else if ( key.startsWith("delay") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    String num = st.nextToken();
		    Local.setSimulatedDelay(Integer.parseInt(num));
		    if (DEBUG) System.out.println("Delay is now: " + Local.getSimulatedDelay());
		}
		else errorFlag = true;
	    }
	    else if ( key.startsWith("numHigh") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    String num = st.nextToken();
		    Local.setNumHighUtilBids(Integer.parseInt(num));
		    if (DEBUG) 
			System.out.println("Num High Bids is now set to: " + Local.getNumHighUtilBids());
		}
		else errorFlag = true;
	    }
            else if ( key.startsWith("numAlt") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    String num = st.nextToken();
		    Local.setNumAltUtilBids(Integer.parseInt(num));
		    if (DEBUG) 
			System.out.println("Num Alt Bids is now set to: " + Local.getNumAltUtilBids());
		}
		else errorFlag = true;
	    }
            else if ( key.startsWith("data") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    String filename = st.nextToken();
		    if (Local.getMyTaskObject() instanceof NumericAgent)
			((NumericAgent)Local.getMyTaskObject()).replaceData(filename, Local);	    
		}
		else errorFlag = true;
	    }
        
		else if ( key.startsWith("userDataString") ) {//order is crucial
		eq = st.nextToken();
		if (eq.equals("=")){
		    String dataString = readUserDataString(st);
		    if (Local.getMyTaskObject() instanceof NumericAgent){
			((NumericAgent)Local.getMyTaskObject()).replaceSomeDataString(dataString, Local);
		    }
		}
		else errorFlag = true;
	    }

		// Added by Evan Steinberger
			/*
		else if (key.startsWith("informedData"))
		{
			eq = st.nextToken();
			if (eq.equals("="))
			{
				String dataString = st;
				if (Local.getMyTaskObject() instanceof NumericAgent)
				{
					((NumericAgent)Local.getMyTaskObject()).replaceSomeDataString(dataString, Local);
				}
			}
			else errorFlag = true;
		}*/

            else if ( key.startsWith("userData") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    String filename = st.nextToken();
		    if (Local.getMyTaskObject() instanceof NumericAgent)
			((NumericAgent)Local.getMyTaskObject()).replaceSomeData(filename, Local);
		}
		else errorFlag = true;
	    }
            else if ( key.startsWith("localBids") ) {
		eq = st.nextToken();
		if (eq.equals("=")){
		    String filename = st.nextToken();
		    BidSet bs = BidSet.readBidsFromFile(filename);
		    bs.setEachBidOwner(Local.getName());
		    Local.clearLocalBids();//don't do this at a critical moment
		    Iterator bsIter = bs.iterator();
		    Bid b;
		    DataSet d;
		    while (bsIter.hasNext()){
			b = (Bid)bsIter.next();
			d = new DataSet(Local.getName());
			d.put("bid", b);
			Local.addLocalBid(b, d.toString());
		    }
		    //System.out.println("Local bids are now: " + Local.getLocalBids());
		    
		}
		else errorFlag = true;
	    }



	    if (errorFlag)
		System.out.println(em + " errorFlag set for key: " + 
				   key + " eq: " + eq + " value: " + value);
	}
   	
    }//	parseUpdate(StringTokenizer st, Agent Local)

    /**
     * Reads new user portfolio data bracketed by []. Designed for use with GUI portfolio screen.
     **/
    private String readUserDataString(StringTokenizer st){

	String next = null;
	StringBuffer data = new StringBuffer();

	next = st.nextToken("[], ");

	if (next.equals("[")){
	    while (st.hasMoreTokens()){
		next = st.nextToken(); 
		if (next.equals("]"))
		    break;
		//if (DEBUG) System.out.println(next);
		if ( next.equals("empty") || next.equals(",") || next.equals(" "))
		    continue;
		else {
		    data.append(next);
		    data.append(" ");
		}
	    }
	    //System.out.println("\nREAD DATA STRING: " + data.toString());
	    return data.toString();
	}
	else 
	    System.out.println(em + "incorrect first token in readUserDataString(StringTok):" + next);
	return null;
    }

    /**
     * Designed to read a Vector of agent names. Return delimiters
     * must be true in st.
     **/
    private Vector readStringVector(StringTokenizer st){

	String next = null;
	Vector v = new Vector(14);
	//StringTokenizer st = new (s.trim(), "[], ", true);

	next = st.nextToken("[], ");

	if (next.equals("[")){
	    while (st.hasMoreTokens()){
		next = st.nextToken(); 
		if (next.equals("]"))
		    break;
		//if (DEBUG) System.out.println(next);
		if ( next.equals("empty") || next.equals(",") || next.equals(" "))
		    continue;
		else 
		    v.add(next);
	    }
	    return v;
	}
	else 
	    System.out.println(em + "incorrect first token in readStringVector(StringTok):" + next);
	return null;
    }

    
}
