import java.io.*;
import java.net.*;
import java.util.*;

/**
 * See acceptRFB.lsp in planeditor.
 *
 * RFB is Request For Bids. Accepts a
 * request from parent, and responds with a msg containing a
 * BidSet. The BidSet contains single bids that represent various
 * combinations of bids from this agent's children. This agent may
 * also have some local bids.
 **/

public class PL_acceptRFB extends GenTask
{
    //static int NUM_HIGH_UTIL_BIDS = 10; see AgentDef
    static final boolean OPPORTUNITY = false;
    static final boolean DEBUG = false;
    static String loc = " in PL_acceptRFB ";
    static String em = " Error in PL_acceptRFB ";
    Integer messageCount = new Integer(0); //modified in countMsgs()
    BidUtility utility;
    StringBuffer bidsFromChildren = new StringBuffer(); //modified in countMsgs()
    String taskReplyWith;
    String symbol = null;
    KQMLmsg startMsg;

    //String myEndTime; //time by which this task should have completed and returned results

   public PL_acceptRFB()
   {
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
      System.out.println("% PL_acceptRFB Zero constructor % <==============");
      System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
   }

   public ProvisionCell generateBids(LinkedListQ Plist, Agent Local)
   {
       if (DEBUG) System.out.println(Local.getName() + " inside generateBids");

       ht = Local.userHash;
       //myEndTime = Local.getEndTime(Local.taskLowerTimeout());

       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
       System.out.println(Local.getName() + loc + "generateBids receiving: " + K);

       job = getMsgJob(K);

       Local.checkMsgTime(K, Local.dataFile1, job);

       bidsMade = (Hashtable)localGet("bidsMade", job);
       bidsAccepted = (Hashtable)localGet("bidsAccepted", job);
       
       startMsg = new KQMLmsg(K.toString());//save copy
       symbol = (String)K.getContentFieldValue("symbol");
       String constraintString = (String)K.getContentFieldValue("constraints");
       String utilityFcnString = (String)K.getContentFieldValue("utilityFcn");
       taskReplyWith = (String)K.getValue("reply-with");
       
       ConstraintSet cs = ConstraintSet.readCSFromString(constraintString);
       Local.constraints = cs; //FIX can't currently have agent
			       //considering different constraints for
			       //different bids; is this an issue?
			       //will be if agent handles more than
			       //one bid at once. When is this used?
       if (DEBUG) System.out.println(Local.getName() + " ConstraintSet: " + cs);
       ht.put(cs.getName(), cs);

       utility = BidUtility.readBidUtilityFromString(utilityFcnString);
       if (DEBUG) System.out.println(Local.getName() + " BidUtility: " + utility);
       Local.setUtilityFcn(utility);
       
       return new ProvisionCell(cs.getName(), "OK");
   }//generateBids(LinkedListQ Plist, Agent Local)

    public ProvisionCell generateLocBids(LinkedListQ Plist, Agent Local)
    {
	if (DEBUG) System.out.println(Local.getName() + " INSIDE generateLocBids");
       
	//KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
	//ConstraintSet cs = (ConstraintSet)ht.get( (String)K.getContentFieldValue("constraintsName") );
	//ConstraintSet cs = (ConstraintSet)ht.get( (String) Util.getValue(Plist, "constraints") );
	//if (DEBUG) System.out.println(Local.getName() + " inside generateLocBids, constraints are " + cs);
       
	if (DEBUG) System.out.println(Local.getName() + " generateLocBids returning: " + Local.getLocalBids(startMsg, Local));
       
	if (DEBUG) System.out.println(Local.getName() + " Leaving genLocBids");
	ProvisionCell P1 =  new ProvisionCell(Local.getLocalBids(startMsg, Local).toString(), "OK");
	return P1;
    }//generateLocBids(LinkedListQ Plist, Agent Local)

   public ProvisionCell generateChildBids(LinkedListQ Plist, Agent Local)
   {
       if (DEBUG) System.out.println(Local.getName() + " INSIDE generateChildBids");
       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));
       ProvisionCell P1 = null;
       StringBuffer msgs = new StringBuffer();

       String constraintsName = (String)Util.getValue(Plist, "in2");

       if (DEBUG) System.out.println(Local.getName() + " PL_acceptRFB: children are: " + Local.children);

       int size = Local.getNumberOfChildren();

       if (size > 0){
	   Iterator kids = Local.getChildren();
	   while (kids.hasNext()){
	       KQMLmsg K2 = new KQMLmsg();
	       K2.addFieldValuePair("sender", Local.getName() );
	       K2.addFieldValuePair("performative", "achieve");
	       K2.addFieldValuePair("reply-with", "whatgoeshere");
	       K2.addFieldValuePair("receiver", (String)kids.next() );
	       K2.addFieldValuePair("ontology", "PL" );
	       //must have :bidset field for countMsgs to see msg
	       K2.addFieldValuePair("content", ":task BidAgent :bidset unused :msg spud :constraints " 
				    + ht.get(constraintsName) + " :utilityFcn " + Local.getUtilityFcn());
	       MsgUtilities.copyGenContent(K, K2, Local);
	       //reduce levels by one to reflect number of levels below and including receiver
	       //int value = Integer.parseInt( K2.getContentFieldValue("levels") );
	       //MsgUtilities.replaceContent(K2, "levels",  Integer.toString(value - 1) );
	       System.out.println(Local.getName() + " generateChildBids writing: " + K2);
	       msgs.append(K2.getKQMLString());
	       msgs.append(Util.addTimeout(Local.lowerMsgTimeout(K.getContentFieldValue("endtime")) ));
	   }
	   P1 =  new ProvisionCell(msgs.toString(),"OK");
       }
       else { //no children
	   KQMLmsg K2 = new KQMLmsg();
	   K2.addFieldValuePair("sender", Local.getName() );
	   K2.addFieldValuePair("receiver", Local.getName() );
	   K2.addFieldValuePair("performative", "tell");
	   K2.addFieldValuePair("content", ":bidset empty");
	   MsgUtilities.copyGenContent(K, K2, Local);
	   System.out.println(Local.getName() + " generateChildBids writing: " + K2);
	   P1 =  new ProvisionCell(K2.getKQMLString(),"OK"); //send msg to count
       }
       if (DEBUG) System.out.println(Local.getName() + " Leaving generateChildBids: "  + P1.getName() + P1.getData());
       return P1;
   }//generateChildBids(LinkedListQ Plist, Agent Local)

    /**
     * First message received will be from agent requesting the
     * counting; this ensures that this Action will receive at least
     * one message. Messages after that will be from children and
     * should be collected/counted.
     **/
   public ProvisionCell countMsgs(LinkedListQ Plist, Agent Local)
   {
       ProvisionCell P1 =  new ProvisionCell("","NotDone"); //default
       String bidSetString, replyWithString, sender;
       KQMLmsg K = new KQMLmsg(Util.getValue(Plist, "MESSAGE"));

       Local.checkMsgTime(K, Local.dataFile1, job);

       synchronized(messageCount){

	   int msgCount = messageCount.intValue(); //for convenience; put it back later

	   msgCount++;//this doesn't work without synch; race
		      //conditions for different instances of this
		      //method.(messageCount.intValue())

	   if (DEBUG) System.out.println("inside " + Local.getName() + " countMsgs: " + K); 
	   if (DEBUG) System.out.println("inside " + Local.getName() + " countMsgs 2: " +  (String)Util.getValue(Plist, "bidset"));
	   int childCount = Local.getNumberOfChildren();
	   if (DEBUG) System.out.println("countMsgs waiting for n children: " + childCount + ", msgCount is: " + msgCount);
       
	   if (childCount == 0){
	       System.out.println("leaving " + Local.getName() + " countMsgs, Done "); 
	       P1 =  new ProvisionCell("empty", "Done");
	   }
	   //       else if (msgCount == 0)
	   //	   P1 =  new ProvisionCell("","NotDone");//wait for another message
	   else {
	       bidSetString = (String)K.getContentFieldValue("bidset");
	       replyWithString = K.getValue("reply-with");
	       sender = K.getValue("sender");
	       ht.put(sender + "EnvReplyWith", replyWithString);//see distributeEnv in acceptEnv
	       System.out.println(Local.getName() + " countMsgs receiving bidset " + msgCount + ": " + bidSetString);       
	       bidsFromChildren.append(bidSetString);
	       P1 =  new ProvisionCell("","NotDone");//wait for another message

	       if (msgCount == childCount){ 
		   System.out.println(  Local.getName() + " countMsgs finished: " + bidsFromChildren.toString());
		   P1 =  new ProvisionCell("[" + bidsFromChildren.toString() + "]","Done");//make Vector string
	       }
	   }
	   messageCount = new Integer(msgCount);
       }

       return P1;
   }//countMsgs(LinkedListQ Plist, Agent Local)

    /**
     * Takes in a msg containing local bidset and bidsets from
     * children, makes combinations of same, then makes a single bid
     * to represent each bidset; returns BidSet of best (or some
     * choice) of those single bids.
     **/
   public ProvisionCell makeNewBidSet(LinkedListQ Plist, Agent Local)
   {
       int NumAltUtilBids = Local.getNumAltUtilBids();
       int NumHighUtilBids = Local.getNumHighUtilBids();

       if (DEBUG) System.out.println(Local.getName() + " inside makeNewBidSet");
       String kids = (Util.getValue(Plist, "childBids"));
       if (DEBUG) System.out.println(Local.getName() + " makeNewBidSet children's bids: " + kids);
       String local = (Util.getValue(Plist, "localBids"));
       if (DEBUG) System.out.println(Local.getName() + " makeNewBidSet local bids: " + local);

       Vector vAll = null;
       BidSet localBids = BidSet.readBidSetFromString(local);

       if (kids.equals("empty")){
	   vAll = new Vector(5);
	   if (localBids != null) vAll.add(localBids);
       }
       else {
	   vAll = BidSet.readBidSetVector(kids);
	   if (localBids != null) vAll.add(localBids);
       }

       //Local.allBidsConsidered = vAll; no, put in ht with bid made + "bidsConsidered" as key
       localPut("bidsConsidered", job, vAll); //is a vector of bidSets

       KQMLmsg K = new KQMLmsg();
       K.addFieldValuePair("sender", Local.getName() );
       K.addFieldValuePair("performative", "tell");
       K.addFieldValuePair("in-reply-to", taskReplyWith);
       K.addFieldValuePair("receiver", Local.getName()); //Local.getParent() );
       K.addFieldValuePair("ontology", "PL" );
       K.addFieldValuePair("language", "DECAF");

       if (vAll.isEmpty()){
       	   K.addFieldValuePair("content", ":newbidset empty");
       }
       else {
	   if (DEBUG) System.out.println(Local.getName() + " local and child bidsets: " + vAll);

	   Vector newBidCombos = BidSet.chooseOneCombos(Local.getName(), vAll);

	   System.out.println(loc + " numHigh: " + NumHighUtilBids + " numAlt: " + NumAltUtilBids);
	   if (DEBUG) for (int i = 0; i<newBidCombos.size(); i++)
	       System.out.println( ((BidSet)newBidCombos.get(i)).getOwner() + " " + 
				   ((BidSet)newBidCombos.get(i)).size() );
	   System.out.println(Local.getName()+ " size new bid combos: " + newBidCombos.size());
	   localPut("numBidSetsConsidered", job, Integer.toString(newBidCombos.size()));
	   
	   //if (Local.getName().equals("Coord"))
	   //writeAgentMsgToFile("\n newBidCombos: \n" + newBidCombos, startMsg.getContentFieldValue("endtime"), Local);
       
	   BidSet bs = BidSet.makeSingleBidsFromBidSets(Local.getName(), newBidCombos);
	   System.out.println("Finished single bid representatives.");
	
	   //if (Local.getName().equals("Coord"))
	   //writeAgentMsgToFile("\n\n bs: \n" + bs, startMsg.getContentFieldValue("endtime"), Local);
          
	   if (DEBUG) System.out.println("acceptRFB.makeNewBidSet new combined bids before constraints: " + bs);
	   
	   ConstraintSet cs = (ConstraintSet)ht.get( (String) Util.getValue(Plist, "constraintName") );       
	   if (DEBUG) System.out.println(Local.getName() + " constraints: " + cs);
	   
	   Bid highestBeforeConstraints = null; //Failing Up
	   if (OPPORTUNITY && Local.isNumeric())
	       highestBeforeConstraints = bs.getHighestUtilityBid(Local.getUtilityFcn());
	   
	   bs.filterUsingConstraints(cs);
	   System.out.println("Filtered with constraints.");
	   
	   Bid hi = bs.getHighestUtilityBid(Local.getUtilityFcn());
	   System.out.println("Got highest bid. Getting " + NumAltUtilBids + " alts.");
	   //System.out.println("OOOOOOOOOOOOOOOOOOOOO hi: " + hi);
	   
	   ///get alternatives before bs gets shrunk to N high U bids; currently kruskal's
	   BidSet bsAlt = null;
	   if (bs.size() > NumHighUtilBids)
	       bsAlt = bs.getNAltUtility(NumAltUtilBids, Local.getUtilityFcn(), hi);
	   System.out.println("Finished getting " + NumAltUtilBids + " alts.");
	   
	   bs = bs.getNHighestUtility(NumHighUtilBids, Local.getUtilityFcn());
	   System.out.println("Got N highest bids.");
	   
	   if (bsAlt != null) bs.addAll(bsAlt);
	   bs.setOwner(Local.getName());
	   setBidsMade(bs);//stores bid=bidset source pairs in Agent
	   
	   if (DEBUG) System.out.println(loc + "new combined bids made: " + getBidsMade());
       
	   ////Failing Up (great utility, but doesn't meet constraints)
	   ////Moved to execution where it makes more sense
	   if (highestBeforeConstraints != null &&
	       ! bs.contains(highestBeforeConstraints)){
	       System.out.println(Local.getName() + " sending special opportunity: " + highestBeforeConstraints);
	       notifyFailUp(highestBeforeConstraints, Local);
	       BidSet bsF = new BidSet(Local.getName());
	       bsF.add(highestBeforeConstraints);
	       setBidsMade(bsF);//stores bid=bidset_source pairs in Agent
	   }

	   if (DEBUG){
	       System.out.println(loc + "Listing bids in bidset: ");
	       Bid b;
	       Iterator bids = bs.iterator();
	       while(bids.hasNext()){
		   b = (Bid)bids.next();
		   System.out.println("bid: " + b);
		   System.out.println("source: " + b.getBidSetSource());
		   //System.out.println("source2: " + Local.getBidComponents(b));
	       }
	   }
	   String bidString = bs.toString();
	   if (bidString == null || bidString.equals("")) bidString = "empty";

	   if (DEBUG) System.out.println(Local.getName() + " new bid set: " + bidString);
       
	   K.addFieldValuePair("content", ":newbidset " + bidString);
	   System.out.println(Local.getName() + " sending msg to BidAgent.propagateBidsUp: " + K);
       }
       return new ProvisionCell(K.getKQMLString(),"OK");
   }//makeNewBidSet

    private void notifyFailUp(Bid highestBeforeConstraints, Agent Local){
	System.out.println("Entering notifyFailUp " + highestBeforeConstraints);

	KQMLmsg K = new KQMLmsg();
	K.addFieldValuePair("sender", Local.getName() );
	K.addFieldValuePair("performative", "achieve");
	K.addFieldValuePair("receiver", Local.getName()); //Local.getParent() );
	K.addFieldValuePair("ontology", "PL" );
	K.addFieldValuePair("language", "DECAF");
	MsgUtilities.copyGenContent(startMsg, K, Local);
	MsgUtilities.addContentFieldValuePair(K, "opportunity", 
					      highestBeforeConstraints.toString());
	MsgUtilities.addContentFieldValuePair(K, "execTask", Local.getExecuteTaskName());
	MsgUtilities.addContentFieldValuePair(K, "execProvName", "env1");
	MsgUtilities.addContentFieldValuePair(K, "task", "FailingUp");
	Local.sendNow(K);
    }

    public static void main(String[] args){

	//	PL_BidAgent ba = new PL_BidAgent();

	BidUtility bu = new BidUtility();
	bu.setAttributeCoeff("cost_exp", -5);
	//bu.setAttributeCoeff("DS_exp", 50);

	ConstraintSet cs = new ConstraintSet();
	cs.put("cost_max", "17");

	KQMLmsg K = new KQMLmsg();

	 K.addFieldValuePair("sender", "spam");
	 K.addFieldValuePair("performative", "tell");
	 K.addFieldValuePair("receiver", "spam");
	 K.addFieldValuePair("content", ":constraints " + cs.toString() + " :utilityFcn " + bu.toString());

	 ProvisionCell P1 =  new ProvisionCell(K.getKQMLString(),"MESSAGE");

	LinkedListQ list = new LinkedListQ();
	list.add(P1);

	PL_acceptRFB arfb = new PL_acceptRFB();
	arfb.ht = new Hashtable();
	//	ProvisionCell P = arfb.generateBids(list);

	//list.add(P);

	//arfb.generateLocBids(list);
    }

}



//       ProvisionCell p = PL_Planner.PL_run(Local); //sends kqml msg with var mappings and root name to PL_AddPlan
//       KQMLmsg K3 = new KQMLmsg(p.getData());
//       String planRootName = K3.getContentFieldValue("root");
//
//       String planFileName = Local.getPlanFile();
//       Local.setTaskTemplate(planFileName);
//
//       KQMLmsg K2 = new KQMLmsg("(achieve :sender " + Local.getName() +" :receiver " + Local.getName() + " :reply-with Message_1 :ontology PL :language DECAF :content (:task " + planRootName + " :?x1 SAP))");
//       Util.send(K2.getKQMLString());
