// Evan Steinberger
// CISC 466
// Spring 2007

/**
 * QuoteComGrabber pulls data from Quote.Com's financial web site(s).
 * 
 * @author Evan Steinberger
 */

public class QuoteComGrabber extends StockInfoGrabber
{
	// Contains P/E Ratio
	private static final String address =
		"http://new.quote.com/stocks/company.action?page=fin&sym=";

	// (no info on QuoteCom about industry)

	private String crSearch = "Current ratio";
	private String peSearch = "P/E</acronym> Ratio";
	private String deTotalSearch = "Total debt to equity";
	private String reTTMSearch = "Return on Equity";

	/**
	 * Constructor
	 * 
	 * @param stockSymbol	the stock ticker symbol
	 * @param at	the Access Threshold for pulling data
	 */
	public QuoteComGrabber(String stockSymbol, AccessThreshold at)
	{
		super.currentAccessThreshold = at;
		super.init(stockSymbol);
		this.validateSymbol();
	}

	/**
	 * Ensures that the selected symbol is a valid.
	 * 
	 */
	protected void validateSymbol()
	{
		super.downloadHTML(this.buildURL(address));
		if (super.pageHTML.startsWith(WebQuery.errorMessageStart))
		{
			super.mineErrorMsg = "Error: URL is not valid.";
			super.validSymbol = false;
		}
		else
		{
			if (!super.pageHTML.contains("Company Information"))
				super.validSymbol = false;
			else
				super.validSymbol = true;
		}
	}

	/**
	 * Builds a URL based on a prefix, stock symbol, and any
	 * other additional information.
	 * 
	 * @param s additional information placed into the URL
	 * 
	 */
	protected String buildURL(String s)
	{
		return s + super.stockSymbol;
	}

	/**
	 * Mines all relevant information.
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	public boolean mineAllInformation()
	{
		if (super.hasHistoricalContent)
			return true;
		if (!super.validSymbol)
		{
			if (super.mineErrorMsg == "")
				super.mineErrorMsg = "Error: Invalid Symbol Provided to Data Miner.";
			return false;
		}
		return this.mineCurrentRatio() && this.minePeTTM() && this.mineDeTotal() &&
			this.mineReTTM() && this.mineIndustry();
	}

	/**
	 * Mines the Current Ratio
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineCurrentRatio()
	{
		String temp = this.mineSingleData(this.crSearch);
		if (temp.equals("N/A"))
			return true;
		try { super.currentRatio = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining Current Ratio: " + ex.toString();
			return false;
		}
		return true;
	}

	/**
	 * Mines Price/Earnings ratio
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean minePeTTM()
	{
		String temp = this.mineSingleData(this.peSearch);
		if (temp.equals("N/A"))
			return true;
		try { super.peTTM = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining P/E TTM: " + ex.toString();
			return false;
		}
		return true;
	}

	/**
	 * Mines the Debt/Equity Ratio.
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineDeTotal()
	{
		String temp = this.mineSingleData(this.deTotalSearch);
		if (temp.equals("N/A"))
			return true;
		try { super.deTotal = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining DE Total: " + ex.toString();
			return false;
		}
		// Specific Rule for QuoteCom
		super.deTotal/=100;
		return true;
	}

	/**
	 * Mines Return on Equity
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineReTTM()
	{
		String temp = this.mineSingleData(this.reTTMSearch);
		if (temp.equals("N/A"))
			return true;
		try { super.reTTM = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining RE TTM: " + ex.toString();
			return false;
		}
		return true;
	}
	
	/**
	 * Mines the Industry.
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineIndustry()
	{
		// No industry info is provided for this site
		return true;
	}
	
	/**
	 * Mines piece of single data
	 * 
	 * @param searchFor	a string to search for
	 * @param beginParser a string where the parsing begins
	 * 
	 * @return	a string of parsed data
	 * 
	 */
	private String mineSingleData(String searchFor)
	{
		try
		{
			String temp = "";
			int start = super.pageHTML.indexOf(searchFor);
			int end;
			if (start < 0)
				return "";
			temp = pageHTML.substring(start);
			temp = temp.substring(searchFor.length(), temp.length()).trim();
			// Specific rules for quotecom
			start = temp.indexOf("strong>");
			end = temp.indexOf("</strong");
			if (start < 0 || end < 0)
				return "";
			temp = temp.substring(start + 7, end);
			if (temp.endsWith("%"))
				temp = temp.substring(0, temp.length() -1);
			return temp;
		}
		catch (Exception ex) { return ""; }
	}
}