import java.util.*;

/**
 * RSTNode.java
 * Children are RSTTrees.
 *
 * Created: Fri Dec 12 22:24:26 2003
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class RSTNode extends Predicate {
    
    /** Error message string. **/
    static String em = "Error in RSTNode.java: ";
    static String wm = "Warning in RSTNode.java: ";
    static String loc = "RSTNode.java: ";
    static boolean DEBUG = false;

    /** For discourse planning **/
    int nuclear = RSTNuc.UNSET;
    /** For discourse planning. myNucleus points to a nuc *within* this structure.**/
    RSTNode myNucleus = null;

    RSTTree myTree = null; //the RSTTree of which this is the root.
    Vector children; //are RSTTrees.
    //    RSTNode parent;

    public RSTNode() {
	children = new Vector(5);
	//parent = null;
    } // RSTNode constructor

    public RSTNode(String n) {
    	super(n);
    	children = new Vector(5);
    	//parent = null;
    }
    public RSTNode(String n, Object o1) {
	super(n, o1);
	children = new Vector(5);
	//parent = null;
    }
    public RSTNode(String n, Object o1, Object o2) {
	super(n, o1, o2);
	children = new Vector(5);
	//parent = null;
    }
    public RSTNode(String n, Object arguments[]) {
	super(n, arguments);
	children = new Vector(5);
	//parent = null;
    }
    public RSTNode(Predicate p){
	super(p.name, p.args);
	children = new Vector(5);
	//parent = null;
    }

    public void setTree(RSTTree t){myTree = t;}
    public RSTTree getTree(){return myTree;}

    public int getNumberOfChildren(){
	if (children != null)
	    return children.size();
	else return 0;
    }

    public void sortChildren(BidUtility bu){
	RSTTreeComp rc = new RSTTreeComp(bu);
	Collections.sort(children, rc);
    }

    public static RSTNode readRSTNode(String s){
	StringTokenizer st = new StringTokenizer(s," \t\n\r\f[]", true );
	st.nextToken(); //get past [
	return readRSTNode(st);
    }

    public static RSTNode readRSTNode(StringTokenizer st){
	RSTNode r = new RSTNode(Predicate.readPredicate(st));
	return r;
    }

    
    public Iterator getChildren(){
	return children.iterator();
    }
    public Vector getChildrenVector(){
	return children;
    }

    public void addChild(RSTTree r){ 
	if (DEBUG) System.out.println(this + " adding child " + r);
	children.addElement(r);
	r.setParent(this);
    } 
    //returns true if successful
    public boolean removeChild(RSTTree t){
	return children.remove(t);
    }

    //why is this needed? to add ?rest child; now add RSTNode named ?rest
    //    public void addChild(String s){ children.addElement(s);} 
    
    /*    public void setParent(RSTNode p){
	parent = p;
	if (DEBUG) System.out.println(this + " setting parent " + p);
    }
    public RSTNode getParent(){
	return parent;
    }
    */
    public int getNuclearStatus(){
	return nuclear;
    }

    public RSTNode getNucleus(){
	return myNucleus;
    }

    //Overrides same method in Predicate.java ASSUME that children are
    //ordered, at least for now.  If a rule has a child String "?rest"
    //then that child matches zero to n remaining children in the
    //other tree. ?rest is handled in Predicate.java version of this
    //method.
    static public boolean isUnifiable(RSTNode r1, RSTNode r2, Hashtable h){
	boolean 
	    nameMatch, 
	    argMatch = true,
	    childMatch = true;

	if (DEBUG) System.out.println("RSTNode.isUnifiable: " + r1 + r2 + h);
	nameMatch = subIsUnifiable(r1.name, r2.name, h);
	if (DEBUG) System.out.println("RSTNode.isUnifiable: nameMatch = " + nameMatch);
	if (! nameMatch) return false;

	int i;
	for(i = 0; i < argNUM; i++){
	    if (r1.args[i] instanceof String &&
		((String)r1.args[i]).equalsIgnoreCase("?rest")){
		System.out.println("RSTNode.isUnifiable: returning true");
		return true; 
	    }
	    else if (r2.args[i] instanceof String &&
		     ((String)r2.args[i]).equalsIgnoreCase("?rest")){
		System.out.println("RSTNode.isUnifiable: returning true");
		return true;
	    }
	    argMatch = argMatch && subIsUnifiable(r1.args[i], r2.args[i], h);
	    if (DEBUG) System.out.println("RSTNode.isUnifiable, args:  " 
			   + r1.args[i] + " " + r2.args[i] + " " + argMatch);
	    if (! argMatch) return false;
	}
	
	int numChildren = r1.children.size() > r2.children.size() ? r2.children.size() : r1.children.size();

	for(i = 0; i < numChildren; i++){

	    if (DEBUG) System.out.println("RSTNode.isUnifiable, starting children:  " 
					  + r1.children.elementAt(i) + " " + r2.children.elementAt(i));
	//if (r1.children.elementAt(i) instanceof String &&
	//	((String)r1.children.elementAt(i)).equalsIgnoreCase("?rest"))
	//	return true;
	//else if (r2.children.elementAt(i) instanceof String &&
	//	     ((String)r2.children.elementAt(i)).equalsIgnoreCase("?rest"))
	//	return true;

	    if (r1.children.elementAt(i) instanceof RSTTree &&
		((RSTTree)r1.children.elementAt(i)).getRoot().getNameWithoutID().equalsIgnoreCase("?rest"))
		return true;
	    else if (r2.children.elementAt(i) instanceof RSTTree &&
		((RSTTree)r2.children.elementAt(i)).getRoot().getNameWithoutID().equalsIgnoreCase("?rest"))
		return true;

	    childMatch = subIsUnifiable(r1.children.elementAt(i), r2.children.elementAt(i), h);
	    if (DEBUG) System.out.println("RSTNode.isUnifiable, children:  " 
					  + r1.children.elementAt(i) + " " + r2.children.elementAt(i) + 
					  " " + childMatch);
	    if (! childMatch) return false;
	}
	if (DEBUG) System.out.println("end of RSTNode.isUnifiable(), returning true");
	return true;
    }

    public Object clone(){
	return new RSTNode((Predicate)super.clone());
    }
    
    public static void main(String[] args){

	//Object a[] = {"i","?a","k","l"};
	//RSTNode p1 = new RSTNode("a", a);
	//System.out.println(p1);
	//Object b[] = {"?j", "?rest"};
	//RSTNode p2 = new RSTNode("?b", b);
	//System.out.println(p2);
	//p1.addChild(p2);
	//System.out.println(p1.getParent());
	//System.out.println(p2.getParent());
	//
	//
	//Object c[] = {"i","c","k","l"};
	//RSTNode p3 = new RSTNode("a", c);
	//System.out.println(p3);
	//Object d[] = {"d", "x", "y"};
	//RSTNode p4 = new RSTNode("d", d);
	//System.out.println(p4);
	//	 
	//p3.addChild(p4);
	//System.out.println(p3.getParent());
	//System.out.println(p4.getParent());
	//
	//String temp = p3.getString();
	//System.out.println("read in: " + readRSTNode(temp));
	
	//RSTNode p1 = readRSTNode("[Inform ?prop2 ]");
	//RSTNode p2 = readRSTNode("[Inform [AND70 This_investment_has_quite_low_debt_risk IR_The_industry_as_a_whole_is_currently_considered_very_stable_stop ]]");

	RSTNode r1 = new RSTNode("?rest");
	System.out.println(r1);
	
	RSTNode r2 = new RSTNode("spam");
	System.out.println(r2);

	System.out.println(isUnifiable(r1,r2));
	System.out.println(isUnifiable(r1,r2));
    }
} // RSTNode

/**
(Inform83 ?prop2 ) (Inform71 (AND70 This investment has quite low debt risk IR The industry as a whole is currently considered very stable stop ) )
RSTNode.isUnifiable: (Inform83 ?prop2 )(Inform71 (AND70 This investment has quite low debt risk IR The industry as a whole is currently considered very stable stop ) ){?act=BuyInvestment, ?prop2=(AND70 This investment has quite low debt risk IR The industry as a whole is currently considered very stable stop ), ?prop=This investment has quite low debt risk, ?MotivateAct=MOTIVATE}
RSTNode.isUnifiable: nameMatch = true
**/
