// Evan Steinberger
// CISC 466
// Spring 2007

/**
 * ReutersGrabber pulls data from Reuters' financial web site(s).
 * 
 * @author Evan Steinberger
 */

public class ReutersGrabber extends StockInfoGrabber
{
	private static final String address =
		"http://www.investor.reuters.wallst.com/stocks/Ratios.asp?rpc=66&ticker=";

	private String crSearch = "Current Ratio (MRQ)";
	private String peSearch = "P/E Ratio (TTM)";
	//private String peHighLast5Search = "P/E High - Last 5 Yrs.";
	//private String peLowLast5Search = "P/E Low - Last 5 Yrs.";
	//private String deLongTermSearch = "LT Debt to Equity (MRQ)";
	private String deTotalSearch = "Total Debt to Equity (MRQ)";
	private String reTTMSearch = "Return On Equity (TTM)";
	//private String re5yrAvgSearch = "Return On Equity - 5 Yr. Avg.";
	private String endingDelim = "</span>";

	/**
	 * Constructor
	 * 
	 * @param stockSymbol	the stock ticker symbol
	 * @param at	the Access Threshold for pulling data
	 */
	public ReutersGrabber(String stockSymbol, AccessThreshold at)
	{
		super.currentAccessThreshold = at;
		super.init(stockSymbol);
		this.validateSymbol();
	}

	/**
	 * Ensures that the selected symbol is a valid.
	 * 
	 */
	protected void validateSymbol()
	{
		super.downloadHTML(this.buildURL(address));
		if (super.pageHTML.startsWith(WebQuery.errorMessageStart))
		{
			super.mineErrorMsg = "Error: URL is not valid.";
			super.validSymbol = false;
		}
		else
		{
			if (super.pageHTML.contains("SECTOR:&nbsp;"))
				super.validSymbol = true;
			else
				super.validSymbol = false;
		}
	}

	/**
	 * Builds a URL based on a prefix, stock symbol, and any
	 * other additional information.
	 * 
	 * @param s additional information placed into the URL
	 * 
	 */
	protected String buildURL(String s)
	{
		return s + super.stockSymbol;
	}

	/**
	 * Mines all relevant information.
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	public boolean mineAllInformation()
	{
		if (super.hasHistoricalContent)
			return true;
		if (!super.validSymbol)
		{
			if (super.mineErrorMsg == "")
				super.mineErrorMsg = "Error: Invalid Symbol Provided to Data Miner.";
			return false;
		}
		return this.mineCurrentRatio() && this.minePeTTM() && this.mineDeTotal() &&
			this.mineReTTM() && this.mineIndustry();
	}

	/**
	 * Mines the Current Ratio
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineCurrentRatio()
	{
		String temp = this.mineSingleData(this.crSearch + this.endingDelim, ">");
		if (temp.equals("N/A"))
			return true;
		try { super.currentRatio = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining Current Ratio: " + ex.toString();
			return false;
		}
		return true;
	}

	/**
	 * Mines Price/Earnings ratio
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean minePeTTM()
	{
		String temp = this.mineSingleData(this.peSearch + this.endingDelim, ">");
		if (temp.equals("N/A"))
			return true;
		try { super.peTTM = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining P/E TTM: " + ex.toString();
			return false;
		}
		return true;
	}

	/**
	 * Mines the Debt/Equity Ratio.
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineDeTotal()
	{
		String temp = this.mineSingleData(this.deTotalSearch + this.endingDelim, ">");
		if (temp.equals("N/A"))
			return true;
		try { super.deTotal = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining DE Total: " + ex.toString();
			return false;
		}
		return true;
	}

	/**
	 * Mines Return on Equity
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineReTTM()
	{
		String temp = this.mineSingleData(this.reTTMSearch + this.endingDelim, ">");
		if (temp.equals("N/A"))
			return true;
		try { super.reTTM = Double.parseDouble(temp); }
		catch (Exception ex)
		{
			super.mineErrorMsg = "Error: Problem mining RE TTM: " + ex.toString();
			return false;
		}
		return true;
	}
	
	/**
	 * Mines the Industry.
	 * 
	 * @return true if successful, false if not
	 * 
	 */
	protected boolean mineIndustry()
	{
		String temp = this.mineSingleData("INDUSTRY:&nbsp;" + this.endingDelim, "center\">");
		if (temp == "")
		{
			super.mineErrorMsg = "Error: Problem mining Industry";
			return false;
		}
		else
		{
			super.industry = temp;
			return true;
		}
	}
	
	/**
	 * Mines piece of single data
	 * 
	 * @param searchFor	a string to search for
	 * @param beginParser a string where the parsing begins
	 * 
	 * @return	a string of parsed data
	 * 
	 */
	private String mineSingleData(String searchFor, String beginParser)
	{
		try
		{
			String temp="";
			int start = super.pageHTML.indexOf(searchFor);
			int end;
			if (start < 0)
				return "";
			temp = pageHTML.substring(start);
			temp = temp.substring(searchFor.length(), temp.length()).trim();
			start = temp.indexOf(beginParser);
			end = temp.indexOf("</");
			if (start < 0 || end < 0)
				return "";
			temp = temp.substring(start+beginParser.length(), end);
			return temp;
		}
		catch (Exception ex) { return ""; }
	}
}