import java.io.*;
import java.util.*;
/**
 * Rule.java
 *
 * Created: Dec 10 20:12:01 2003
 *
 * It would be nice to be able to mark the rule trees with adjoin
 * points, and then mark the trees being unified at the relevant
 * portions as they are being traversed; then simply make one point
 * the child of another. Could rewrite tree unify to do the marking
 * after unifiability has been determined.
 */
public abstract class Rule {


    static final boolean DEBUG = false;
    static String em = "Error in Rule.java: ";
    static String wm = "Warning in Rule.java: ";
    static String loc = "Rule.java: ";

    RSTTree structure; //a tree into which the matching trees will be added
    Vector attachPoints; //numbered points in the tree that correspond to patterns
    Vector patterns; // patterns against which to match actual RSTTrees
    int ruleNumber = 0;
    int ruleValue = 0;
    boolean worksUnderMixedDirectionTrees = true; //not true for concession, contrast
    boolean worksOnSubTrees = true;

    public Rule() {
	
    } // Rule constructor
    
    public int ruleNumber(){return ruleNumber;}
    public int ruleValue(){return ruleValue;}

    public abstract RSTTree apply(RSTTree t1, RSTTree t2, String agentName);
    public boolean worksOnSubTrees(){return worksOnSubTrees;}
    public boolean worksUnderMixedDirection(){return worksUnderMixedDirectionTrees;}

} // Rule

//b.find( new RSTNode((String)h.get("?MotivateAct"), (String)h.get("?act"), (String)h.get("?prop2")) );

/* see RuleSet.java
 * rule1: AND under simple motivation
 * rule2 Concession
 * rule3 Contrast
 * rule4 Joint under concession
 * rule5 JointUnderNEQ (elab) under motivation
 * rule6 Motivation under JOINT , for trees with diff topic
 * rule7 JointUnderNEQ (ELAB) under Motivation, for trees with same topic
 * rule8 Joint
 */
