import java.util.*;
/**
 * Rule1.java
 *
 *
 * Created: Wed Mar 24 17:36:21 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class Rule1 extends Rule {

    int ruleNumber = 1;
    int ruleValue = 1;

    public Rule1() {
	
    } // Rule1 constructor 
    public int ruleNumber(){return ruleNumber;}
    public int ruleValue(){return ruleValue;}

   
    /** RULE 1
     * This rule takes two simple trees that agree and conjoins
     * ?prop under MOT using AND. The trees must be the same in strength.
     */
    /*
(GOAL12 (DO9 BuyInvestment ) ) //first tree
    (RECOMMEND11 BuyInvestment )
    (PERSUADE10 (DO9 BuyInvestment ) )
        (MOTIVATE8 BuyInvestment too much short term debt )
            (INFORM7 too much short term debt )
(GOAL6 (DO3 BuyInvestment ) ) //second tree
    (RECOMMEND5 BuyInvestment )
    (PERSUADE4 (DO3 BuyInvestment ) )
        (MOTIVATE2 BuyInvestment too much long term debt )
            (INFORM1 too much long term debt )
(Goal48 (DO44 BuyInvestment ) ) //new combined tree
    (Recommend47 BuyInvestment )
    (Persuade45 (DO44 BuyInvestment ) )
	(MOTIVATE2 BuyInvestment (AND too much long term debt too much short term debt))
                (INFORM1 (AND too much long term debt too much short term debt))
     **/
    public RSTTree apply(RSTTree a, RSTTree b, String agentName){

	///check Strength numbers first - its fast!
	Strength sa = a.getTotalStrength();
	Strength sb = b.getTotalStrength();

	//Two trees must be single topic, same topic:
	if (a.isMixedTopic() || b.isMixedTopic()) return null;
	if (! a.getTopic().equals(b.getTopic())) return null;
	//Trees that will satisfy this rule are either all positive or
	//all negative, and both must be the same polarity.
	if (!  ((sa.direction() == sb.direction()) && (sa.direction() != 0) &&
		//(sa.posCurr() == sb.posCurr())))
		(sa.absCurr() == sb.absCurr())))
	    return null;

	RSTNode i,m,p,r,g,i2,m2,p2,r2,g2;
	Predicate d,d2;
	RSTTree template1,template2;

	if (sa.direction() > 0){ //positive trees
	    //make these static!
	    i = new RSTNode("Inform", "?prop");
	    m = new RSTNode("?MotivateAct", "?act", "?prop");
	    m.addChild( new RSTTree(i));
	    d = new Predicate("DO", "?act");
	    p = new RSTNode("Persuade", d);
	    p.addChild( new RSTTree(m));
	    r = new RSTNode("Recommend", "?act");
	    g = new RSTNode("Goal", d);
	    g.addChild( new RSTTree(r));
	    g.addChild( new RSTTree(p));
	    template1 = new RSTTree(g);

	    i2 = new RSTNode("Inform", "?prop2");
	    m2 = new RSTNode("?MotivateAct", "?act", "?prop2");
	    m2.addChild( new RSTTree(i2));
	    d2 = new Predicate("DO", "?act");
	    p2 = new RSTNode("Persuade", d2);
	    p2.addChild( new RSTTree(m2));
	    r2 = new RSTNode("Recommend", "?act");
	    g2 = new RSTNode("Goal", d2);
	    g2.addChild( new RSTTree(r2));
	    g2.addChild( new RSTTree(p2));
	    template2 = new RSTTree(g2);
	}
	else {
	    i = new RSTNode("Inform", "?prop");
	    m = new RSTNode("?DemotivateAct", "?act", "?prop");
	    m.addChild( new RSTTree(i));
	    d = new Predicate("DO", "?act");
	    p = new RSTNode("Dissuade", d);
	    p.addChild( new RSTTree(m));
	    r = new RSTNode("Recommend", new Predicate("NOT", d));
	    g = new RSTNode("Goal", new Predicate("NOT", d));
	    g.addChild( new RSTTree(r));
	    g.addChild( new RSTTree(p));
	    template1 = new RSTTree(g);

	    i2 = new RSTNode("Inform", "?prop2");
	    m2 = new RSTNode("?DemotivateAct", "?act", "?prop2");
	    m2.addChild( new RSTTree(i2));
	    d2 = new Predicate("DO", "?act");
	    p2 = new RSTNode("Dissuade", d2);
	    p2.addChild( new RSTTree(m2));
	    r2 = new RSTNode("Recommend", new Predicate("NOT", d2));
	    g2 = new RSTNode("Goal", new Predicate("NOT", d2));
	    g2.addChild( new RSTTree(r2));
	    g2.addChild( new RSTTree(p2));
	    template2 = new RSTTree(g2);
	}

	
	if (DEBUG){
	    System.out.println("r1 two pattern trees: ");
	    template1.print();
	    template2.print();}
	
	patterns = new Vector(2);
	patterns.addElement(template1);
	patterns.addElement(template2);

	Hashtable h = new Hashtable();

	int flag = 0;
	if (RSTTree.isUnifiable(template1, a, h)){
	    if (DEBUG) System.out.println("Rule1 template1 unifies with tree a");
	    if (RSTTree.isUnifiable(template2, b, h)){
		if (DEBUG) System.out.println("Rule1 template2 unifies with tree b");
		flag = 1;
	    }
	}
	if (flag == 0) return null;

	//create new superstructure with correct instantiations
	RSTNode i3 = null;
	RSTNode g3 = null;
	if (sa.direction() > 0){ //positive trees
	    Predicate and = new Predicate("AND", h.get("?prop"), h.get("?prop2"));
	    i3 = new RSTNode("Inform", and);
	    Predicate d3 = new Predicate("DO", h.get("?act"));
	    RSTNode p3 = new RSTNode("Persuade", d3);
	    RSTNode m3 = new RSTNode((String)h.get("?MotivateAct"), (String)h.get("?act"), and);
	    p3.addChild( new RSTTree(m3));
	    m3.addChild( new RSTTree(i3));
	    RSTNode r3 = new RSTNode("Recommend", h.get("?act"));
	    g3 = new RSTNode("Goal", d3);
	    g3.addChild( new RSTTree(r3));
	    g3.addChild( new RSTTree(p3));
	}
	else{ //negative trees
	    Predicate and = new Predicate("AND", h.get("?prop"), h.get("?prop2"));
	    i3 = new RSTNode("Inform", and);
	    RSTNode m3 = new RSTNode((String)h.get("?DemotivateAct"), (String)h.get("?act"), and);
	    m3.addChild( new RSTTree(i3));
	    Predicate d3 = new Predicate("DO", h.get("?act"));
	    RSTNode p3 = new RSTNode("Dissuade", d3);
	    p3.addChild( new RSTTree(m3));
	    RSTNode r3 = new RSTNode("Recommend", new Predicate("NOT", d3));
	    g3 = new RSTNode("Goal", d3);
	    g3.addChild( new RSTTree(r3));
	    g3.addChild( new RSTTree(p3));
	}
	RSTTree newTree = new RSTTree(g3);
	//newTree = RuleSet.makeNewTreeFromTwo(a, b, g3, ruleNumber(), ruleValue(), agentName);

	BidSet bs = new BidSet(agentName);
	bs.add(a.getBid());
	bs.add(b.getBid());
	
	i3.getTree().setBid( bs.makeOneBidFromBidSet(agentName) );
	
	return newTree;
    }// apply(RSTTree a, RSTTree b)

    public static void main(String[] args){
	Bid b1 = new Bid();
	Bid b2 = new Bid();

	RSTTree x  = RSTTree.GoalDoNot("goaldonot", DSType.RISK, -3, b1);
	RSTTree y = RSTTree.GoalDoNot("goaldonot1", DSType.RISK, -3, b2);

	Rule r1 = new Rule1();
	RSTTree z = r1.apply(x, y,  "spam");
	z.print();
    }

} // Rule1
