import java.util.*;
/**
 * Rule4.java
 *
 * For introducing a Joint under one side of a Concession
 * Concession must be more important tree (Integrator does this).
 * Identify Concession; identify 2nd tree as single non-zero direction.
 * Make 2nd tree joint with Satellite side of Concession (under what circumstances? must be less important than sat?)
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class Rule4 extends Rule {

    int ruleNumber = 4;
    int ruleValue = 1;
    boolean worksUnderMixedDirectionTrees = true;

    public Rule4() {
	
    } // Rule4 constructor
    
    public boolean worksUnderMixedDirection(){return worksUnderMixedDirectionTrees;}

    public int ruleNumber(){return ruleNumber;}
    public int ruleValue(){return ruleValue;}

    public RSTTree apply(RSTTree a, RSTTree b, String agentName){

	//a(don't act), but concede b
	Strength sa = a.getTotalStrength();
	Strength sb = b.getTotalStrength();
	if (DEBUG) System.out.println(" " + sa.direction() + " " + sb.direction() + " sa:" + sa + " sb:" + sb);

	if ( ! ((sa.direction() != sb.direction()) &&
		(sa.direction() != 0) &&
		(sb.direction() != 0) &&
		( Math.abs(sa.absCurr() - sb.absCurr()) < 2) &&
		sa.absCurr() != 0 && //redundant?
		sb.absCurr() != 0 )){
	    if (DEBUG) System.out.println("rule3 failed, direction & strength");
	    return null;
	}

	//trees must have different direction, not same curr strength
	if ( sa.direction() == sb.direction() ){
	    if (DEBUG) System.out.println("rule4 failed, direction");
	     //(((sa.direction()==1 && sb.direction()==-1 ) && (sa.posCurr()- sb.negCurr()  >= 1)) ||
	     //   (sa.direction()==-1 && sb.direction()==1 )  && (sa.negCurr() - sb.posCurr() >= 1)))
	    return null;
	}

	String type = null;
	if ( Math.abs(sa.absCurr() - sb.absCurr()) > 2)
	    type = "Concession";
	else type = "Contrast";

	RSTTree posTree = sa.direction() == 1 ? a : b;
	RSTTree negTree = sa.direction() == 1 ? b : a;

	RSTNode r1 = new RSTNode("?rest");
	RSTNode g = new RSTNode("Concession");
	g.addChild( new RSTTree(r1));
	RSTTree template1 = new RSTTree(g);

	if (DEBUG) System.out.println("GOT HERE2!");
	
	Hashtable h = new Hashtable();
	int flag = 0;

	if (! RSTTree.isUnifiable(template1, a, h)){
	    if (DEBUG) System.out.println("Rule4 templates do not unify");
	    return null;       
	}

	//get children of concession, figure out which is satellite;


	//create superstructure with correct instantiations

	RSTNode g3 = new RSTNode(type);
	RSTTree aPoint = null, bPoint = null;

	

	//newTree = RuleSet.makeNewTreeFromTwo(posTree, negTree, g3, ruleNumber(), ruleValue(), agentName);

	if (posTree.getTotalStrength().absCurr() > negTree.getTotalStrength().absCurr()){
	    
	    aPoint = (RSTTree)posTree.clone();
	    bPoint = (RSTTree)negTree.clone();
	}
	else{
	    aPoint = (RSTTree)negTree.clone();
	    bPoint = (RSTTree)posTree.clone();
	}
	if (DEBUG) System.out.println("two attach points are : " + aPoint + " " + bPoint);

	g3.addChild(aPoint);
	g3.addChild(bPoint);
	RSTTree newTree = new RSTTree(g3);

	DebugFile df = new DebugFile("rule4output");
	df.write("tree A: " + aPoint.getRoot() + " " + aPoint.getBid());
	df.write("________________\n");
	df.write("tree B: " + bPoint.getRoot() + " " + bPoint.getBid());
	df.write("________________\n");
	
	df.write("\n" + newTree.getString() + "\n" );
	df.close();
	
	return newTree;
    }// apply(RSTTree a, RSTTree b)

    public static void main(String[] args){
	Bid b1 = new Bid();
	Bid b2 = new Bid();

	RSTTree x  = RSTTree.GoalDoNot("goaldonot", DSType.RISK, -4, b1);
	RSTTree y = RSTTree.GoalDo("goaldo1", DSType.RISK, 2, b2);

	RSTNode r2 = new RSTNode("?rest");
	Predicate d2 = new Predicate("DO", "?act");
	RSTNode g2 = new RSTNode("Goal", d2);
	g2.addChild( new RSTTree(r2));
	RSTTree template2 = new RSTTree(g2);
	template2.print();

	
	Hashtable h = new Hashtable();

	if (RSTTree.isUnifiable(template2, y, h))
	    System.out.println("yes");
	else
	    System.out.println("no");
	
	Rule r4 = new Rule4();
	RSTTree z = r4.apply(x, y,  "spam");
	z.print();
    }
}
