import java.util.*;
/**
 * Rule5.java
 *
 *
 * Created: Wed Mar 24 17:36:21 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class Rule5 extends Rule {

    int ruleNumber = 5;
    int ruleValue = 1;

    public Rule5() {
	
    } // Rule5 constructor 
    public int ruleNumber(){return ruleNumber;}
    public int ruleValue(){return ruleValue;}

    /**
     * ELABORATION under Motivation, for trees with same topic, with
     * Strength that is not identical (if not same topic should be
     * joint; is strength is identical should be rule1.
     **/
    public RSTTree apply(RSTTree a, RSTTree b, String agentName){

	//for elab between motivates, must be same topic
	if (! a.getTopic().equals(b.getTopic())){
	    if (DEBUG) System.out.println("Rule5 failed, not same topic");
	    return null;
	}

	///check Strength numbers first - its fast!
	Strength sa = a.getTotalStrength();
	Strength sb = b.getTotalStrength();

	//Trees that will satisfy this rule are either all positive or
	//all negative, and both must be the same polarity.
	if (!  ((sa.direction() == sb.direction()) && (sa.direction() != 0) &&
		((sa.absCurr() - sb.absCurr()) != 0)))
	    {
	    if (DEBUG) System.out.println("Rule5 failed, direction and strength");
	    return null;
	}
	
	RSTTree template1 = null;
	RSTTree template2 = null;

	if (sa.direction() == 1){
	    RSTNode i = new RSTNode("Inform", "?prop");
	    RSTNode m = new RSTNode("?MotivateAct", "?act", "?prop");
	    m.addChild( new RSTTree(i));
	    Predicate d = new Predicate("DO", "?act");
	    RSTNode p = new RSTNode("Persuade", d);
	    p.addChild( new RSTTree(m));
	    RSTNode r = new RSTNode("Recommend", "?act");
	    RSTNode g = new RSTNode("Goal", d);
	    g.addChild( new RSTTree(r));
	    g.addChild( new RSTTree(p));
	    template1 = new RSTTree(g);
	    
	    RSTNode i2 = new RSTNode("Inform", "?prop2");
	    RSTNode m2 = new RSTNode("?MotivateAct", "?act", "?prop2");
	    m2.addChild( new RSTTree(i2));
	    Predicate d2 = new Predicate("DO", "?act");
	    RSTNode p2 = new RSTNode("Persuade", d2);
	    p2.addChild( new RSTTree(m2));
	    RSTNode r2 = new RSTNode("Recommend", "?act");
	    RSTNode g2 = new RSTNode("Goal", d2);
	    g2.addChild( new RSTTree(r2));
	    g2.addChild( new RSTTree(p2));
	    template2 = new RSTTree(g2);
	}
	else {
	    
	    RSTNode i = new RSTNode("Inform", "?prop");
	    RSTNode m = new RSTNode("DemotivateAct", "?act", "?prop");
	    m.addChild( new RSTTree(i));
	    Predicate d = new Predicate("DO", "?act");
	    RSTNode p = new RSTNode("Dissuade", d);
	    p.addChild( new RSTTree(m));
	    RSTNode r = new RSTNode("Recommend", new Predicate("NOT", d));
	    RSTNode g = new RSTNode("Goal", new Predicate("NOT", d));
	    g.addChild( new RSTTree(r));
	    g.addChild( new RSTTree(p));
	    template1 = new RSTTree(g);

	    RSTNode i2 = new RSTNode("Inform", "?prop2");
	    RSTNode m2 = new RSTNode("DemotivateAct", "?act", "?prop2");
	    m2.addChild( new RSTTree(i2));
	    Predicate d2 = new Predicate("DO", "?act");
	    RSTNode p2 = new RSTNode("Dissuade", d2);
	    p2.addChild( new RSTTree(m2));
	    RSTNode r2 = new RSTNode("Recommend", new Predicate("NOT", d2));
	    RSTNode g2 = new RSTNode("Goal", new Predicate("NOT", d2));
	    g2.addChild( new RSTTree(r2));
	    g2.addChild( new RSTTree(p2));
	    template2 = new RSTTree(g2);

	}
	
	if (DEBUG){
	    System.out.println("R5: two pattern trees: ");
	    template1.print();
	    template2.print();}
	
	patterns = new Vector(2);
	patterns.addElement(template1);
	patterns.addElement(template2);

	Hashtable h = new Hashtable();

	int flag = 0;
	if (RSTTree.isUnifiable(template1, a, h)){
	    if (DEBUG) System.out.println("Rule5 template1 unifies with tree a");
	    	if (RSTTree.isUnifiable(template2, b, h)){
		    if (DEBUG) System.out.println("Rule1 template2 unifies with tree b");
		    flag = 1;
		}
	}

	if (flag == 0) return null;

	RSTTree aInform = a.find(new RSTNode("Inform", "?rest")).getTree();
	RSTTree bInform = b.find(new RSTNode("Inform", "?rest")).getTree();

	//create new superstructure with correct instantiations
	RSTNode g3 = null;
	if (sa.direction() == 1){
	    //	Predicate and = new Predicate("AND", h.get("?prop"), h.get("?prop2"));
	    RSTNode i3 = new RSTNode("Inform", h.get("?prop"));
	    RSTNode i4 = new RSTNode("Inform", h.get("?prop2"));
	    RSTTree i3Tree = new RSTTree(i3);
	    i3Tree.setBid( (Bid)aInform.getBid().clone() );
	    RSTTree i4Tree = new RSTTree(i4);
	    i4Tree.setBid( (Bid)bInform.getBid().clone() );
	    RSTNode e1 = new RSTNode("Elaborate");
	    e1.addChild(i3Tree );
	    e1.addChild(i4Tree );
	    Predicate d3 = new Predicate("DO", h.get("?act"));
	    RSTNode m3 = new RSTNode((String)h.get("?MotivateAct"), (String)h.get("?act"), e1);
	    m3.addChild( new RSTTree(e1));
	    RSTNode p3 = new RSTNode("Persuade", d3);
	    p3.addChild( new RSTTree(m3));
	    RSTNode r3 = new RSTNode("Recommend", h.get("?act"));
	    g3 = new RSTNode("Goal", d3);
	    g3.addChild( new RSTTree(r3));
	    g3.addChild( new RSTTree(p3));
	}
	else {
	    RSTNode i3 = new RSTNode("Inform", h.get("?prop"));
	    RSTNode i4 = new RSTNode("Inform", h.get("?prop2"));
	    RSTTree i3Tree = new RSTTree(i3);
	    i3Tree.setBid( (Bid)aInform.getBid().clone() );
	    RSTTree i4Tree = new RSTTree(i4);
	    i4Tree.setBid( (Bid)bInform.getBid().clone() );
	    RSTNode e1 = new RSTNode("Elaborate");
	    e1.addChild(i3Tree );
	    e1.addChild(i4Tree );
	    RSTNode m3 = new RSTNode("DemotivateAct", h.get("?act"), e1);
	    m3.addChild( new RSTTree(e1));
	    Predicate d3 = new Predicate("DO", h.get("?act"));
	    RSTNode p3 = new RSTNode("Dissuade", d3);
	    p3.addChild( new RSTTree(m3));
	    RSTNode r3 = new RSTNode("Recommend", new Predicate("NOT", d3));
	    g3 = new RSTNode("Goal", new Predicate("NOT", d3));
	    g3.addChild( new RSTTree(r3));
	    g3.addChild( new RSTTree(p3));
	}
	RSTTree newTree = new RSTTree(g3);
	//newTree = RuleSet.makeNewTreeFromTwo(a, b, g3, ruleNumber(), ruleValue(), agentName);
	return newTree;
    }// apply(RSTTree a, RSTTree b) ELABORATION

    public static void main(String[] args){
	Bid b1 = new Bid();
	Bid b2 = new Bid();

	RSTTree x  = RSTTree.GoalDoNot("goaldo", DSType.RISK, -3, b1);
	RSTTree y = RSTTree.GoalDoNot("goaldo1", DSType.RISK, -2, b2);

	Rule r5 = new Rule5();
	RSTTree z = r5.apply(x, y,  "spam");
	RSTTree newTree = RuleSet.makeNewTreeFromTwo(x, y, z.getRoot(), 5, 17, "asdf");
	z.print();
	newTree.print();
    }
}
