import java.util.*;
/**
 * Rule7.java
 *
 *
 * Created: Wed Mar 24 17:36:21 2004
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class Rule7 extends Rule {

    int ruleNumber = 7;
    int ruleValue = 1;

    public Rule7() {
	
    } // Rule7 constructor 
    public int ruleNumber(){return ruleNumber;}
    public int ruleValue(){return ruleValue;}

    

    /**
     * ELABORATION under Motivation, for trees with same topic, with
     * Strength that is not identical (if not same topic should be
     * joint; if strength is identical should be rule1.
     **/
    public RSTTree apply(RSTTree a, RSTTree b, String agentName){

	//for elab between motivates, must be same topic
	if (! a.getTopic().equals(b.getTopic())){
	    if (DEBUG) System.out.println("Rule7 failed, not same topic");
	    return null;
	}

	///check Strength numbers first - its fast!
	Strength sa = a.getTotalStrength();
	Strength sb = b.getTotalStrength();

	//Trees that will satisfy this rule are either all positive or
	//all negative, and both must be the same polarity.
	if (!  ((sa.direction() == sb.direction()) && (sa.direction() != 0) &&
		((sa.absCurr() - sb.absCurr()) != 0)))
	    {
	    if (DEBUG) System.out.println("Rule7 failed, direction and strength" + sa + sb);
	    return null;
	}
	
	RSTTree template1 = null;
	RSTTree template2 = null;

	if (sa.direction() == 1){
	    RSTNode e = new RSTNode("Elaborate", "?rest");
	    RSTNode m = new RSTNode("?MotivateAct", "?act", "?prop");
	    m.addChild( new RSTTree(e));
	    Predicate d = new Predicate("DO", "?act");
	    RSTNode p = new RSTNode("Persuade", d);
	    p.addChild( new RSTTree(m));
	    RSTNode r = new RSTNode("Recommend", "?act");
	    RSTNode g = new RSTNode("Goal", d);
	    g.addChild( new RSTTree(r));
	    g.addChild( new RSTTree(p));
	    template1 = new RSTTree(g);
	    
	    RSTNode i2 = new RSTNode("Inform", "?prop2");
	    RSTNode m2 = new RSTNode("?MotivateAct", "?act", "?prop2");
	    m2.addChild( new RSTTree(i2));
	    Predicate d2 = new Predicate("DO", "?act");
	    RSTNode p2 = new RSTNode("Persuade", d2);
	    p2.addChild( new RSTTree(m2));
	    RSTNode r2 = new RSTNode("Recommend", "?act");
	    RSTNode g2 = new RSTNode("Goal", d2);
	    g2.addChild( new RSTTree(r2));
	    g2.addChild( new RSTTree(p2));
	    template2 = new RSTTree(g2);
	}
	else {
	    
	    RSTNode e = new RSTNode("Elaborate", "?rest");
	    RSTNode m = new RSTNode("DemotivateAct", "?act", "?prop");
	    m.addChild( new RSTTree(e));
	    Predicate d = new Predicate("DO", "?act");
	    RSTNode p = new RSTNode("Dissuade", d);
	    p.addChild( new RSTTree(m));
	    RSTNode r = new RSTNode("Recommend", new Predicate("NOT", d));
	    RSTNode g = new RSTNode("Goal", new Predicate("NOT", d));
	    g.addChild( new RSTTree(r));
	    g.addChild( new RSTTree(p));
	    template1 = new RSTTree(g);

	    RSTNode i2 = new RSTNode("Inform", "?prop2");
	    RSTNode m2 = new RSTNode("DemotivateAct", "?act", "?prop2");
	    m2.addChild( new RSTTree(i2));
	    Predicate d2 = new Predicate("DO", "?act");
	    RSTNode p2 = new RSTNode("Dissuade", d2);
	    p2.addChild( new RSTTree(m2));
	    RSTNode r2 = new RSTNode("Recommend", new Predicate("NOT", d2));
	    RSTNode g2 = new RSTNode("Goal", new Predicate("NOT", d2));
	    g2.addChild( new RSTTree(r2));
	    g2.addChild( new RSTTree(p2));
	    template2 = new RSTTree(g2);

	}
	
	if (DEBUG){
	    System.out.println("7 two pattern trees: ");
	    template1.print();
	    template2.print();}
	
	patterns = new Vector(2);
	patterns.addElement(template1);
	patterns.addElement(template2);

	Hashtable h = new Hashtable();

	int flag = 0;
	if (RSTTree.isUnifiable(template1, a, h)){
	    if (DEBUG) System.out.println("Rule7 template1 unifies with tree a");
	    if (DEBUG) a.print();
	    	if (RSTTree.isUnifiable(template2, b, h)){
		    if (DEBUG) System.out.println("Rule7 template2 unifies with tree b");
		    if (DEBUG) b.print();
		    flag = 1;
		}
	}

	if (flag == 0) return null;
	

	//create new superstructure with correct instantiations
	RSTTree g3 = null;
	//if (sa.direction() == 1){

	g3 = (RSTTree)a.clone();
	RSTNode g3Elab = g3.find(new RSTNode("Elaborate", "?rest"));
	System.out.println("g3Elab: " + g3Elab);
	//b.print();
	System.out.println("node from b: " + b.find(new RSTNode("Inform", "?rest")) );
	
	RSTTree bInform = b.find(new RSTNode("Inform", "?rest")).getTree();
	//System.out.println("tree: " + bInform);
	
	g3Elab.addChild((RSTTree)bInform.clone());

	
	//newTree = RuleSet.makeNewTreeFromTwo(a, b, g3.getRoot(), ruleNumber(), ruleValue(), agentName);

	//System.out.println("rule 7 final new ");
	///newTree.print();
	return g3;
    }// apply(RSTTree a, RSTTree b) ELABORATION

    public static void main(String[] args){
	Bid b1 = new Bid();
	Bid b2 = new Bid();
	Bid b3 = new Bid();

	RSTTree w  = RSTTree.GoalDo("goaldo3", DSType.RISK, 1, b3);
	RSTTree x  = RSTTree.GoalDo("goaldo1", DSType.RISK, 3, b1);
	RSTTree y = RSTTree.GoalDo("goaldo2", DSType.RISK, 2, b2);

	Rule r5 = new Rule5();
	RSTTree z = r5.apply(x, y,  "spam");
	z.print();
	w.print();

	Rule r7 = new Rule7();
	RSTTree z2 = r7.apply(z, w,  "spam");
	z2.print();
    }
}
