import java.util.*;
import java.text.NumberFormat;
/**
 * Strength.java
 *
 *
 * Created: Tue Jan 4 17:55:12 2004
 *
 * Format [[currPosStrength, posSum, posCount][[currNegStrength, negSum, negCount]]
 *
 * A given RSTTree has a Bid containing (at least) four Strengths: R,V,G,DStot
 *
 * Note: an empty constructor "new Strength()" is not the same as "new Strength(0)". The latter has count = 1.
 *
 * @author <a href="mailto:terry@Terry-Harveys-Computer.local.">Terry Harvey</a>
 * @version 1.0
 */
public class Strength {

    static final boolean DEBUG = false;
    static final String loc = "in Strength ";
    static final String em = "ERROR in Strength ";
    
    static NumberFormat nf = NumberFormat.getInstance();

    Vector data = new Vector(6);

    public Strength() {
	nf.setMaximumFractionDigits(2);
	for (int i = 0; i<6; i++)
	    data.add(i, new Double(0));
    } // Strength constructor

    public Strength(double i) {
	nf.setMaximumFractionDigits(2);
	for (int j = 0; j<6; j++)
	    data.add(j, new Double(0));
	if (i >= 0){
	    data.setElementAt(new Double(i), 0);
	    data.setElementAt(new Double(i), 1);
	    setPosCount(1);
	}
	else if (i < 0){
	    data.setElementAt(new Double(-i), 3);
	    data.setElementAt(new Double(-i), 4);
	    setNegCount(1);
	}
    } // Strength constructor
    
    public Strength(Double i) {
	nf.setMaximumFractionDigits(2);
	for (int j = 0; j<6; j++)
	    data.add(j, new Double(0));
	if (i.doubleValue() >= 0){
	    data.setElementAt(i, 0);
	    data.setElementAt(i, 1);
	    setPosCount(1);
	}
	else if (i.doubleValue() < 0){
	    i = new Double( Math.abs(i.doubleValue()) );
	    data.setElementAt(i, 3);
	    data.setElementAt(i, 4);
	    setNegCount(1);
	}
    } // Strength constructor
    
    public double posCurr() {return ((Double)data.get(0)).doubleValue();}
    public double posSum()  {return ((Double)data.get(1)).doubleValue();}
    public double posCount(){return ((Double)data.get(2)).doubleValue();}
    public double negCurr() {return ((Double)data.get(3)).doubleValue();}
    public double negSum()  {return ((Double)data.get(4)).doubleValue();}
    public double negCount(){return ((Double)data.get(5)).doubleValue();}

    private void setPosCurr(double i) {data.setElementAt(new Double(i), 0);}
    private void setPosSum(double i)  {data.setElementAt(new Double(i), 1);}
    private void setPosCount(double i){data.setElementAt(new Double(i), 2);}
    private void setNegCurr(double i) {data.setElementAt(new Double(i), 3);}
    private void setNegSum(double i)  {data.setElementAt(new Double(i), 4);}
    private void setNegCount(double i){data.setElementAt(new Double(i), 5);}

    public boolean equals(Object o){
	if (o instanceof Strength)
	    return this.toString().equals(o.toString());
	else return false;
    }

    //Returns 1 if only Strength is positive; -1 if only Strength is negative; and 0 otherwise.
    public int direction(){
	if (DEBUG)
	    System.out.println("nc, pc: " + negCount() + " " +  posCount());
	
	if (negCount() == posCount()) return 0;
	else if (negCount() == 0) return 1;
	else if (posCount() == 0) return -1;
	else return 0;
    }

    //Combine two Strengths; Curr becomes MAX of two Currs; Count and Sum are summed.
    public static Strength merge(Strength s1, Strength s2){
	if (DEBUG) System.out.println(loc + ".merge: s1 " + s1 + " s2 " + s2);
	
	Strength result = new Strength();
	result.setPosCurr( s1.posCurr() > s2.posCurr() ? s1.posCurr() : s2.posCurr() );
	result.setPosSum( s1.posSum() + s2.posSum());
	result.setPosCount( s1.posCount() + s2.posCount());
	result.setNegCurr( s1.negCurr() > s2.negCurr() ? s1.negCurr() : s2.negCurr() );
	result.setNegSum( s1.negSum() + s2.negSum());
	result.setNegCount( s1.negCount() + s2.negCount());
	if (DEBUG) System.out.println(loc + ".merge: result " + result);
	return result;
    }

    //Merge two Strengths into first one
    public void add(Strength s2){
	if (DEBUG) System.out.println(loc + ".add: this " + this + " s2 " + s2);
	
	this.setPosCurr( this.posCurr() > s2.posCurr() ? this.posCurr() : s2.posCurr() );
	this.setPosSum( this.posSum() + s2.posSum());
	this.setPosCount( this.posCount() + s2.posCount());
	this.setNegCurr( this.negCurr() > s2.negCurr() ? this.negCurr() : s2.negCurr() );
	this.setNegSum( this.negSum() + s2.negSum());
	this.setNegCount( this.negCount() + s2.negCount());
	if (DEBUG) System.out.println(loc + ".add: result " + this);
    }

    //returns highest of posCurr, abs(negCurr).
    public String getDS(){
	return Double.toString( posCurr() > negCurr() ? posCurr() : negCurr() );
    }
    //returns highest of posCurr, abs(negCurr).
    public double absCurr(){
	return posCurr() > negCurr() ? posCurr() : negCurr();
    }
    public double getDSSumsForUtility(){
	return  posSum() + negSum();
    }
    //returns sum of posSum and negSum
//   public String getDSSums(){
//	return Double.toString( posSum() + negSum() );
//   }


    public String toString(){
	StringBuffer s = new StringBuffer("[[");
	s.append(nf.format(posCurr()) + " "); 
	s.append(nf.format(posSum())  + " "); 
	s.append(nf.format(posCount()));
	s.append("][");
	s.append(nf.format(negCurr())  + " ");
	s.append(nf.format(negSum())   + " ");
	s.append(nf.format(negCount()) + "]]");
	return s.toString();
    }

    public static Strength readStrength(String s){
	return readStrength(new StringTokenizer(s," \t\n\r\f[]", true ));
    }

    public static Strength readStrength(StringTokenizer st){
	Strength result = new Strength();
	String test;
	do {
	    test = st.nextToken("[], {}=");
	    if (DEBUG) System.out.println( "readStrength ignoring: " + test);
	} while(! test.equals("["));
	
	st.nextToken(); //second [
	result.setPosCurr( Double.parseDouble(st.nextToken()) );
	st.nextToken(); //space
	result.setPosSum(   Double.parseDouble(st.nextToken()) );
	st.nextToken(); //space
	result.setPosCount( Double.parseDouble(st.nextToken()) );
	st.nextToken(); //]
	st.nextToken(); //[
	result.setNegCurr(  Double.parseDouble(st.nextToken()) );
	st.nextToken(); //space
	result.setNegSum(   Double.parseDouble(st.nextToken()) );
	st.nextToken(); //space
	result.setNegCount( Double.parseDouble(st.nextToken()) );
	st.nextToken(); //]
	st.nextToken(); //]
	return result;
    }

    public static void main(String[] args){
	Strength s1 = new Strength(3);
	Strength s2 = new Strength(1);
	Strength s3 = Strength.merge(s1,s2);
	String temp = s1.toString();
	Strength s4 = readStrength(temp);
	System.out.println(s4.toString());

//	DataSet d1 = DataSet.readDataSetFromString("{bid={topic=r, length_exp=10, cost_exp=0, DSr_exp=[[1 1 1][0 0 0]], time_exp=1, DSg_exp=[[0 0 0][0 0 0]], DSv_exp=[[0 0 0][0 0 0]], RV=[[] 0], symbol=IBM, U=null, bidder=DR}, source=DR}");
//	double ds;
//	Bid b = (Bid)d1.get("bid");
//	Strength dsr = (Strength)b.get("DSr_exp");
//	if (dsr.direction() > 0)
//	    ds = dsr.posCurr();
//	else if (dsr.direction() < 0)
//	    ds = dsr.negCurr();
//	else
//	    System.out.println(em + "DataSet has mixed or no Strength: " + dsr );
//	
//	System.out.println("direction: " + dsr.direction());
	

    }

} // Strength
