// Evan Steinberger
// CISC 466
// Spring 2007

import java.net.*;
import java.net.URL;
import java.io.*;

/**
 * Base Clase for pulling information from the WWW
 * 
 * @author Evan Steinberger
 */

public class WebQuery
{
	public static final String errorMessageStart = "Error! getCompleteHTML request failed: ";
	
	/**
	 * Returns Raw HTML, for use quickly in a static way without
	 * having to construct an object
	 * 
	 * @param webPage the web site do download from
	 * @return the complete HTML from the web site
	 */
	public static String getCompleteHTML(String webPage)
	{
		try
		{
			System.out.print("Pulling data from " + webPage + " ... ");
			URL page = new URL(webPage);
			URLConnection conn = page.openConnection();
			BufferedReader br = new BufferedReader(new InputStreamReader(conn.getInputStream()));
			String inputLine, htmlString="";
			while ((inputLine = br.readLine()) != null)
				htmlString+=inputLine + "\r\n";
			br.close();
			System.out.println("done");
			return htmlString;
		}
		catch (Exception ex)
		{
			System.out.println("error");
			return errorMessageStart + ex.toString();
		}
	}

	// Members
	private String urlString;
	private String urlHTML;
	
	/**
	 * Constructor
	 * 
	 * @param urlString the URL to download from
	 */
	public WebQuery(String urlString)
	{
		this.urlString = urlString;
		this.urlHTML = getCompleteHTML(urlString);
	}
	
	/**
	 * Returns the complete HTML
	 * 
	 * @return the HTML that was downloaded
	 */
	public String getCompleteHTML()
	{
		return this.urlHTML;
	}
	
	/**
	 * Counts A string pattern in an HTML page
	 * 
	 * @param pattern the pattern to check for
	 * @return the number times the pattern occurs in the HTML page
	 */
	public int countPattern(String pattern) 
	{
		int origLength = pattern.length();
		int total=0;
		if (origLength > 0 && !urlHTML.startsWith("Error ! getCompleteHTML")) 
		{ 
			int placeHolder = urlHTML.indexOf(pattern);
			while (placeHolder != -1) 
			{
				total++;
				placeHolder = urlHTML.indexOf(pattern, placeHolder+origLength);
			}
		}
		System.out.println(total);
		return total;
	}
}
	