// Evan Steinberger
// CISC 466
// January 2007

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
import java.lang.reflect.*;

public class WebQueryFrame extends JFrame implements ActionListener
{
	// Constants 
	final private int xSize = 640;
	final private int ySize = 480;

	// GUI Components
	private JLabel l_info;
	private JTextField t_queryfield;
	private JTextField t_blindprofile;
	private JButton b_submit;
	private JTextArea t_result;
   
	public WebQueryFrame() 
	{
		this.setLayout(null);

		l_info = new JLabel("Enter Stock Symbol:");
		l_info.setBounds((xSize/2)-150, 30, 200, 20);
		this.add(l_info);
		
		t_queryfield = new JTextField();
		t_queryfield.setBounds((xSize/2)-150, 50, 200, 20);
		t_queryfield.setText("IBM");
		this.add(t_queryfield);

		t_blindprofile = new JTextField();
		t_blindprofile.setBounds((xSize / 2)+ 150, 25, 100, 20);
		this.add(t_blindprofile);

		b_submit = new JButton("Submit");
		b_submit.setBounds((xSize/2) + 50, 50, 100, 20);
		b_submit.setActionCommand("Submit");
		b_submit.addActionListener(this);
		this.add(b_submit);

		t_result = new JTextArea();
		//t_result.setLineWrap(true);
		t_result.setEditable(false);
		JScrollPane jsp = new JScrollPane(t_result);
		jsp.setBounds(5, 75, xSize-15, ySize-125);
		this.add(jsp);
		
		this.setTitle("Web Query Window, Evan Steinberger 2007");
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.setSize(xSize,ySize);
		this.setLocationRelativeTo(null);
		this.setResizable(false);
		this.setVisible(true);
	}

	public void actionPerformed(ActionEvent e) 
	{
		if("Submit".equals(e.getActionCommand())) 
		{
			StockInfoGrabber.AccessThreshold at = StockInfoGrabber.AccessThreshold.Week;
			ReutersGrabber rg = new ReutersGrabber( t_queryfield.getText(), at);
			YahooGrabber yg = new YahooGrabber(t_queryfield.getText(), at);
			QuoteComGrabber qcg = new QuoteComGrabber(t_queryfield.getText(), at);
			CNNMoneyGrabber cmg = new CNNMoneyGrabber(t_queryfield.getText(), at);
			MSNGrabber mg = new MSNGrabber(t_queryfield.getText(), at);
			AolGrabber ag = new AolGrabber(t_queryfield.getText(), at);
			NetscapeGrabber ng = new NetscapeGrabber(t_queryfield.getText(), at);

			ArrayList a = new ArrayList();
			a.add(rg);
			a.add(yg);
			a.add(qcg);
			a.add(cmg);
			a.add(mg);
			a.add(ag);
			a.add(ng);

			this.mineInformation(a);
			t_result.setText( this.displayAllInfo(a) );

			for (Object o : a)
				((StockInfoGrabber)o).writeHistoricalContent();
			t_result.setText(t_result.getText() + "\r\n\r\n"+ this.analyzeInfo(a));

			try {
            Class c = Class.forName("StockInfoGrabber");
            Method m[] = c.getDeclaredMethods();
            for (int i = 0; i < m.length; i++)
            System.out.println(m[i].toString());
			 }
			 catch (Throwable exxx) {
            System.err.println(exxx);
			}
		}
	}

	private String displayInfo(StockInfoGrabber sg)
	{
		String s = "Stock: " + sg.getStockSymbol() + " \r\n";
		s += "Industry: " + sg.getIndustry() + "\r\n";
		s += "Current Ratio: " + sg.getCurrentRatio() + "\r\n";
		s += "P/E TTM: " + sg.getPeTTM() + "\r\n";
		s += "Total DE: " + sg.getDeTotal() + "\r\n";
		s += "RE TTM: " + sg.getReTTM() + "\r\n";
		return s;
	}

	private String displayAllInfo(ArrayList a)
	{

		String result="";
		for (Object o : a)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			result+= "\r\n\r\n" + sig.getClass().getName() + " Provides:\r\n";
			if (sig.mineErrorMsg=="")
				result+= this.displayInfo(sig);
			else
				result+= sig.mineErrorMsg;
		}
		return result;
	}

	private void mineInformation(ArrayList a)
	{
		for (Object o : a)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			sig.mineAllInformation();
		}
	}

	private void setAccessThresholds(ArrayList a, StockInfoGrabber.AccessThreshold at)
	{
		for (Object o : a)
		{
			StockInfoGrabber sig = (StockInfoGrabber)o;
			sig.setThreshold(at);
		}
	}


	private String analyzeInfo(ArrayList a)
	{
		BlindAnalyzer ba;
		if (!t_blindprofile.getText().equals(""))
			ba = new BlindAnalyzer(t_blindprofile.getText());
		else
			ba = new BlindAnalyzer();
		for (Object o : a)
			ba.add((StockInfoGrabber)o);

		ba.analyze();
		
		double crMean = ba.cr_mean;//this.calcMean(a, "CR");
		double peMean = ba.pe_mean;//this.calcMean(a, "PE");
		double deMean = ba.de_mean;//this.calcMean(a, "DE");
		double reMean = ba.re_mean;//this.calcMean(a, "RE");

		double crSTD = ba.cr_std;//this.calcSTD(crMean, a, "CR");
		double peSTD = ba.pe_std;//this.calcSTD(peMean, a, "PE");
		double deSTD = ba.de_std;//this.calcSTD(deMean, a, "DE");
		double reSTD = ba.re_std;//this.calcSTD(reMean, a, "RE");

		int crUsed = ba.cr_count;
		int peUsed = ba.pe_count;
		int deUsed = ba.de_count;
		int reUsed = ba.re_count;

		int crOut = ba.cr_outliers.size();
		int peOut = ba.pe_outliers.size();
		int deOut = ba.de_outliers.size();
		int reOut = ba.re_outliers.size();

		double crCon = ba.cr_confidence;
		double peCon = ba.pe_confidence;
		double deCon = ba.de_confidence;
		double reCon = ba.re_confidence;

		String s = "Averages and Confidence:\r\n";
		s+= "CR: " + crMean + " with " + crCon + "% confidence." + "\r\n";
		s+= "PE: " + peMean + " with " + peCon + "% confidence." + "\r\n";
		s+= "DE: " + deMean + " with " + deCon + "% confidence." + "\r\n";
		s+= "RE: " + reMean + " with " + reCon + "% confidence." + "\r\n";

		s+= "\r\nStandard Deviations:\r\n";
		s+= "CR: " + crSTD + "\r\n";
		s+= "PE: " + peSTD + "\r\n";
		s+= "DE: " + deSTD + "\r\n";
		s+= "RE: " + reSTD + "\r\n";

		s+= "\r\nSources Used:\r\n";
		s+= "CR: " + crUsed + "/" + a.size() + "\r\n";
		s+= "PE: " + peUsed + "/" + a.size() + "\r\n";
		s+= "DE: " + deUsed + "/" + a.size() + "\r\n";
		s+= "RE: " + reUsed + "/" + a.size() + "\r\n";

		s+= "\r\nProbable Outliers:\r\n";
		s+= "CR: " + crOut + "\r\n";
		s+= "PE: " + peOut + "\r\n";
		s+= "DE: " + deOut + "\r\n";
		s+= "RE: " + reOut + "\r\n";

		return s;
	}
	
	/*
	public static double sdFast ( double[] data )
	{
		// sd is sqrt of sum of (values-mean) squared divided by n - 1
		// Calculate the mean
		double mean = 0;
		final int n = data.length;
		if ( n < 2 )
		{
			return Double.NaN;
		}
		for ( int i=0; i<n; i++ )
		{
			mean += data[i];
		}
		mean /= n;
		// calculate the sum of squares
		double sum = 0;
		for ( int i=0; i<n; i++ )
		{
			double v = data[i] - mean;
			sum += v * v;
		}
		return Math.sqrt( sum / ( n - 1 ) );
	}
	*/

	public static void main (String[] args)
	{
		WebQueryFrame wqf = new WebQueryFrame();
	}

}