(define MAX-NUMBER-OF-BIDS 3) ;max number of bids returned by single child
(define DEBUG-FLAG #f)
(define DEBUG-FLAG2 #t)
(define border "\n;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n")
(define limitingConstraints '((length . length_max) (DS . DS_min) (cost . cost_max)))

(define (debug . args)
  (if DEBUG-FLAG (begin (map display args) (newline))))

(define (debug2 . args)
  (if DEBUG-FLAG2 (begin (map display args) (newline))))

;A Bid is an assoc list of expected values, with the name of the
;bidder too, e.g. ((bidder . risk)(length_exp . 10) (cost_exp . 3))

;passes initial constraints in to process arg is an assoc list;
;attribute values not appearing in the alist are given a default value
;of NULL. If a constraint is nulll, then it is assumed there is no
;constraint in that direction by intent of the user.
(define (start agent constraints) (agent 'plan constraints))

(define (make-agent name)
  (define children ())
  (define utility ())
  (define local-bids ())
  (define bids-submitted ())
  (define bid-components ())
  (define (dispatch msg . args)
    (cond ((eq? msg 'plan) (apply plan args))
	  ((eq? msg 'name) name)
	  ((eq? msg 'sendRFP) (apply sendRFP args)) ;private?
          ((eq? msg 'acceptRFP) (apply acceptRFP args))
          ((eq? msg 'acceptEnvelope) (apply acceptEnvelope args))
	  ((eq? msg 'execute) (apply execute envelope))
	  ((eq? msg 'set-children) (set! children args))
	  ((eq? msg 'set-utility) (set! utility (car args)))
	  ((eq? msg 'set-local-bids) (set! local-bids (car args)))
	  ((eq? msg 'set-bid-components) (set! bid-components (car args)))
	  ((eq? msg 'children) children)
	  ((eq? msg 'utility)  utility)
	  ((eq? msg 'local-bids) local-bids)
	  ((eq? msg 'bid-components) bid-components) ;how bids decompose into child bids
	  ((eq? msg 'g) (apply generate-sets-of-bids args)) ;for debugging
	  (else (error "unknown method in " name))
          ))

  (define (plan constraints)
    (debug2 border name " starting to plan for constraints: " constraints border)
    (let* ((bidset (acceptRFP constraints)))
      (cond ((null? bidset) (debug2 name ": no bids received for constraints: " constraints) 'NoBidsReceived)
	    (else
	     (sendEnvelopes (car bidset) 
			    (if (null? (cdr bidset)) () (cadr bidset)) 
			    children 
			    constraints)))))

  ;;returns a list of a set of bids from each child agent
  (define (sendRFP constraints children)
    (debug name " sendRFP, constraints, children: " constraints " " children)
    (map (lambda (child) (child 'acceptRFP constraints)) children))      

  (define (acceptRFP constraints)
    (debug2 border name " acceptRFP, constraints: " constraints border )
    (let ((bids (return-bids constraints)))
      (set! bids-submitted bids)
      bids))

  ;returns a bidset
  (define (return-bids constraints)
    (let* ((bidsets (generate-sets-of-bids constraints)) ;list of bidsets - sends RFP to children
	   (bids (make-single-bids-from-sets bidsets)) ;this is a bidset
	   (ok-bids ;;bids that fall within constraints
	    (filter (lambda (b) (within-constraints? b constraints)) bids)))
      (set! bid-components (pair-assoc bids bidsets))
      (debug2 border name " return-bids will return: " (choose-n-bids ok-bids MAX-NUMBER-OF-BIDS) border )
      (choose-n-bids ok-bids MAX-NUMBER-OF-BIDS)))

  ;check envelope(?) accept bid with envelope
  (define (acceptEnvelope bid envelope constraints) 
    (debug2 border name " accepting envelope: " envelope "\n     for bid: " bid "\n     with constraints: " constraints border)
    (if (not (null? children))
	;(plan envelope)))
	(sendEnvelopes (car bids-submitted) 
			    (if (null? (cdr bids-submitted)) () (cadr bids-submitted)) 
			    children 
			    constraints)))

  ;;Look at chosen bid. 
  (define (sendEnvelopes chosenBid altBid children constraints)
    (define envelope-alist (empty-alist))
    (debug2 name " sendEnvelope, chosen bid, children: " chosenbid " "
	   children)
    (debug2 name " sendEnvelope bid-components " bid-components)
    (let ((childConstraints 
	   (calcChildConstraints children chosenBid bid-components constraints)))
      (let loop ((child-list children))
	(cond ((null? child-list) 
	       (for-each (lambda (c-e) (if (not (null? (car c-e))) 
					   ((car c-e) 
					    'acceptEnvelope 
					    (findChildsPartOfBid (car c-e) chosenBid)
					    (cdr c-e)
					    constraints)))
			 envelope-alist))
	      (else
	       (let* 
		   ((thisChildsConstraints 
		     (let ((c (assoc (car child-list) childConstraints ))) (if c (cdr c) ())))
		    (childsBid (begin (debug2 "setting childsbid")
				      (findChildsPartOfBid (car child-list) chosenBid)))
		    (childsAltBid (findChildsPartOfBid (car child-list) altBid))
		    (env (calcEnvelope (car child-list) childsBid childsAltBid thisChildsConstraints)))
		 (debug2 name " SendEnvelopes child: " ((car child-list) 'name)) 
		 (debug2 name " SendEnvelopes chosenBid: " chosenBid) 
		 (debug2 name " SendEnvelopes childsBid: " childsBid) 
		 (debug2 name " SendEnvelopes childsAltBid: " childsAltBid) 
		 (debug2 name " SendEnvelopes this child's constraints: " thisChildsConstraints) 
		 (debug2 name " SendEnvelopes envelope: " env)
		 (if (not (null? env))
		     (acons (car child-list) env envelope-alist))
		 (debug2 name " SendEnvelopes envelope-alist: " envelope-alist)
		 (loop (cdr child-list) )))))))

  ;returns an alist of (child . constraints) pairs
  (define (calcChildConstraints children chosenBid bid-components constraints)

    (define (calcTotalExpecteds valuesSoFar childBids)
      (debug "calcTotalExpecteds valuesSoFar childBids: " valuesSoFar " " childBids)
      (cond ((null? childBids) valuesSoFar)
	    (else (let 
		      ((length (cdr (assoc 'length_exp (car childbids))))
		       (DS (cdr (assoc 'DS_exp (car childbids))))
		       (utility (getUtilityFromBid (car childbids))))
		    (acons 'length_exp (+ length (cdr (assoc 'length_exp valuesSoFar))) valuesSoFar)
		    (acons 'DS_exp (+ DS (cdr (assoc 'DS_exp valuesSoFar))) valuesSoFar)
		    (acons 'utility (+ utility (cdr (assoc 'utility valuesSoFar))) valuesSoFar)
		    (debug2 "calcTotalExpecteds final valuesSoFar : " valuesSoFar))
		  (calcTotalExpecteds valuesSoFar (cdr childBids)))))

    (define (getConstraintToCheck name)
      (cdr (assoc name limitingConstraints)))

    ;Excess is the amount by which a resource constraint exceeds the
    ;expected value. For maxes, max - exp; for mins, exp - min. If a
    ;bid is within constraints, both values will be positive.
    (define (calcExcesses excessAlist totalAlist)
      (debug "calcExcesses starting: excess, total: " excessAlist " " totalAlist)
      (acons 'length_max (- (cdr (assoc 'length_max constraints) )
			    (cdr (assoc 'length_exp totalAlist)))
	     excessAlist)
      (acons 'length_min (- (cdr (assoc 'length_exp totalAlist))
			    (cdr (assoc 'length_min constraints) ))
	     excessAlist)
      (acons 'DS_max (- (cdr (assoc 'DS_max constraints)) 
			(cdr (assoc 'DS_exp totalAlist)))
	     excessAlist)
      (acons 'DS_min (- (cdr (assoc 'DS_exp totalAlist))
			(cdr (assoc 'DS_min constraints)) )
	     excessAlist)
      (debug "calcExcesses returning: " excessAlist)
      excessAlist)

    ;next time try looping on the components in the bid
    ;not every child necessarily participates in a bid. Only participating children appear in the resulting alist.
    (define (child-iter parentbid children child-alist excesses totals)
      (if (null? children) 
	  (begin (debug2 "child-iter returning alist of children, constraints: " child-alist) child-alist)
	  (let ((child (car children))
		(bid (findChildsPartOfBid (car children) parentBid)))
	    (if (null? bid)
		(child-iter parentbid (cdr children) child-alist excesses totals)
		(let
		    ((length (cdr (assoc 'length_exp bid)))
		     (DS (cdr (assoc 'DS_exp bid)))
		     (utility (getUtilityFromBid bid))
		     (length_max (cdr (assoc 'length_max constraints)))
		     (length_min (cdr (assoc 'length_min constraints))) 
		     (DS_max (cdr (assoc 'DS_max constraints))) 
		     (DS_min (cdr (assoc 'DS_min constraints)))
		     (utilityTotal (cdr (assoc 'utility totals)))
		     (child-constraints (empty-alist)))
		  (debug2 "child-iter totals " totals)
		  (debug2 "child-iter constraints " constraints)
		  (if (null? length_max)
		      (acons 'length_max null child-constraints)
		      (acons 'length_max
			     (+ length (* (/ utility utilityTotal)
					  (cdr (assoc 'length_max excesses))))
			     child-constraints))
		  (if (null? length_min)
		      (acons 'length_min null child-constraints)
		      (acons 'length_min
			     (- length (* (/ length (cdr (assoc 'length_exp totals)))
					  (cdr (assoc 'length_min excesses))))
			     child-constraints))
		  (if (null? DS_max)
		      (acons 'DS_max null child-constraints)
		      (acons 'DS_max
			     (+ DS (* (/ DS (cdr (assoc 'DS_exp totals)))
				      (cdr (assoc 'DS_max excesses))))
			     child-constraints))
		  (if (null? DS_min)
		      (acons 'DS_min null child-constraints)
		      (acons 'DS_min
			     (- DS (* (/ utility utilityTotal)
				      (cdr (assoc 'DS_min excesses))))
			     child-constraints))
		  (acons child child-constraints child-alist)
		  (child-iter parentbid (cdr children) child-alist excesses totals))))))

    (debug "calcChildConstraints children chosenBid bid-components constraints: " 
	   children chosenBid bid-components constraints)
    (debug "(cdr (assoc chosenBid bid-components)): " (getBidParts chosenBid bid-components))
    (let ((child-alist (empty-alist))
	  (values-so-far (empty-alist))
	  (excesses (empty-alist)))
      (acons 'length_exp 0 values-so-far)
      (acons 'DS_exp 0 values-so-far)
      (acons 'utility 0 values-so-far)
      (calcTotalExpecteds values-so-far (getBidParts chosenBid bid-components))
      (calcExcesses excesses values-so-far)
      (child-iter chosenBid children child-alist excesses values-so-far)))
;;;end calcChildConstraints

   ;returns child's bid to parent or null.
  (define (findChildsPartOfBid child bid)
    (debug2 name " calling findChildsPartOfBid child  bid bid-components " (child 'name) " for " bid)
    ;(debug2 " in "  bid-components) 
    (let* ((childName (child 'name))
	  (allChildBids (getBidParts bid bid-components))
	  (result (memberf allChildBids 
			   (lambda (x) (equal? childName (cdr (assoc 'bidder x)))))))
      (debug2 "findChildsPartOfBid found bid " result " for " (child 'name) " in " allChildBids)
      result))

  (define (calcEnvelope child childsBid childsAltBid constraints)
    ;(let ((childsBid (findChildsPartOfBid child parentBid))
	;  (childsAltBid (findChildsPartOfBid child altParentBid)))
    (debug2 name " calling (calcEnvelope child childsBid childsAltBid constraints):")
    (debug2 (child 'name))
    (debug2  childsbid )
    (debug2 childsAltBid)
    (debug2 constraints)
      (cond ((null? constraints) ())
	    ((and (not (null? childsBid)) (not (null? childsAltBid)))
	     (calcEnvelopeFromTwoBids child childsBid childsAltBid constraints))
	    ((not (null? childsBid)) constraints)
	    (else (debug "ERROR " name " calling (calcEnvelope child childsBid childsAltBid constraints): ")
		  (debug (child 'name))
		  (debug  childsbid )
		  (debug childsAltBid)
		  (debug constraints))))
	  

  ;Before this is called, have to decompose parent's bid to find the
  ;bid from this child AND distribute constraints across all children.
  ;
  ;When min or max of constraints are null, then user has indicated no
  ;bound in that direction, and envelope should not restrict values in
  ;that direction (i.e. if max is null, then more is better). When
  ;there is a min or max, check to see if the alternate bid has a
  ;value between the constraint and the expected value of the chosen
  ;bid.
  (define (calcEnvelopeFromTwoBids child chosenBid altBid constraints)
    (debug2 name " calling calcEnvelopeFromTwoBids on " (child 'name))
    (debug2 "calcEnvelopeFromTwoBids chosenBid: " chosenBid)
    (debug2 "calcEnvelopeFromTwoBids altBid: " altBid)
    (debug2 "calcEnvelopeFromTwoBids constraints: " constraints)
    (let ((length (cdr (assoc 'length_exp chosenBid)))
	  (DS (cdr (assoc 'DS_exp chosenBid)))
	  (altLength (cdr (assoc 'length_exp altBid)))
	  (altDS (cdr (assoc 'DS_exp altBid)))
	  (length_max (cdr (assoc 'length_max constraints)))
	  (length_min (cdr (assoc 'length_min constraints))) 
	  (DS_max (cdr (assoc 'DS_max constraints))) 
	  (DS_min (cdr (assoc 'DS_min constraints)))
	  (envelope (empty-alist)))
      
      ;length_min is non-critical
      (acons 'length_min length_min envelope)
      ;length_max is critical
      (if (and (not (null? length_max)) (> length_max altLength)(> altLength length))
	  (acons 'length_max altLength envelope)
	  (acons 'length_max length_max envelope))
      ;non-critical
      (acons 'DS_max DS_max envelope)
      (if (and (not (null? DS_min)) (> DS altDS)(> altDS DS_min))
	  (acons 'DS_min altDS envelope)
	  (acons 'DS_min DS_min envelope))

      (debug2 "calcEnvelopeFromTwoBids envelope: " envelope)
      envelope))
  
  ;Returns a single bidset. Takes a set of bidsets.
  ;set of 3 bidsets ({1 bid}{2 bids}{1 bid})looks like this (note nesting):
  ;((((bidder . risk) (length_exp . 10) (ds_exp . 3))) 
  ; (((bidder . risk) (length_exp . 10) (ds_exp . 3)) 
  ;  ((bidder . industry-risk) (length_exp . 5) (ds_exp . 2))) 
  ; (((bidder . industry-risk) (length_exp . 5) (ds_exp . 2)))) 
  ;which returns a bidset of three bids:  
  ;(((bidder . RISK1) (length_exp . 10) (ds_exp . 3))
  ; ((bidder . RISK1) (length_exp . 15) (ds_exp . 5)) 
  ; ((bidder . RISK1) (length_exp . 5) (ds_exp . 2)))
  (define (make-single-bids-from-sets bidsets)
    (define (iter bidset length DS)
      (cond ((null? bidset) 
	     (list (cons 'bidder name)(cons 'length_exp length)(cons 'DS_exp DS)))
	     (else 
	      (let ((newlength (cdr (assoc 'length_exp (car bidset))))
		    (newDS (cdr (assoc 'DS_exp (car bidset)))))
		(iter (cdr bidset) (+ length newlength) (+ DS newDS))))))
    (debug "make-single-bids-from-sets: " bidsets)
    (let ((result (map (lambda (set) (iter set 0 0)) bidsets)))
      (debug "make-single-bids-from-sets returning: " result)
      (map (lambda (set) (iter set 0 0)) bidsets)))

;returns a list of lists of bids, where each outermost list represents
;all the bids from a single source
;(((length_exp . 10)) ((cost_exp . 3))
  (define (generate-sets-of-bids constraints)
    (debug name " generate-sets-of-bids, constraints: " constraints )
    (let ((result 
	   (if (< 25 (percent-time-elapsed)) ;;if over 25% of the time is gone, short-cut the bid process
	       (generate-local-bids constraints)
	       (make-child-bid-combos (cons (generate-local-bids constraints)
					    (sendRFP constraints children))))))
	  (debug name " generate-sets-of-bids returning: " result )
	  result))

  (define (percent-time-elapsed) 10)
  
  ;returns a list of BidSets
  (define (make-child-bid-combos bidsets)
    (debug name " make-child-bid-combos, BidSets: " bidsets )
    (debug name " make-child-bid-combos will return: " (choose-1-combos bidsets))
    (choose-1-combos bidsets)
    )

	 
  ;;return all combinations possible by taking 0-1 element from each set in list
  ;> (choose-1-combos '((1 2 3) (a b)))
  ;((1) (2) (3) (1 a) (1 b) (2 a) (2 b) (3 a) (3 b) (a) (b))
  (define (choose-1-combos list-of-sets)
	  (debug "choose-1-combos list-of-sets: " list-of-sets)
    (if (null? (cdr list-of-sets)) (map list (car list-of-sets))
	(let ((new-sets (choose-1-combos (cdr list-of-sets))))
	  (debug "choose-1-combos new sets: " new-sets)
	  (append (map list (car list-of-sets))
		  (apply append 
                         (map (lambda (x) (map (lambda (y) (cons x y)) 
                                               new-sets)) 
                              (car list-of-sets)))
		  new-sets))))

  ;;returns set of bids that can be made without consulting children
  (define (generate-local-bids constraints)
					;use cached info to make one or more bids
    local-bids
    )

  ;;should find highest U, second highest(?), then most unlike
  (define (choose-n-bids bids number-of-bids)
    (define (get-first-n elements count)
      (if (or (null? elements) (<= count 0)) ()
	  (cons (car elements) (get-first-n (cdr elements) (- count 1)))))
    ;(debug "choose-n-bids: " number-of-bids " " bids)
    (let ((sorted-bids (insertsort bids utility-greater-than?)))
      ;(debug "choose-n-bids sorted: " sorted-bids)
      (get-first-n sorted-bids number-of-bids)))

  (define (within-constraints? bid constraints)
    (debug "within-constraints? bid " bid) 
    (let ((length (cdr (assoc 'length_exp bid)))
	  (DS (cdr (assoc 'DS_exp bid)))
	  (length_max (cdr (assoc 'length_max constraints)))
	  (length_min (cdr (assoc 'length_min constraints))) 
	  (DS_max (cdr (assoc 'DS_max constraints))) 
	  (DS_min (cdr (assoc 'DS_min constraints))))
					;check constraints
      (if (or (and (not (null? length_max)) (> length length_max))
	      (and (not (null? length_min)) (< length length_min))
	      (and (not (null? DS_max)) (> DS DS_max))
	      (and (not (null? DS_min)) (< DS DS_min)))
	  #f
	  #t)))

;assumes bid is within constraints
  (define (calc-utility-bid bid) 
    (let ((u (assoc 'utility bid)))
      (if u 
	  (cdr u)
	  (let ((length (cdr (assoc 'length_exp bid)))
		(DS (cdr (assoc 'DS_exp bid))))
	    (acons 'utility (utility length DS) bid)
	    (utility length DS)))))
  
  (define (getUtilityFromBid bid)
    (let ((u (assoc 'utility bid)))
      (if u (cdr u) (calc-utility-bid bid))))

  (define (utility-greater-than? bid1 bid2)
    (> (calc-Utility-bid bid1) (calc-Utility-bid bid2)))
    
  dispatch
  );make-agent

(define (acons name value alist)
  (let ((current (assoc name alist)))
    (cond ((null? alist) (error "null alist in acons"))
	  (current (set-car! (member current alist) (cons name value)))
	  (else (let ((temp (car alist)))
		  (set-car! alist (cons name value))
		  (set-cdr! alist (cons temp (cdr alist))))))))
	  
;(set-cdr! alist (cons (cons name value) ())))))
(define (empty-alist) (list (cons () ())))


(define (insert ele numlist greater-than?)
  (cond ((null? numlist) (list ele))
        ((greater-than? ele (car numlist))
         (cons ele numlist))
	(else (cons (car numlist) (insert ele (cdr numlist) greater-than?)))))

(define (insertsort elements greater-than?)   
  (accumulate (lambda (x rest) (insert x rest greater-than?)) 
	      () elements))

(define (accumulate op init alist)
  (if (null? alist) init
      (op (car alist) (accumulate op init (cdr alist)))))


(define (filter pred? elements)
  (cond ((null? elements) ())
	((pred? (car elements)) (cons (car elements) (filter pred? (cdr elements))))
	(else (filter pred? (cdr elements)))))

(define (getBidParts bid bid-components) ;pair-assoc makes list of second element
  (let ((result (assoc bid bid-components)))
    (if result (cadr result)
	())))

(define (pair-assoc lista listb)
  (debug "pair-assoc: \n" lista "\n" listb)
  (cond ((null? lista) ())
	(else (cons (cons (car lista) (list (car listb)))
		    (pair-assoc (cdr lista) (cdr listb))))))

;returns element which is true for pred (f element), or null.
(define (memberf lst f)
  (cond ((null? lst) ())
	((f (car lst)) (car lst))
	(else (memberf (cdr lst) f))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define c1 '((length_min . 3) (length_max . 30)(DS_min . 0)(DS_max . 100)))
(define b1 '((length_exp . 20) (DS_exp . 10)))
(define (u1 length DS) (+ (- (* length 2)) (* DS 5)))

(define industry-risk1 (make-agent 'industry-risk1))
(industry-risk1 'set-utility u1)
(industry-risk1 'set-local-bids '(((bidder . industry-risk1)(length_exp . 5) (DS_exp . 2))))

(define debt-risk1 (make-agent 'debt-risk1))
(debt-risk1 'set-utility u1)
(debt-risk1 'set-local-bids '(((bidder . debt-risk1)(length_exp . 25) (DS_exp . 20))
			   ((bidder . debt-risk1)(length_exp . 15) (DS_exp . 10))
			   ((bidder . debt-risk1)(length_exp . 3) (DS_exp . 0))))

(define risk1 (make-agent 'RISK1))
(risk1 'set-utility u1)
(risk1 'set-local-bids '(((bidder . risk1)(length_exp . 10) (DS_exp . 3))))

(risk1 'set-children industry-risk1 debt-risk1)


